#!/bin/sh
# usage: build.sh [makefile_target]
#
# Shell script to start a top-level make.
#
# RCS: $Header: /vol/cooley/cooley1/kennard/src/xpole-dist/root/RCS/build.sh,v 1.3 93/08/29 18:02:41 kennard Exp Locker: kennard $
#

# The full path name of this directory.  This will not be "hardcoded" anywhere;
# it only used to find ourselves while cd'ing around.
ThisDir=`pwd`

# The makefiles required GNU Make.  If your "make" is not gnumake, then
# change the line below.
GnuMake=${GnuMake-make}

# This tells the makefiles where to find its auxiliary makefiles
# and scripts.  It must be a full path.
MakeDir=${MakeDir-$ThisDir/etc/make}

# The architecture you are running on (e.g., "sun4", "mips").  If the script
# used below (arch.sh) fails, please tell me about your arch.
ARCH=${ARCH-`$MakeDir/arch.sh`}

# The sub-class of the architecture to you want.  Subclasses include:
# "opt"= optimized, "dbg"=debug, "pur"=purify.  (examples: sun4.opt, mips.dbg)
MakeArch=${MakeArch-$ARCH.opt}

if [ ! -r $MakeDir/$MakeArch.arch ] ; then
    echo "The architecture \`\`$MakeArch'' is not yet defined."
    echo "Please create an arch file \`\`$MakeArch.arch'' in the"
    echo "directory \`\`$MakeDir'' and try again."
    exit 1
fi

# Examine the file ./etc/make/std.conf, especially the section titled
# "External Package Directories".  The following symbols must be
# properly configured for XPole (the rest may be disregarded):
#    TCL_INCLUDE, TCL_DASHLIBS		For Tcl versions 6.6 or 6.7
#    TK_INCLUDE, TK_DASHLIBS		For Tk versions 3.1 or 3.2
#    TCLTCP_INCLUDE, TCLTCP_DASHLIBS	For the "tcpTCP" extension
#					  version 1.0 by Kevin Kenny.
# Note: XPole will work best if the "focus3.1a" patch to tk is made.
# Note: XPole requires the TkPixmap1.5 patch to tk.
#
# Copy std.conf to xxx.conf, and edit xxx.conf: delete everything that is
# correct and modify everything that is incorrect.  Note that first
# std.conf will be included, then your xxx.conf.
MakeConf=${MakeConf-xxx}

# Some advanced features of the make files can be enabled by setting
# MakeExpert non-empty.  If you're using this script, you dont want to do that.
MakeExpert=


export ARCH MakeDir MakeArch MakeConf MakeExpert

if [ "$*" = "" ] ; then
    exec $GnuMake -f mf.build All
else
    # The syntax below invokes gnumake exactly as we were called.
    exec $GnuMake -f mf.build "$@"
fi
exit	# just-in-case

###############################################################################
#
#	Installing XPole from source distribution
#
###############################################################################
#
# After building XPole (using "build.sh All"), you will most likely want
# to install it somewhere so you can remove the source code.  Before
# installing, you should build the docs in src/xpole/doc.  Note that
# these are not built by starting a build at the top-level; this is because
# many people dont have latex/latexinfo/emacs installed.

To install XPole, you have 4 options:
#  a)	build.sh ptinstall
#	Install XPole into your local Ptolemy tree (~ptolemy).
#  b)	build.sh tmpinstall
#	Install XPole into /tmp.  You can then tar and/or copy it
#	into where ever you eventually want it to go.
#  c)	build.sh relinstall
#	Install XPole into the source tree (creates bin and lib dirs at
#	the top level).
#  d)	build.sh reltarz
#	Do a relinstall, and then create a binary distribution.
#
# In case the makefiles fail to work and you want to do it manually, here
# is what must be installed:
#  a)	src/xpole/<arch>/xpole, the binary itself
#  b)	src/xpole/lib/*.tcl, some scripts
#  c)	src/tkgraph/lib/*.tcl, some more scripts
#  d)	src/xpole/doc/*/{*.ps,*.info*}, help files for xpole
#  e)	lxpole/pxpole, a driver/config script based on lxpole.{rel/abs}.
#  Items (b-d) should all be installed in the same directory, and
#  the environment variable XPOLE_LIBRARY set to that directory.


###############################################################################
#
#	Compile problems?
#
###############################################################################
#
# 1. When I run "make" or "build.sh" it says:
#	"make: Fatal error in reader: Makefile, line 8: Unexpected ..."
#    This is because you are used normal make, not gnumake.  Get gnumake,
#    and set GnuMake above.
# 2. When I run "make" or "build.sh" it says:
#	"Making DependInfo in xxx"
#    or
#	"Makefile is out of date, remaking in xxx"
#    This shouldn't happen; if it does, it means that somehow the files
#    in etc/make/ are newer than the Makefiles elsewhere.  This is not
#    necessarily fatal: it will chug for a while and proceed.  However,
#    the scripts used to do this are not as portable and may break.
#    If they do break, restore the Makefile and DependInfo files and
#    "touch" them so that they are newer.
# 3. When I run "make" or "build.sh" it says:
#	"make: fopen: xxx.yyy/DependInfo: No such file or directory"
#    This is not a problem: it is part of the normal operation of the
#    makefiles the first time you build.
# 4. What other commands are available?  A short list:
#	build.sh Depend		Build *.h file dependencies.
#	build.sh Clean		Remove *.o, libs and bins for cur arch.
#	build.sh Pristine	Remove *.o, libs and bins for all archs.
#    Also see above for install commands.	

# The End
