#!/bin/sh

# This script "wraps" a variety of different types of source code.  These
# are described below.  The wrap'ing is idempotent: no change will be made
# if the file is already properly wrapped.
# .h files:	Adds "ifdef/endif" around file so that it wont be processed
#		more than once.  Adds rcsid/sccsid inside comments.
# .c files:	Adds rcsid/sccsid inside proper condidionals.

MakeDir=${MakeDir-$HOME/etc/make}

CUSTOMKEY=NOWRAPKEY

RCS_HEADER='$Header'

if [ "$1" = "" ] ; then
    echo Adding wraps to all .h files.
    exec $0 *.h
    exit 0
fi

SAVE=\#save$$
mkdir $SAVE

while [ "$1" != "" ] ; do
    F=$1
    FP=$1.$$
    shift
    # we dont want to remove the existing file, or the perms&ownerships
    # will get mixed up
    if [ ! -r "$F" ] ; then
	echo $F: Doesnt exist: skipping...
	continue
    fi
    if [ "$F" = "copyright.h" ] ; then
	continue
    fi
    if grep -s $CUSTOMKEY $F ; then
	echo $F: Custom wrap: skipping...
	continue
    fi
    if [ "`egrep '(Copyright \(c\))|(CUSTOMCOPYRIGHT)' $F`" = "" ] ; then
	echo $F: Missing copyright, adding before wrapping...
	$MakeDir/copyright.sh $F
    fi
    DO_IFDEF=
    DO_COMMENT=
    DO_HID=
    DO_CID=
    DO_AUTHOR=
    COMSTART='\+^/\*[ \*]*$+'
    COMEND='\+\*/$+'
    TEXT_author=`grep Author: $F`
    if [ "$TEXT_author" = "" ] ; then
	DO_AUTHOR=yes;
    fi
    TEXT_rcsid=`grep "$RCS_HEADER:" $F`
    case "$F" in
    *.h)
	newtag=`echo _${F}_ | tr '.+a-z' '_XA-Z'`
	headline=`head -1 $F`
	oldtag=`echo "$headline" | sed -n -e 's/^#if !defined(\([^)]*\))$/\1/p'`
	if [ "$newtag" != "$oldtag" ] ; then
	    if [ "$oldtag" != "" ] ; then
		DO_STRIP_IFDEF=yes
	    fi
	    DO_IFDEF=yes
	fi
	if [ "$TEXT_rcsid" = "" ] ; then
	    DO_HID=yes
	fi
	;;
    *.c)
        TEXT_rcsid=`grep rcsid $F`
        if [ "$TEXT_rcsid" = "" ] ; then
	    DO_CID=yes
	fi
	;;
    *)
	echo $F: Dont know how to wrap: skipping...
	continue
	;;
    esac


    if [ "$DO_IFDEF" -o "$DO_AUTHOR" -o "$DO_HID" -o "$DO_CID" ] ; then
	echo "$F: Wrapping (ifdef=$DO_IFDEF com=$DO_COMMENT author=$DO_AUTHOR hid=$DO_HID cid=$DO_CID)..."
    else
    	echo $F: Pre-wrapped: skipping...
	continue
    fi

    sccsflag=0; rcsflag=0
    if [ ! -w "$F" ] ; then
	if [ -d SCCS ] ; then
	    sccs edit $F
	    sccsflag=1
	elif [ -d RCS ] ; then
	    co -l $F
	    rcsflag=1
	fi
        if [ ! -w $F ] ; then
	    echo $F: is not writable: cant insert copyright
	    continue
	fi
        echo $F: checked out to add copyright
    fi
    if [ -d $SAVE ] ; then
        cp $F $SAVE/$F
	if [ ! -s $SAVE/$F ] ; then
	    echo $F: Save file empty...disk may be full
	    continue
	fi
    fi


    #
    # Stuff that comes before the comment area
    #
    if [ "$DO_IFDEF" ] ; then
        cp $F $FP
	echo "#if !defined($newtag)"	>  $F
	echo "#define $newtag 1"	>> $F
	if [ "$DO_STRIP_IFDEF" ] ; then
	    sed -e '1,2 d' -e '$ d' $FP	>> $F
	else
	    cat $FP			>> $F
	fi
	echo "#endif /* $newtag */"	>> $F
	rm -f $FP
    fi

    #
    # Make the comment area
    #
    COMSTART=`head -10 $F | sed -n -e "$COMSTART =" | head -1`
    if [ "$COMSTART" ] ; then
        COMEND=`tail +$COMSTART $F | head -100 | sed -n -e "$COMEND =" | head -1`
	if [ "$COMEND" = "" ] ; then
	    DO_COMMENT=yes
	fi
    else
	DO_COMMENT=yes
    fi
#    echo coms XXX${COMSTART}XXX YYY${COMEND}YYY
    if [ "$DO_COMMENT" ] ; then
        cp $F $FP
	case "$F" in
	*.h)
	    head -2 $FP			> $F
	    echo "/*"			>> $F
	    echo "*/"			>> $F
	    tail +3 $FP			>> $F
	    COMSTART=3; COMEND=4
	    ;;
	*.c)
	    echo "/*"			> $F
	    echo "*/"			>> $F
	    cat $FP			>> $F
	    COMSTART=1; COMEND=2
	    ;;
	esac
	rm -f $FP
    fi

    #
    # Stuff that goes in the comment block
    #
    COMBLK=/tmp/comblk.$$; cp /dev/null $COMBLK
    if [ "$DO_AUTHOR" ] ; then
	echo "    Author: (maybe $USER)"	>> $COMBLK
	echo "$F: Inserting author, please fix."
    fi
    if [ "$DO_HID" ] ; then
	echo "    RCS: $RCS_HEADER:$"	>> $COMBLK
    fi

    #
    # Stuff that goes after the comment block
    #
    AFTBLK=/tmp/aftblk.$$; cp /dev/null $AFTBLK
    if [ "$DO_CID" ] ; then
	echo "#if !defined(lint) && !defined(SABER)"	>> $AFTBLK
	# The funny ``H\e'' is to keep RCS from screwing with this file
	echo "static char rcsid[] = \"$RCS_HEADER:$\";"	>> $AFTBLK
	echo "#endif"					>> $AFTBLK
    fi
    if [ -s $COMBLK -o -s $AFTBLK ] ; then
#	echo "Merging at $COMSTART and $COMEND"
	cp $F $FP
	sed -e "$COMSTART r $COMBLK" -e "$COMEND r $AFTBLK" $FP > $F
	rm -f $FP
    fi

    rm -f $FP $COMBLK $AFTBLK

    if [ "$sccsflag" = "1" ] ; then
	sccs delget $F -ycopyright
    elif [ "$rcsflag" = "1" ] ; then
	ci -mwrap $F
    fi
done
