/* 
    Author: Kennard White

    Copyright (c) 1990-93 The Regents of the University of California.
    All rights reserved.  See the file "./copyright.h" for full copyright
    notice, limitation of liability, and disclaimer of warranty provisions.
#include "copyright.h"				COPYRIGHTENDKEY
*/
#if !defined(lint) && !defined(SABER)
static char rcsid[] = "$Header: /vol/cooley/cooley1/kennard/src/eem/RCS/cplxTcl.c,v 1.14 93/08/27 14:29:41 kennard Exp $";
#endif

#include "topFixup.h"
#include <stdarg.h>
#include <tcl.h>

#include "topStd.h"
#include "topTcl.h"
#include "topMisc.h"

#include "eemDflt.h"
#include "eemCplx.h"

int
eemCplxPrsTcl( Tcl_Interp *ip, EEMComplex *pC, int aC, char **aV) {
    TOPLogical		degB = TOP_FALSE, mpB = TOP_FALSE;
    char		*cvtcode = NULL;

    pC->r = 0.0;
    pC->i = 0.0;
    if ( aC >= 2 && isalpha(aV[0][0]) ) {
	char *code = aV[0];
	char code0 = TOP_ToLower(code[0]);
	char code1 = TOP_ToLower(code[1]);
	char code2 = TOP_ToLower(code[2]);
	if ( ! (code[1] == '\0' || code[2] == '\0' || code[3] == '\0') )
	    code0 = '\0';	/* force an error */
	/*IF*/ if ( code0 == 'r' && (code1=='i'||code1=='\0') ) {
	} else if ( code0 == 'm' && (code1=='p'||code1=='\0') ) {
	    mpB = TOP_TRUE;
	    if ( code2 == 'r' )		degB = TOP_FALSE;
	    if ( code2 == 'd' )		degB = TOP_TRUE;
	} else {
	    Tcl_AppendResult(ip,"bad coordinate-system code ``",code,"''",NULL);
	    return TCL_ERROR;
	}
	--aC; ++aV;
    }
    if ( aC == 0 ) {
	Tcl_AppendResult(ip,"complex number format:\t?ri? real ?imag?\n",
	  "\t\t  or:\tmp?r|d? mag ?phase?",NULL);
	return TCL_ERROR;
    }
    if ( aC > 0 ) {
	if ( topPrsDblTcl( ip, aV[0], &pC->r) != TCL_OK ) {
	    Tcl_AppendResult(ip,"\nbad ",(mpB?"magnitude":"real")," part",NULL);
	    return TCL_ERROR;
	}
	--aC; ++aV;
    }
    if ( aC > 0 ) {
	char	*str = aV[0], *code, oldchar, trychar;

	for ( code=str+strlen(str)-1; code>=str && isalpha(*code); code--)
	    ;
	oldchar = *++code; trychar = TOP_ToLower(oldchar);
	if ( trychar=='r' || trychar=='d' || trychar=='i' ) {
	    cvtcode = code; *code = '\0';
	} else if ( *code != '\0' && isalpha(code[1]) ) {
	    oldchar = *++code; trychar = TOP_ToLower(oldchar);
	    if ( trychar=='r' || trychar=='d' || trychar=='i' )
	        cvtcode = code; *code = '\0';
	}
        if ( topPrsDblTcl( ip, aV[0], &pC->i) != TCL_OK ) {
	    *code = oldchar;
	    Tcl_AppendResult(ip,"\nbad ",(mpB?"phase":"imag")," part",NULL);
	    return TCL_ERROR;
	}
	*code = oldchar;
        --aC; ++aV;
    }
    if ( cvtcode == NULL && aC > 0 ) {
	cvtcode = aV[0];
	--aC; ++aV;
    }
    if ( cvtcode != NULL ) {
	int	len = strlen(cvtcode);
	if ( len==0 ) {
	    ;
	} else if ( strncasecmp(cvtcode,"imaginary",len)==0 ) {
	    mpB = TOP_FALSE;
	} else if ( strncasecmp(cvtcode,"radians",len)==0 ) {
	    mpB = TOP_TRUE; degB = TOP_FALSE;
	} else if ( strncasecmp(cvtcode,"degrees",len)==0 ) {
	    mpB = TOP_TRUE; degB = TOP_TRUE;
	} else {
	    Tcl_AppendResult(ip,"bad code ``",cvtcode,"''",NULL);
	    return TCL_ERROR;
	}
    }
    if ( aC != 0 ) {
	Tcl_AppendResult(ip,"Junk at end of complex number ``",aV[0],"''",NULL);
	return TCL_ERROR;
    }
    if ( mpB ) {
	if ( degB )
	    pC->i *= (M_PI/180.0);
	eemCplxRealImag( pC, pC);
    }
    return TCL_OK;
}

int
eemCplxStrPrsTcl( Tcl_Interp *ip, char *arg, EEMComplex *pC) {
    int			sC;
    char		**sV;

    if (Tcl_SplitList( ip, arg, &sC, &sV) != TCL_OK) {
	Tcl_AppendResult(ip,"\nBad complex number ``",arg,"'' (list)", NULL);
	return TCL_ERROR;
    }
    if ( eemCplxPrsTcl( ip, pC, sC, sV) != TCL_OK ) {
	free( (char*) sV);
	Tcl_AppendResult(ip,"\nBad complex number ``",arg,"'' (prs)",NULL);
	return TCL_ERROR;
    }
    free( (char*) sV);
    return TCL_OK;
}

int
eemCplxFmtTcl( Tcl_Interp *ip, EEMComplex *pC, TOPLogical NestB, char *fmt) {
    char	buf[100];

    eemCplxFmt( buf, pC, fmt);
    if ( NestB )	Tcl_AppendElement( ip, buf, 0);
    else		Tcl_AppendResult( ip, buf, NULL);
    return TCL_OK;
}

static int 
_eemCplxCmds( ClientData cld, Tcl_Interp *ip, int aC, char **aV) {
    int			opts = 1, len, i;
    TOPLogical		mpB = TOP_FALSE, degB = TOP_FALSE;
    TOPLogical		fmtB = TOP_FALSE;
    char		*cmd, c, *fmtStr = NULL;
    EEMComplex		c1, c2, c3, cr, cr2;
    char		buf[100];

    while ( aC > opts && aV[opts][0] == '-' ) {
	cmd = aV[opts++];
	if ( strcmp(cmd,"-fmt")==0 ) {
	    fmtB = TOP_TRUE;
	    if ( aV[opts][0] == '%' || aV[opts][0] == '\0' ) {
		fmtStr = aV[opts];
		++opts;
	    }
	} else if ( strcmp(cmd,"-mp")== 0 ) {
	    mpB = TOP_TRUE;
	} else if ( strcmp(cmd,"-mpr")== 0 ) {
	    mpB = TOP_TRUE; degB = TOP_FALSE;
	} else if ( strcmp(cmd,"-mpd")== 0 ) {
	    mpB = TOP_TRUE; degB = TOP_TRUE;
	} else if ( strcmp(cmd,"-ri")== 0 ) {
	    mpB = TOP_FALSE;
	} else {
	    Tcl_AppendResult(ip,aV[0],": bad option ``",cmd,"''",NULL);
	    return TCL_ERROR;
	}
    }

    if ( aC < 2+opts ) {
	Tcl_AppendResult(ip,"usage: ",aV[0],
	  " ?-fmt ?format?? ?-ri? ?-mp? ?-mpd?",
	  " idendity|add|sub|mul|div op1 ?opts?",NULL);
	return TCL_ERROR;
    }
    cmd = aV[opts];	c = *cmd; len = strlen(cmd);
    if ( eemCplxStrPrsTcl( ip, aV[opts+1], &cr) != TCL_OK ) {
	Tcl_AppendResult(ip,"\n",aV[0]," ",cmd,": bad op1",NULL);
	return TCL_ERROR;
    }
    i = opts+2;
    /*IF*/ if ( c=='i' && strncmp(cmd,"identity",len)==0 ) {
	;
    } else if ( c=='a' && strncmp(cmd,"add",len)==0 ) {
	for ( ; i<aC; i++) {
    	    if ( eemCplxStrPrsTcl( ip, aV[i], &c2) != TCL_OK ) goto bad_i;
	    eemCplxAdd2( &cr, &c2);
	}
    } else if ( c=='m' && strncmp(cmd,"multiply",len)==0 ) {
	for ( ; i<aC; i++) {
    	    if ( eemCplxStrPrsTcl( ip, aV[i], &c2) != TCL_OK ) goto bad_i;
	    eemCplxMul2( &cr, &c2);
	}
    } else if ( c=='s' && strncmp(cmd,"subtract",len)==0 ) {
	if ( i+1 != aC ) goto bad_num;
    	if ( eemCplxStrPrsTcl( ip, aV[i], &c2) != TCL_OK ) goto bad_i;
	eemCplxSub2( &cr, &c2);
    } else if ( c=='d' && strncmp(cmd,"divide",len)==0 ) {
	if ( i+1 != aC ) goto bad_num;
    	if ( eemCplxStrPrsTcl( ip, aV[i], &c2) != TCL_OK ) goto bad_i;
	eemCplxDiv2( &cr, &c2);
    } else if ( c=='s' && (strcmp(cmd,"sqr")==0 || strcmp(cmd,"square")==0) ) {
	if ( i != aC ) goto bad_num;
	eemCplxSqr1( &cr);
    } else if ( c=='s' 
      && (strcmp(cmd,"sqrt")==0||strcmp(cmd,"squareroot")==0) ) {
	if ( i != aC ) goto bad_num;
	eemCplxSqrt1( &cr);
    } else if ( (c=='q' && strncmp(cmd,"quadroots",len)==0)
             || (c=='q' && strncmp(cmd,"quadraticroots",len)==0)
             || (c=='p' && strncmp(cmd,"poly2roots",len)==0) ) {
	if ( i+2 != aC ) {
	    Tcl_AppendResult(ip,"usage: ",aV[0]," ",cmd," a b c",NULL);
	    return TCL_ERROR;
	}
	c1 = cr;
    	if ( eemCplxStrPrsTcl( ip, aV[i], &c2) != TCL_OK ) goto bad_i;
	++i;
    	if ( eemCplxStrPrsTcl( ip, aV[i], &c3) != TCL_OK ) goto bad_i;
	eemCplxQuadraticRoots( &c1, &c2, &c3, &cr, &cr2);
        eemCplxFmtTcl( ip, &cr, TOP_TRUE, fmtStr);
        eemCplxFmtTcl( ip, &cr2, TOP_TRUE, fmtStr);
	return TCL_OK;
    } else {
	Tcl_AppendResult(ip,aV[0],": unknown opcode: ``",cmd,"''",NULL);
	return TCL_ERROR;
    }
    if ( mpB ) {
	eemCplxMagPhase( &cr, &cr);
        if ( degB ) {
	    cr.i *= 180.0 / M_PI;
	}
    }
    eemCplxFmtTcl( ip, &cr, TOP_FALSE, fmtStr);
    if ( fmtB ) {
	if ( mpB ) {
	    Tcl_AppendResult(ip,degB?"deg":"rad",NULL);
	}
    }
    return TCL_OK;

bad_i:
    sprintf( buf, "%d", i-opts-1);
    if ( i >= aC ) {
        Tcl_AppendResult(ip,"\n",aV[0]," ",cmd,": missing operand #",buf,NULL);
    } else {
        Tcl_AppendResult(ip,"\n",aV[0]," ",cmd,": bad operand #",buf,
          " ``",aV[i],"''",NULL);
    }
    return TCL_ERROR;

bad_num:
    sprintf( buf, "%d", aC-opts-1);
    Tcl_AppendResult(ip,"\n",aV[0]," ",cmd,": wrong num operands ",buf,NULL);
    return TCL_ERROR;
}

int
eemCplxRegisterCmds( Tcl_Interp *ip) {
    eemDfltRegisterCmds( ip);
    Tcl_CreateCommand( ip, "cplx", _eemCplxCmds, (ClientData)NULL, NULL);
    return TCL_OK;
}
