#if !defined(_EEMPFE_H_)
#define _EEMPFE_H_ 1
/* 
    RCS: $Header: /vol/cooley/cooley1/kennard/src/eem/RCS/eemPfe.h,v 1.14 93/08/27 14:29:20 kennard Exp $
	eemPfe.h :: EEM Library

	Partial Fraction Expansion functions.

    Author: Mei-Tjng Huang & Kennard White
    Copyright (c) 1990-93 The Regents of the University of California.
    All rights reserved.  See the file "./copyright.h" for full copyright
    notice, limitation of liability, and disclaimer of warranty provisions.
#include "copyright.h"				COPYRIGHTENDKEY
*/

#if !defined(_TOPVEC_H_) && __TOP_AUTOINCLUDE__
#include "topVec.h"
#endif
#if !defined(_EEMCPLX_H_) && __TOP_AUTOINCLUDE__
#include "eemCplx.h"
#endif
#if !defined(_EEMSILAR_H_) && __TOP_AUTOINCLUDE__
#include "eemSilar.h"
#endif
#if !defined(_EEMPOLY_H_) && __TOP_AUTOINCLUDE__
#include "eemPoly.h"
#endif

/**
    Each PFE represents a term of the form:
	gain * z^numerpow * (z - pole)^power
    where z is the free variable, and power is almost always negative.
    Note that most pfe algorithms assume that numerpow is 0.
**/

typedef struct _EEMPfe EEMPfe;
struct _EEMPfe {
    EEMComplex		gain;
    EEMComplex          pole;
    int                 power;		/* same as EEMSilar (negative) */
    int			numerpow;	/* normally 0 */
};

typedef struct _EEMPfeGrp EEMPfeGrp;
struct _EEMPfeGrp {
    int			shift;		/* # zeros at or origin */
    EEMPolyCplx         poly;		/* whole-part */
    TOPVector		terms;		/* of EEMPfe */
};


#define EEM_PFE_Num(pGrp)	TOPVEC_Num(&(pGrp)->terms)
#define EEM_PFE_Get(pGrp,i)	TOPVEC_Get(&(pGrp)->terms,(i),EEMPfe)

#define EEM_PFE_LOOP_BEGIN(pGrp,pPfe) \
    TOPVEC_LOOP_FORW_BEGIN(&(pGrp)->terms,EEMPfe,pPfe)

#define EEM_PFE_LOOP_END() \
    TOPVEC_LOOP_FORW_END()

#define EEM_PFE_IDX_LOOP_FORW_BEGIN(pGrp,idx,pPfe) \
    TOPVEC_IDX_LOOP_FORW_BEGIN(&(pGrp)->terms,idx,EEMPfe,pPfe)

#define EEM_PFE_IDX_LOOP_FORW_END() \
    TOPVEC_IDX_LOOP_FORW_END()
	
#define EEM_PFE_IDX_LOOP_BACK_BEGIN(pGrp,idx,pPfe) \
    TOPVEC_IDX_LOOP_BACK_BEGIN(&(pGrp)->terms,idx,EEMPfe,pPfe)

#define EEM_PFE_IDX_LOOP_BACK_END() \
    TOPVEC_IDX_LOOP_BACK_END()

#if __TOP_DEBUG__
#define EEM_PFE_Print(pkg,pPfe,title) {		\
    if ( topMsgPkgDbgOnB(pkg) ) {		\
	eemPfePrint(pPfe,title);		\
    }						\
}

#define EEM_PFEGRP_Print(pkg,pPfeGrp,title) {	\
    if ( topMsgPkgDbgOnB(pkg) ) {		\
	eemPfeGrpPrint(pPfeGrp,title);		\
    }						\
}

#else
#define EEM_PFE_Print(pkg,pPfe,title) /* */
#define EEM_PFEGRP_Print(pkg,pPfeGrp,title) /* */
#endif

/*
 * Specialized tokens
 */
#define EEM_PFE_ExtractOrigin	TOP_BIT(0)

/*
 * pfe.c :: EEMPfe & EEMPfeGrp core functions
 */
extern void             eemPfeGrpConstructor( EEMPfeGrp *pGrp);
extern void             eemPfeGrpDestructor( EEMPfeGrp *pGrp);
extern void             eemPfeGrpSetNum(  EEMPfeGrp *pGrp, int n);
extern EEMPfe*		eemPfeGrpAppend( EEMPfeGrp *pGrp, EEMComplex *pGain,
			  EEMComplex *pPole, int order);

extern void		eemPfePrint(EEMPfe *pPfe, char *title);
extern void		eemPfeGrpPrint( EEMPfeGrp *pGrp, char *title);
extern void 		eemPfeDenOneToPolyCplx(EEMPolyCplx *pD, EEMPfe *pS);
extern void 		eemPfeDenToPolyCplx(EEMPolyCplx *pD, EEMPfe *pS);
extern void 		eemPfeGrpToTime(EEMComplex *ptime,
			  double t0, double dt, 
			  int NumStep, EEMPfeGrp *pPG);
extern void		eemPfeToTimeS(EEMPfe *pPfe, double t0,
			  EEMComplex *val);
extern void		eemPfeToTimeZ(EEMPfe *pPfe, int n,
			  EEMComplex *val);
extern void		eemPfeGrpToTimeS(EEMPfeGrp *pPfe, double t0,
			  EEMComplex *val);
extern void		eemPfeGrpToTimeZ(EEMPfeGrp *pPfe, int n,
			  EEMComplex *val);
extern void 		eemConjPfeToTime(EEMPfe *pPfe, double t0,
			  EEMComplex *val);

/*
 * pfesilar.c :: Conversions between PfeGrp and SilarGrp
 */
extern void     eemSilarGrpToPfeGrp(EEMSilarGrp *pSG, EEMPfeGrp *pPG,
		  TOPMask flags);
extern void     eemPfeToSilar(EEMPfe *pS, EEMSilar *pD);
extern void     eemPfeGrpToSilarGrp(EEMPfeGrp *pS, EEMSilarGrp *pD,
		  int numzeros);

#endif /* _EEMPFE_H_ */
