#include "topFixup.h"
#include <stdio.h>
#include <math.h>

#include "topStd.h"
#include "topMem.h"
#include "topVec.h"

#include "eemDflt.h"
#include "eemCplx.h"
#include "eemSilar.h"
#include "eemPoly.h"
#include "eemLoci.h"

char _EemRootSubPackage[] = "eem.root";
#define SPKG _EemRootSubPackage

#define DELTAH 		1e-7
#define EPSILON		1e-2
#define DIFF_EPSILON	1e-3
#define Q_EPSILON	1e-10
#define Q_NON_ZERO	1e-5
#define MAXPOS		1e300 
#define MAXNEG         -1e300
#define MAX(x, y)		 	(x > y) ? x : y
#define MIN(x, y)		 	(x > y) ? y : x 

TOPLogical
eemBSearchK(EEMLocus *pLocus, double k, EEMComplex *pOut) {
    double midK, midK1, ratio;
    int lo, mid, hi=EEM_LOCUS_Num(pLocus)-1;
    EEMComplex temp;
    EEMPolyCplxRat *pRat = &pLocus->MasterGrp->Rat;
    lo = 0;

    while (lo <= hi ) {
	mid = (lo +  hi)/2;
	midK = eemEvalK(pRat,&EEM_LOCUS_Get(pLocus,mid));
	if (k < midK) {
	    midK1 = eemEvalK(pRat, &EEM_LOCUS_Get(pLocus,mid-1)); 
	    if (k > midK1) {
		ratio = ( k - midK1)/(midK - midK1);
		eemCplxSub3(&EEM_LOCUS_Get(pLocus, mid),
		 	&EEM_LOCUS_Get(pLocus, mid-1), &temp);
		pOut->r = EEM_LOCUS_Get(pLocus, mid).r + ratio*temp.r;
		pOut->i = EEM_LOCUS_Get(pLocus, mid).i + ratio*temp.i;
		return TOP_TRUE;
            } else 
		hi = mid-1;		  
        } else if (k > midK) {
	    midK1 = eemEvalK(pRat, &EEM_LOCUS_Get(pLocus,mid+1));
	    if (k < midK1) {
		ratio = (k - midK)/ (midK1 - midK);
		eemCplxSub3(&EEM_LOCUS_Get(pLocus, mid+1),
		 	&EEM_LOCUS_Get(pLocus, mid), &temp);
		pOut->r = EEM_LOCUS_Get(pLocus, mid).r + ratio*temp.r;
		pOut->i = EEM_LOCUS_Get(pLocus, mid).i + ratio*temp.i;
		return TOP_TRUE;
            } else 
	        lo = mid +1;
	} else { 
	    *pOut = EEM_LOCUS_Get(pLocus, mid);
	    return TOP_TRUE;
	}
    }
    return TOP_FALSE; 
}

void
eemLocusPtAtK (EEMLociGrp *pLGrp, double k, EEMSilarGrp *pSGrp) {
    int i, m, n = EEM_LOCI_Num(pLGrp);
    EEMLocus *pLocus;
    EEMSilar *pSilar;
 
    eemSilarGrpSetNum(pSGrp, n);
    for (i=0, m=0; i<n; i++){
	pLocus = &EEM_LOCI_Get(pLGrp, m);
	if (k>0.) {
	    pLocus->CurLocus = &pLocus->LocusPos;
	    if (eemBSearchK(pLocus, k, &EEM_SILAR_Get(pSGrp, m).loc)){
		EEM_SILAR_Get(pSGrp, m).power=1;
		m++;
	    }
        } else {
	    pLocus->CurLocus = &pLocus->LocusNeg;
	    if( eemBSearchK(pLocus, k, &EEM_SILAR_Get(pSGrp, m).loc)){
		EEM_SILAR_Get(pSGrp, m).power=1;
		m++;
	    }
	}
    } 
    TOP_PDBG((SPKG, "\nIn eemLocusPtAtK: k = %g", k));
    EEM_SILARGRP_Print(SPKG, pSGrp, "\nIn eemLocusPtAtK: list of locus points at same k"); 
    
}


/**
Suggested use of eemRootLociMaster and eemZoomIn when used to plot
root loci:

When the program is first started, eemRootLociMaster is called.  
Drag should be set to false so that the box size
would be determined based on the location of the singularities.
(the box is centered at the origin, and its length and width are 
approximatley four times the largest distance between a singularity
and the origin).

If Drag is TOP_TRUE, pBox is not reset.
If Drag is TOP_FALSE, pBox is reset.
**/

void
eemRootLociMaster( EEMLociGrp *pLGrp, EEMBox *pB, EEMSilarGrp *pSGrp, double k0, TOPLogical Drag)  { 
    int n, z;
    EEMPolyCplx temp, abp, ADoubleP, BDoubleP;
    EEMPolyCplxRat *pRat;
    EEMLocus *pLocus, Locus;
    EEMSilarGrp *pSing, pSG;
    EEMArray *pKlist;
    EEMBox BSearch, *pBS,  *pBox, Box = *pB; 
    pBox = &Box;
    pBS = &BSearch; 

    if (EEM_SILAR_Num(pSGrp) == 0)
	return;
    pLGrp->SG = pSGrp;
    pLocus = &Locus;
    pKlist = &pLGrp->Klist;
    pRat = &pLGrp->Rat;
    pSing = &pLGrp->Sing;
    pBox->delta.x = (pBox->hi.x - pBox->lo.x) * .005;
    pBox->delta.y = (pBox->hi.y - pBox->lo.y) * .005;

    eemPolyCplxConstructor(&temp);
    eemLocusConstructor(&Locus, pLGrp);

    eemSilarGrpSetNum(pSing, 0);
    eemSilarToPolyCplxRat(pRat, pSGrp);
    eemLociGrpSetNum(pLGrp, 0);

    eemPolyCplxDiff2(&pRat->Num, &pLGrp->ADiff);
    eemPolyCplxDiff2(&pRat->Den, &pLGrp->BDiff);

    eemPolyCplxConstructor( &abp);
    eemPolyCplxMul3(&pLGrp->ADiff, &pRat->Den, &pLGrp->KDiff);
    eemPolyCplxMul3(&pRat->Num, &pLGrp->BDiff, &abp);
    eemPolyCplxSub2(&pLGrp->KDiff, &abp);
    eemPolyCplxDestructor( &abp);
    
    eemPolyCplxRatPrint(pRat, "\npRat");
    EEM_POLYCPLXRAT_Print(SPKG, pRat, "\npRat");
    EEM_POLYCPLX_Print(SPKG, &pLGrp->ADiff, "\nADiff");
    EEM_POLYCPLX_Print(SPKG, &pLGrp->BDiff, "\nBDiff");
    EEM_POLYCPLX_Print(SPKG, &pLGrp->KDiff, "\nKDiff");

    eemAppendPolyCplxToSilarGrp(pSing, &pLGrp->KDiff);
    EEM_SILARGRP_Print (SPKG, pSing, "\n\nList of Singular Points:");

    z = EEM_SILAR_Num(pSing);
    eemArraySetNum (pKlist, z);
    eemStoreK (pSing, pRat, pBox); 
    /**
      No box specified.  Box dimensions calculated.
    **/
    if (Drag == TOP_FALSE) { 
	eemResizeBox(pSGrp, pBox);
	pBS = pBox;
    } else { 
	BSearch.lo.x = 1.5 * pBox->lo.x - .5 * pBox->hi.x;
	BSearch.lo.y = 1.5 * pBox->lo.y - .5 * pBox->hi.y;
	BSearch.hi.x = 1.5 * pBox->hi.x - .5 * pBox->lo.x;
	BSearch.hi.y = 1.5 * pBox->hi.y - .5 * pBox->lo.y;

	pBS->delta.x = (pBS->hi.x - pBS->lo.x) * .005;
	pBS->delta.y = (pBS->hi.y - pBS->lo.y) * .005;
    }
    /*
     * If a silar has conjugate, then the silargroup
     * is expanded
     */
    eemSilarGrpFlattenConj(pSGrp);

    eemDeleteMultSing(pSing);

    n = EEM_SILAR_Num(pSing);
    eemArraySetNum(pKlist,n);
    eemSilarToArrayCompact (pSing, pKlist);

    EEM_ARRAY_Print(SPKG, pKlist, "This is pKlist"); 
    EEM_SILARGRP_Print(SPKG, pSing, "Singular list");
    EEM_LOCIGRP_Print(SPKG, pLGrp, "LociGrp");

    eemDrawMaster(pLGrp, pSGrp, pBS, pBox);

    eemPolyCplxDestructor(&temp);
    eemLocusDestructor(&Locus);
}

void
eemResizeBox(EEMSilarGrp *pSGrp, EEMBox *pBox){
    double c, d=0.0;
    int i;
    for (i=0; i<EEM_SILAR_Num(pSGrp); i++) {
	c = EEM_CPLX_MAGSQ(&EEM_SILAR_Get(pSGrp, i).loc); 
	if (d < c) 		d = c;
    }
    d = sqrt(d);
    pBox->lo.x = -2 * d;
    pBox->lo.y = -2 * d;
    pBox->hi.x = 2 * d;
    pBox->hi.x = 2 * d;
}

/** 
Add K  and  Q to pSing
K in this case is pSilar->extra.val
Q in this case is pSilar->power
**/
void
eemStoreK(EEMSilarGrp *pSing, EEMPolyCplxRat *pRat, EEMBox *pBox) {
    int i;
    for (i=0; i<EEM_SILAR_Num(pSing); i++) {
	EEM_SILAR_Get(pSing,i).extra.val=
	  eemEvalK(pRat, &EEM_SILAR_Get(pSing, i).loc);
        eemFindQ(pRat, pBox, &EEM_SILAR_Get(pSing,i).loc,
	  &EEM_SILAR_Get(pSing,i).power); 
    }
}

void 
eemSilarToArrayCompact(EEMSilarGrp *pSing, EEMArray *pKlist){
    int i, j, m, n = EEM_SILAR_Num(pSing);
    double c, d;

    if (n == 0)
	return;
    for (i=0, m=0; i<n; i++, m++) {
	c=EEM_SILAR_Get(pSing,i).extra.val;
	for (j=0; j==0 || j<m; j++){
	    d = EEM_ARRAY_Get(pKlist, j);
	    if ( TOP_ABS(c - d) < .001*c) {
		m--;
	        break;
	    } else if (j == 0 || j == m-1)  
	        EEM_ARRAY_Get(pKlist, m) = EEM_SILAR_Get(pSing,i).extra.val;
	}
    }
    eemArraySetNum(pKlist,m);
}


void
eemDrawMaster(EEMLociGrp *pLGrp, EEMSilarGrp *pSGrp, EEMBox *pBS, EEMBox *pBox){
    int i, m, n, w;
    double c, d, NextKNeg, Dtemp, NextK=0;
    EEMLocus *pLocus;
    EEMComplex Ctemp, MagPhase, posStart, negStart;

    for ( i=0; i<EEM_SILAR_Num(pSGrp); i++){
	if (EEM_SILAR_Get(pSGrp, i).power < 0) {	
	    n = -EEM_SILAR_Get(pSGrp, i).power;
	    w = EEM_LOCI_Num(pLGrp);
            eemLociGrpSetNum(pLGrp, w + n);	
            /*
	     * In dealing with repeated poles, the algorithm moves
	     * the starting point by a little bit, and changes the
	     * corresponding starting value of k
             *   K = epsilon^n/p 
	     * p*exp(j*phi) = - A(s) (s-p_k)^n/B(s) evaluated at
	     *   p_k
	     * "New" poles are at S_old + epsilon
	     * 	  * exp[j(phi + 2*pi*m)/n]
	     */
	    for (m=0; m<n; m++) {
		pLocus = &EEM_LOCI_Get( pLGrp, w + m);  
                if (n != 1) {
		    eemEvalSilarA0(pSGrp, i, &Ctemp);
		    eemCplxMagPhase( &MagPhase, &Ctemp); 
                    MagPhase.i += M_PI; 
                    Dtemp = (MagPhase.i+2.0*M_PI*(double)m)/(double)n; 
		    c = EPSILON * cos (Dtemp);
		    d = EPSILON * sin (Dtemp);
		    posStart.r=EEM_SILAR_Get(pSGrp, i).loc.r+c; 
		    posStart.i=EEM_SILAR_Get(pSGrp, i).loc.i+d; 
		    negStart.r=EEM_SILAR_Get(pSGrp, i).loc.r-c; 
		    negStart.i=EEM_SILAR_Get(pSGrp, i).loc.i-d; 
		    EEM_CPLX_Print(SPKG, &posStart, "\nPositive Start");
		    EEM_CPLX_Print(SPKG, &negStart, "\nNegative Start");

		    NextK =  pow(EPSILON, (double)n)/MagPhase.r; 
                    NextKNeg = -NextK;
	        } else {
		    posStart = negStart = EEM_SILAR_Get(pSGrp, i).loc;
		    NextK = NextKNeg = 0.0;
		}
		/*
		 * For positive K
		 */
		pLocus->CurLocus = &pLocus->LocusPos;
		pLGrp->DeltaH = DELTAH;
		pLGrp->DiffEpsilon = DIFF_EPSILON;

		EEM_CPLX_Print(SPKG,&EEM_SILAR_Get(pSGrp,i).loc,"\nstore");
		EEM_LOCUS_Append(pLocus) = EEM_SILAR_Get(pSGrp, i).loc;
		eemDrawLocus (pLocus, &posStart, pBS, pBox, &NextK, 0.0, MAXPOS);
		/*
		 * For negative K
		 */
		pLocus->CurLocus = &pLocus->LocusNeg;
		pLGrp->DeltaH = DELTAH;
		pLGrp->DiffEpsilon = DIFF_EPSILON;
		/* 
		 * storing multiple poles as they are skipped over
		 * in eemDrawLocus
		 */
		EEM_CPLX_Print(SPKG,&EEM_SILAR_Get(pSGrp,i).loc,"\nstore");
		EEM_LOCUS_Append(pLocus) = EEM_SILAR_Get(pSGrp, i).loc;
		eemDrawLocus (pLocus, &negStart, pBS, pBox, &NextKNeg, 0.0, MAXNEG);
	    }
	}
    }
}

TOPLogical
eemInsideBox(EEMLocusPt *pLocusPt, EEMBox *pB){
    if (pLocusPt->r >= pB->lo.x && pLocusPt->r <= pB->hi.x &&
      pLocusPt->i >= pB->lo.y && pLocusPt->i <= pB->hi.y) 
        return TOP_TRUE;
    else 
	return TOP_FALSE;
}

/**
q is the q_th derivative such that the the differentiated polynomial 
is not equal to 0 evaluated at pLocus.

If val < Q_EPSILON, val is assumed to be zero, and differentiation
continues.  If val > Q_NON_ZERO, val is assumed to be non-zero,
and q has been found.  Otherwise, pLocusPt is not close enough to
the exact singular point, and Newton-Raphson is applied to obtain a 
more accurate singular point.
**/
void 
eemFindQ(EEMPolyCplxRat *pRat, EEMBox *pDB, EEMLocusPt *pLocusPt, int *pq) {
    double n,d,i,j,val= 0.0;
    EEMComplex Ctemp, a, b, s, *ps;
    EEMPolyCplxRat PCRat, PCRat_Diff, PCRat_Temp;
    ps = &s;
    eemPolyCplxRatConstructor(&PCRat);
    eemPolyCplxRatConstructor(&PCRat_Diff);
    eemPolyCplxRatConstructor(&PCRat_Temp);
    eemPolyCplxRatCopy (&PCRat, pRat);
    EEM_POLYCPLXRAT_Print( SPKG, pRat, "\nIn FindQ:pRat");
    *pq = 0;
    n = EEM_POLYCPLX_Num(&pRat->Num); 
    d = EEM_POLYCPLX_Num(&pRat->Den);
    for (i=0; i< n+d-2; i++) { 

	eemPolyCplxRatDiff1( &PCRat);
	eemEvalPolyCplxRat( &PCRat, pLocusPt, &Ctemp);
	val = EEM_CPLX_MAGSQ( &Ctemp);

	(*pq)++;
	#if __TOP_DEBUG__
	printf("\nval = %g, q = %d", val, *pq);
	#endif
	if (val < Q_EPSILON){
	    continue;
        } else if (val > Q_NON_ZERO) {
	    EEM_CPLX_Print(SPKG, pLocusPt, "\n\tIn FindQ: LocusPt in Q>NonZero");
	    #if __TOP_DEBUG__
	    printf("\nval = %g, q = %d", val, *pq);
	    #endif
	    return;
        } else {
	    #if __TOP_DEBUG__
	    printf("\nUndetermined val: %g \t q:%d", &val, *pq);
	    #endif
	    eemPolyCplxRatCopy(&PCRat_Temp, &PCRat);
	    s = *pLocusPt;  
	    for (j=0; j<5; j++){
		eemPolyCplxRatDiff2(&PCRat_Temp, &PCRat_Diff);
		eemEvalPolyCplxRat(&PCRat_Temp, ps, &a); 
		eemEvalPolyCplxRat(&PCRat_Diff, ps, &b); 
		eemCplxDiv2(&a, &b);
		ps->r = ps->r - a.r;
		ps->i = ps->i - a.i;
		eemPolyCplxRatCopy(&PCRat_Temp, &PCRat_Diff);

		eemEvalPolyCplxRat( &PCRat, ps, &Ctemp);
		val = EEM_CPLX_MAGSQ( &Ctemp);
		if (val<Q_EPSILON || val>Q_NON_ZERO){ 
		    #if __TOP_DEBUG__
		    printf("In FindQ: undetermined val becomes:%g", val);
		    #endif
		    EEM_CPLX_Print(SPKG, ps, "\n\tIn FindQ: ps in Q>NonZero");
		    break;
	        }
		if (val > Q_NON_ZERO) 
		    /*
		     * Note: Newton Iteration might have caused
		     * s to move off locus
		     */
		    if (eemInsideBox (ps, pDB)) 
			*pLocusPt = s;
			return;
	    }
	}
    }
    eemPolyCplxRatDestructor(&PCRat);
    eemPolyCplxRatDestructor(&PCRat_Diff);
    eemPolyCplxRatDestructor(&PCRat_Temp);
}


/**
Newton Interation:
s_k+1 = s_k - [B(s_k) + K*A(s_k)]/[B'(s_k)+K*A'(s_k)]

t is the number of times Newton's interation is to be run
**/


void
eemNewtonIteration(EEMLocusPt *pLocusPt, EEMLociGrp *pLGrp, double k){
    int i;
    EEMPolyCplxRat *pRat=&pLGrp->Rat;
    EEMComplex CtempA, CtempB, CtempAPrime, CtempBPrime;
   
    EEM_POLYCPLXRAT_Print(SPKG, pRat, "\nIn NewtonInter: pRat");
        eemEvalPolyCplx (&pRat->Num, pLocusPt, &CtempA);
        eemEvalPolyCplx (&pRat->Den, pLocusPt, &CtempB);

	CtempB.r += CtempA.r * k;
	CtempB.i += CtempA.i * k;
    
	eemEvalPolyCplx (&pLGrp->ADiff, pLocusPt, &CtempAPrime);
	eemEvalPolyCplx (&pLGrp->BDiff, pLocusPt, &CtempBPrime);

	CtempBPrime.r += CtempAPrime.r * k;
	CtempBPrime.i += CtempAPrime.i * k;

	eemCplxDiv2( &CtempB, &CtempBPrime);

	pLocusPt->r = pLocusPt->r - CtempB.r;
        pLocusPt->i = pLocusPt->i - CtempB.i;
}



double
eemEvalK(EEMPolyCplxRat *pRat, EEMLocusPt *pLocusPt) {
    EEMLocusPt A, B;

    eemEvalPolyCplx( &pRat->Num, pLocusPt, &A);
    eemEvalPolyCplx( &pRat->Den, pLocusPt, &B);

    return (-B.r/A.r);
}


/**
Suggested use:

eemRootLociRange should called whenever a range of k is specified.
When zooming out, if a range of k is specified, then eemRootLociRange is 
called.  Otherwise, eemRootLociMaster is called.`

pLGrp is the master locus.
pLGP is the partial locus.
**/

void
eemRootLociKRange(EEMLociGrp *pLGrp, EEMLociGrp *pLGP, EEMBox *pB, double KStart, double KEnd, TOPMask InfFlag) {
    int i, j, n, p, m;
    EEMLocus *pLP, *pLocus;
    EEMPolyCplxRat *pRat=&pLGrp->Rat;
    EEMLocusPt LocusPt;
    double k, KS, KE, KD = KEnd-KStart;
    TOPLogical Sign;

    n = EEM_LOCI_Num(pLGrp);
    eemLociGrpSetNum(pLGP, n); 

    if (TOP_IsOn(InfFlag, EEMLocusF_PosInf))
	KEnd = MAXPOS;
    else if (TOP_IsOn(InfFlag, EEMLocusF_NegInf))
	KEnd = MAXNEG; 

    for (i=0; i<EEM_LOCI_Num(pLGrp); i++){
	#if __TOP_DEBUG__
	printf("\nthis is locus #%d",i);
        printf("\nFor positive k");
	#endif
	pLocus = &EEM_LOCI_Get(pLGrp, i);
	pLP = &EEM_LOCI_Get(pLGP, i);
        pLocus->CurLocus = &pLocus->LocusPos;
        pLP->CurLocus = &pLP->LocusPos;
        eemLocusSetNum(pLP, 0);  
	/**
	  Case 1:
	     Ex:  3 < k < 10
		 - infinity < k < 3
		  3 < k < + inifinity
	**/
	if (KStart > 0.0) {
	    Sign = TOP_TRUE;
	    pLocus->CurLocus = &pLocus->LocusPos;  
	    pLP->CurLocus = &pLP->LocusPos;  
	    if (KEnd > KStart) { 
		KE = KEnd;
		KS = KStart;
	    } else {
		KE = KStart;
		KS = 0.0;
	    }
	    m = EEM_LOCUS_Num(pLocus);
	    for (p=0; p<m ;p++){
		k = eemEvalK(pRat, &EEM_LOCUS_Get(pLocus, p));
		if (k > KS)
		    break;
	    }
	    for ( ; p<m ; p++){
		k = eemEvalK(pRat, &EEM_LOCUS_Get(pLocus, p));
		if (k <= KE) {
		    EEM_LOCUS_Append(pLP) = EEM_LOCUS_Get(pLocus,p);
		    EEM_CPLX_Print(SPKG, &EEM_LOCUS_Get(pLocus,p), "\nAppend");
		    #if __TOP_DEBUG__
		    printf("k=%g",k);
		    #endif
		} else {
		    break;
		    p++;
		}	
	    }
	    LocusPt = EEM_LOCUS_Get(pLocus, p-1);	
	    if (k < KE)
		eemDrawLocus(pLocus, &LocusPt, pB, pB, &k, KS, KE);
	    /*
	     * positive k for range of k from minus infinity to KStart
	     */
	    if (KEnd < 0.0) {
		printf("\nFor Negative K:");
	        Sign = TOP_FALSE;
		pLocus->CurLocus = &pLocus->LocusNeg;	
		pLP->CurLocus = &pLP->LocusNeg;
		eemLocusSetNum(pLP, 0);
	
		for (p=0; p<EEM_LOCUS_Num(pLocus); p++) {
		    EEM_LOCUS_Append(pLP) = EEM_LOCUS_Get(pLocus,p);
		    EEM_CPLX_Print(SPKG, &EEM_LOCUS_Get(pLocus, p), "\nAppend:");
		    #if __TOP_DEBUG__
		    k = eemEvalK(pRat, &EEM_LOCUS_Get(pLocus,p));
		    printf("\n k=%g",k);
		    #endif
		}
		KS = 0.0;
		KE = MAXNEG;
		LocusPt = EEM_LOCUS_Get(pLocus,p-1);
	        k = eemEvalK(pRat, &EEM_LOCUS_Get(pLocus,p-1));
		eemDrawLocus(pLocus, &LocusPt, pB, pB, &k, KS, KE);
	    }
	} else {
	    /*
	     * For LocusPos:
	     */
	    if (KEnd > 0.0){
		Sign = TOP_TRUE;
		pLocus->CurLocus = &pLocus->LocusPos;
		pLP->CurLocus = &pLP->LocusPos;
		m = EEM_LOCUS_Num(pLocus);
		KS = 0.0;
		KE = KEnd;

		for (p=0 ; p<m ;p++){
		    k = eemEvalK(pRat, &EEM_LOCUS_Get(pLocus, p));
		    if (k < KE){ 
			EEM_LOCUS_Append(pLP)=EEM_LOCUS_Get(pLocus,p);
			EEM_CPLX_Print(SPKG, &EEM_LOCUS_Get(pLocus,p), "\nAppend");
			#if __TOP_DEBUG__
			printf("k=%g",k);
		   	#endif
		    }
		    else {
			p++;
			break;
		    }
		} 
		LocusPt = EEM_LOCUS_Get(pLocus, p-1);	
		if (k < KE)
		    eemDrawLocus(pLocus, &LocusPt, pB, pB, &k, KS, KE);
	    }	
	    /*
	     * For LocusNeg:
	     */
	    printf("\nFor Negative k");
	    Sign = TOP_FALSE;
	    pLocus->CurLocus = &pLocus->LocusNeg;  
	    pLP->CurLocus = &pLP->LocusNeg;  
	    eemLocusSetNum(pLP, 0);
	    if (KEnd > 0.0) {
		KE = KStart; 
		KS = 0.0;
	    }
	    else {
		KE = KEnd;
		KS = KStart;
	    }

	    m = EEM_LOCUS_Num(pLocus);
	    for (p=0; p<m ;p++){
		k = eemEvalK(pRat, &EEM_LOCUS_Get(pLocus, p));
		if (k < KS)
		    break;
	    }
	    for ( ; p<m ;p++){
		k = eemEvalK(pRat, &EEM_LOCUS_Get(pLocus, p));
		if (k > KE){
		    EEM_LOCUS_Append(pLP)=EEM_LOCUS_Get(pLocus,p);
		    EEM_CPLX_Print(SPKG, &EEM_LOCUS_Get(pLocus,p),"\nAppend");
		    #if __TOP_DEBUG__
		    printf("k=%g",k);
		    #endif
		} else {
		    p++;
		    break;
		}
	    }
	    if (k > KE){
		LocusPt = EEM_LOCUS_Get(pLocus, p-1);	
		eemDrawLocus(pLocus, &LocusPt, pB, pB, &k, KS, KE);
	    }
	    /*
	     * positive k for range of k from minus infinity to KStart
	     */
		

	}
    }
}


/**
pLGrp is the master locus
pLGP is the partial locus to be determined
**/
void
eemRootLociZoomInK( EEMLociGrp *pLGrp, EEMLociGrp *pLGP, EEMBox *pB, double KStart, double KEnd){
    int i,n;
    EEMLocus *pLocus, *pLP; 
    EEMBox BSearch, *pBS; 
    TOPLogical Sign;
    EEMLocusPt LocusPt, *pLocusPtNew;
    double KS, KE,tmp;
    n = EEM_LOCI_Num(pLGrp);
    eemLociGrpSetNum(pLGP, n);

    if (KStart > KEnd) {
       tmp = KStart;
       KStart = KEnd;
       KEnd = tmp;
    }
    pBS = &BSearch;
    /*
     * BSearch box is twice as big (perimeter wise ) as the 
     * original bounding box
     */
    BSearch.lo.x = 1.5 * pB->lo.x - .5 * pB->hi.x;
    BSearch.lo.y = 1.5 * pB->lo.y - .5 * pB->hi.y;
    BSearch.hi.x = 1.5 * pB->hi.x - .5 * pB->lo.x;
    BSearch.hi.y = 1.5 * pB->hi.y - .5 * pB->lo.y;

    pBS->delta.x = (pBS->hi.x - pBS->lo.x) * .005;
    pBS->delta.y = (pBS->hi.y - pBS->lo.y) * .005;

    pB->delta.x = (pB->hi.x - pB->lo.x) * .005;
    pB->delta.y = (pB->hi.y - pB->lo.y) * .005;

    pLocus = &EEM_LOCI_Get(pLGrp, i); 

    printf("\nBox pBS:\t lo:(%g,%g), \t hi:(%g,%g), \t delta:(%g,%g)\n",pBS->lo.x, pBS->lo.y, pBS->hi.x, pBS->hi.y, pBS->delta.x, pBS->delta.y);

    for (i=0; i<EEM_LOCI_Num(pLGrp); i++){
        pLocus = &EEM_LOCI_Get(pLGrp, i); 
        pLP = &EEM_LOCI_Get(pLGP, i); 
	pLP->PntGrp = pLGP;

	if (KEnd > 0.0) {
	    KE = KEnd;
	    KS = MAX(KStart, 0.0);
	    pLP->CurLocus = &pLP->LocusPos;  
	    pLocus->CurLocus = &pLocus->LocusPos;  
	    eemDrawLocusPartK(pLocus, pLP, pBS, pB, KS, KE);
	    EEM_LOCUS_Print(SPKG, pLP, "ZoomLocusPos");
        }
	if (KStart < 0.0) {
	    KE = KStart;
	    KS = MIN (0.0, KEnd);
	    pLP->CurLocus = &pLP->LocusNeg; 
	    pLocus->CurLocus = &pLocus->LocusNeg;
	    eemDrawLocusPartK(pLocus, pLP, pBS, pB, KS, KE);
	    EEM_LOCUS_Print(SPKG, pLP, "ZoomLocusNeg");
	}
    }
}

/**
This function searches for points inside pLocusDst that are inside the search box.  Upon finding a point inside the search box, it
calls eemDrawLocus to calculate successive points, which are 
stored in pLocus->LocusPosPart or pLocus->LocusNegPart, 
depending upon Sign
**/

void
eemDrawLocusPartK(EEMLocus *pLocus, EEMLocus *pLP, EEMBox *pBS, EEMBox *pB, double KStart, double KEnd) {
    int k, j = 0;
    double NextK, thisK, absNextK, ABS_KE;
    EEMLocusPt LocusPtNew, *pLocusPtNew;
    EEMSilarGrp *pSG = pLocus->PntGrp->SG;
    EEMPolyCplxRat *pRat = &pLocus->PntGrp->Rat;
    int	numLocusPts = EEM_LOCUS_Num(pLocus);
    pLocusPtNew = &LocusPtNew;
    
    ABS_KE = TOP_ABS(KEnd);
    *pLocusPtNew = EEM_LOCUS_Get(pLocus, 0);
    
    for (k=0; k<EEM_SILAR_Num(pSG); k++){
	if (eemCplxSameB(&EEM_SILAR_Get(pSG,k).loc, pLocusPtNew)) {
	    *pLocusPtNew = EEM_LOCUS_Get(pLocus, 1);
	    j = 1;
	    break;
	}
    }
    for ( ; j<numLocusPts; j++) {
	*pLocusPtNew = EEM_LOCUS_Get(pLocus, j);
	if (eemInsideBox(pLocusPtNew, pBS)){ 
	    NextK = eemEvalK(pRat, &EEM_LOCUS_Get(pLocus,j)); 
	    pLP->PntGrp->DeltaH = DELTAH;
	    if (TOP_ABS(NextK) > ABS_KE)
		return;	
	    eemDrawLocus(pLP, pLocusPtNew, pBS, pB, &NextK, KStart, KEnd);
	    absNextK = TOP_ABS(NextK);
	    if (absNextK > ABS_KE)
		return;
	    for ( ; j<numLocusPts; j++) {
		thisK = eemEvalK(pRat, &EEM_LOCUS_Get(pLocus,j));
		 if ( TOP_ABS(thisK) > absNextK )
		     break;
	    }
	    if (TOP_ABS(thisK) > ABS_KE)
		return;
	    for (; j<numLocusPts ; j++) {
		if (eemInsideBox(&EEM_LOCUS_Get(pLocus,j), pBS)) 
		    ;
	    NextK = eemEvalK (pRat, &EEM_LOCUS_Get(pLocus, j));
	    if (TOP_ABS(NextK) > ABS_KE)
		return;
	    }
	}
    }
}


/**
pLGrp is the master locus
pLGP is the partial locus to be determined
**/
void
eemRootLociZoomIn( EEMLociGrp *pLGrp, EEMLociGrp *pLGP, EEMBox *pB){
    int i,n;
    EEMLocus *pLocus, *pLP; 
    EEMBox BSearch, *pBS; 
    TOPLogical Sign;
    EEMLocusPt LocusPt, *pLocusPtNew;

    n = EEM_LOCI_Num(pLGrp);
    eemLociGrpSetNum(pLGP, n);

    pBS = &BSearch;
    /*
     * BSearch box is twice as big (perimeter wise ) as the 
     * original bounding box
     */
    BSearch.lo.x = 1.5 * pB->lo.x - .5 * pB->hi.x;
    BSearch.lo.y = 1.5 * pB->lo.y - .5 * pB->hi.y;
    BSearch.hi.x = 1.5 * pB->hi.x - .5 * pB->lo.x;
    BSearch.hi.y = 1.5 * pB->hi.y - .5 * pB->lo.y;

    pBS->delta.x = (pBS->hi.x - pBS->lo.x) * .005;
    pBS->delta.y = (pBS->hi.y - pBS->lo.y) * .005;
    pB->delta.x = (pB->hi.x - pB->lo.x) * .005;
    pB->delta.y = (pB->hi.y - pB->lo.y) * .005;

    pLocus = &EEM_LOCI_Get(pLGrp, i); 

    printf("\nBox pBS:\t lo:(%g,%g), \t hi:(%g,%g), \t delta:(%g,%g)\n",pBS->lo.x, pBS->lo.y, pBS->hi.x, pBS->hi.y, pBS->delta.x, pBS->delta.y);

    for (i=0; i<EEM_LOCI_Num(pLGrp); i++){
        pLocus = &EEM_LOCI_Get(pLGrp, i); 
        pLP = &EEM_LOCI_Get(pLGP, i); 
	pLP->PntGrp = pLGP;
        Sign = TOP_TRUE;
        pLP->CurLocus = &pLP->LocusPos;  
        pLocus->CurLocus = &pLocus->LocusPos;  
	eemDrawLocusPart(pLocus, pLP, pBS, pB, Sign);

	#ifdef notdef
	Sign = TOP_FALSE;
	pLP->CurLocus = &pLP->LocusNeg; 
	pLocus->CurLocus = &pLocus->LocusNeg;
	eemDrawLocusPart(pLocus, pLP, pBS, pB, Sign);
	#endif
    }
}


/**
This function searches for points inside pLocusDst that are inside the search box.  Upon finding a point inside the search box, it
calls eemDrawLocus to calculate successive points, which are 
stored in pLocus->LocusPosPart or pLocus->LocusNegPart, 
depending upon Sign
**/

void
eemDrawLocusPart(EEMLocus *pLocus, EEMLocus *pLP, EEMBox *pBS, EEMBox *pB, TOPLogical Sign) {
    int j,k;
    double NextK, thisK, absNextK;
    EEMLocusPt LocusPtNew, *pLocusPtNew;
    EEMSilarGrp *pSG = pLocus->PntGrp->SG;
    EEMPolyCplxRat *pRat = &pLocus->PntGrp->Rat;
    int	numLocusPts = EEM_LOCUS_Num(pLocus);
    pLocusPtNew = &LocusPtNew;

    for (j=0; j<numLocusPts; j++) {
	*pLocusPtNew = EEM_LOCUS_Get(pLocus, j);
	for (k=0; k<EEM_SILAR_Num(pSG); k++){
	    if (eemCplxSameB(&EEM_SILAR_Get(pSG,k).loc, pLocusPtNew)) {
		j++;    
		*pLocusPtNew = EEM_LOCUS_Get(pLocus, j);
		break;
	    }
	}
	if (eemInsideBox(pLocusPtNew, pBS)){ 
	    NextK = eemEvalK(pRat, &EEM_LOCUS_Get(pLocus,j)); 
	    pLP->PntGrp->DeltaH = DELTAH;
	    if (Sign == TOP_TRUE) { 
		pLP->CurLocus = &pLP->LocusPos;
	        eemDrawLocus(pLP, pLocusPtNew, pBS, pB, &NextK, 0.0, MAXPOS);
	    } else {
		pLP->CurLocus = &pLP->LocusNeg;
	        eemDrawLocus(pLP, pLocusPtNew, pBS, pB, &NextK, 0.0, MAXNEG);
	    }
	    #if __TOP_DEBUG__
	    if (Sign == TOP_TRUE) { 
		EEM_LOCUS_Print(SPKG, pLP, "ZoomLocusPos");
	    } else
		EEM_LOCUS_Print(SPKG, pLP, "ZoomLocusNeg");
	    #endif
	    absNextK = TOP_ABS(NextK);
	    for ( ; j<numLocusPts; j++) {
		thisK = eemEvalK(pRat, &EEM_LOCUS_Get(pLocus,j));
		 if ( TOP_ABS(thisK) > absNextK )
		     break;
	    }
	    for (; j<numLocusPts ; j++) {
		if (eemInsideBox(&EEM_LOCUS_Get(pLocus,j), pBS)) 
		    ;
	    }
	}
    }
}

/**
Calculating each root loci corresponding to 
each pole
If a locus appears outside the bounding box, 
calculation ends
There are three boxes.  pBox is the large search
box for zooming in.  pDrawBox is within the pBox;
points are stored only if inside pDrawBox.  In the
master plot, pBox is the same as pDrawBox 
**/

void
eemDrawLocus( EEMLocus *pLocus, EEMLocusPt *pLocusPtNew, EEMBox *pBox, EEMBox *pDrawBox, double *pk0, double KStart, double KEnd){
    int j, m, np, y, y2, r,z, u=0, v=-1;
    EEMBox DB;
    double pr, pi, qr, qi;
    EEMComplex Atemp, Btemp, PDen, PNum, Pt;  
    EEMComplex N,D,R,P;
    double k,t, w, KD = KEnd - KStart;
    EEMLociGrp *pLGrp = pLocus->MasterGrp;
    EEMSilarGrp *pSing = &pLGrp->Sing;
    EEMPolyCplxRat *pRat = &pLGrp->Rat;
    EEMLocusPt LastTwo, Last; 

    y=0;	y2 = 0;
    k = *pk0;

    LastTwo = Last = *pLocusPtNew;
    if (eemInsideBox(pLocusPtNew, pDrawBox))  
	if (TOP_ABS(k) >= TOP_ABS(KStart) && TOP_ABS(k) <= TOP_ABS(KEnd)){
	    EEM_LOCUS_Append(pLocus) = *pLocusPtNew; 
	    EEM_CPLX_Print(SPKG, pLocusPtNew, "\nstore"); 
	    #if __TOP_DEBUG__
	    printf( "\nk=%g",k);
	    #endif
    	}
    DB.lo.x = pLocusPtNew->r - pBox->delta.x;
    DB.hi.x = pLocusPtNew->r + pBox->delta.x;
    DB.lo.y = pLocusPtNew->i - pBox->delta.y;
    DB.hi.y = pLocusPtNew->i + pBox->delta.y;
    /*
     * If the Klist is empty, then u = -1
     */
    if (EEM_ARRAY_Num(&pLGrp->Klist) > 0) 
	t = EEM_ARRAY_Get(&pLGrp->Klist, y);
    else {
	u = -1;
	t = 0.0;
    }	
    printf("\nIn EEMDrawLocus: Box \t lo:(%g,%g), \t hi:(%g,%g), \t delta:(%g,%g)\n",pBox->lo.x, pBox->lo.y, pBox->hi.x, pBox->hi.y, pBox->delta.x, pBox->delta.y);

    for (j=0; eemInsideBox(pLocusPtNew, pBox); j++) {  
	if (j > 1000000) {
	    #if __TOP_DEBUG__
	    printf("\nError: non-convergence\n");
	    #endif
	    #if __TOP_DEBUG__
	    printf("Locus required %d major iterations\n", j);
	    #endif
	    *pk0 = k;
	    return;
	}
	/*
	 * Conditions are checked after every 2 
	 * points are calculated 
	 */
	for ( m=0; m<2; m++) { 
	    eemEvalPolyCplx ( &pLGrp->ADiff, pLocusPtNew, &Atemp);
	    eemEvalPolyCplx ( &pLGrp->BDiff, pLocusPtNew, &Btemp);
	    PDen.r = k * Atemp.r  + Btemp.r;
	    PDen.i = k * Atemp.i  + Btemp.i;
	    eemEvalPolyCplx ( &pRat->Num, pLocusPtNew, &Atemp);
	    eemEvalPolyCplx ( &pRat->Den, pLocusPtNew, &Btemp);

	    if (KEnd > 0.0) {
		PNum.r = (k + 1.0) * Atemp.r + Btemp.r;
		PNum.i = (k + 1.0) * Atemp.i + Btemp.i;
		k += (pLocus->PntGrp->DeltaH);
	    } else {
		PNum.r = (k - 1.0) * Atemp.r + Btemp.r;
		PNum.i = (k - 1.0) * Atemp.i + Btemp.i;
		k -= (pLocus->PntGrp->DeltaH);
	    }
	    eemCplxDiv2(&PNum, &PDen);
	    pLocusPtNew->r = pLocusPtNew->r - pLocus->PntGrp->DeltaH * PNum.r;
	    pLocusPtNew->i = pLocusPtNew->i - pLocus->PntGrp->DeltaH * PNum.i;
	}
	++y2;

	/*
	 * Checking for singularity:
	 * A LocusPt is singular if B'(s) + KA'(s) = 0
	 * KList is sorted list of K values corresponding
	 * to singular points
	 */
        if (u>=0 && 1.05*TOP_ABS(t)>TOP_ABS(k)&& TOP_ABS(t-k)<.05*TOP_ABS(t)){ 
	    if (v == -1)  
	        v = eemCloseToNthSilarInSilarGrp (pSing, pLocusPtNew);
	    w = eemEvalK(pRat, &EEM_SILAR_Get(pSing, v).loc);
	    if (TOP_ABS(t-w) < .00001*t)
		if (eemInsideBox(&EEM_SILAR_Get(pSing,v).loc,&DB)) {
		    *pLocusPtNew = EEM_SILAR_Get(pSing, v).loc; 
		    if (EEM_LOCUS_Num(pLocus) < 2)
			if (Last.r == LastTwo.r && Last.i == LastTwo.i)
			    printf("Error: In eemDrawLocus: A singular point has been met, but there are not enough points calculated to determine the direction vector, and the result may be incorrect");
		    eemMeetSingAndCorrect(pLocusPtNew, v, pLocus, &DB,
		      &Atemp, &LastTwo);
		    EEM_CPLX_Print(SPKG, pLocusPtNew, "\n\tpLocusPtNew ");
		    EEM_CPLX_Print(SPKG, &EEM_SILAR_Get(pSing, v).loc, 
		      "\tis close to Singular point");
		    Pt = EEM_SILAR_Get(pSing, v).loc;
		    if (pDrawBox == pBox) {
			if (TOP_ABS(k) >= TOP_ABS(KStart) && 
			  TOP_ABS(k) <= TOP_ABS(KEnd))
			    if (eemInsideBox(&Pt, pDrawBox)){
				EEM_LOCUS_Append(pLocus) = Pt;
				#if __TOP_DEBUG__
				printf( "\nk=%g",k);
				#endif
			        EEM_CPLX_Print(SPKG, &Pt, "\nStore");
			    }	
			    if (eemInsideBox(pLocusPtNew, pDrawBox)){ 
				EEM_LOCUS_Append(pLocus) = *pLocusPtNew; 
				#if __TOP_DEBUG__
				printf( "\nk=%g",k);
				#endif
			        EEM_CPLX_Print(SPKG, pLocusPtNew, "\nStore");	
			    }
	            } 
		    /*
		     * This is to deal with the problem that occurs when
		     * a singular point is met, but there is there are no
		     * previous points stored to calculate the direction
		     * vector. 
		     */
		    DB.lo.x = pLocusPtNew->r - pBox->delta.x;
		    DB.hi.x = pLocusPtNew->r + pBox->delta.x;
		    DB.lo.y = pLocusPtNew->i - pBox->delta.y;
		    DB.hi.y = pLocusPtNew->i + pBox->delta.y;

		    /**
		    #if __TOP_DEBUG__
		    printf("\ny2 = %d, y =%d  h = %g, ", y2, y, pLocus->PntGrp->DeltaH);
		    #endif
		    **/
		    y2 = 0; 	y = 0; 		v = -1; 	u++;
		    if ( u == EEM_SILAR_Num(pSing))
			u = -1;
		    else 	
			t = EEM_ARRAY_Get(&pLGrp->Klist, u);
                }
	} else if (u>=0 && (TOP_ABS(k)>1.05*TOP_ABS(t))) {
	    if (u !=EEM_SILAR_Num(pSing)-1 ) {
                if (t != 0.0 && k>1.01*t) {
		    pLocus->PntGrp->DeltaH /= 10;
		    z = EEM_LOCUS_Num(pLocus);
		    if (z>0)
		        *pLocusPtNew = EEM_LOCUS_Get(pLocus, z-1); 
		}		    
               	u++;
                t = EEM_ARRAY_Get(&pLGrp->Klist, u);
            }
	    else 		u = -1;
	    v = -1;
        }

	if (eemInsideBox(pLocusPtNew, &DB)) { 
	    if ( TOP_ABS(PNum.r) < ((pBox->delta.x) * .0001)
	      && TOP_ABS(PNum.i) < ((pBox->delta.y) * .0001)) {
	       #if __TOP_DEBUG__
	       printf("\nLocus terminated in a zero\n");
	       #endif
	       break;
	    }
	    if ( ++y > 30 ) {
		pLocus->PntGrp->DeltaH *= (y-10) / 15.0;
		y = 0;
	    }
	} else {
	    if ( y < 10 ) 
		pLocus->PntGrp->DeltaH *= (y+1) / 15.0;
	    /*
	    #if __TOP_DEBUG__
	    printf("\ny2 = %d, y =%d  h = %g,", y2, y, pLocus->PntGrp->DeltaH);
	    eemCplxDiv2 (&Btemp, &Atemp);
	    Btemp.r = -Btemp.r;
	    Btemp.i = -Btemp.i;
	    EEM_CPLX_Print(SPKG, &Btemp, "\n\t K = -B/A =");
	    #endif
	    */
	    y2 = 0;
	    y = 0;
	    if (eemInsideBox(pLocusPtNew, pDrawBox)) 
		if (EEM_LOCUS_Num(pLocus) < 3) {
		    LastTwo = Last; 
		    Last = *pLocusPtNew;
		}
		if (TOP_ABS(k) > TOP_ABS(KStart) && TOP_ABS(k) < TOP_ABS(KEnd)){
		    EEM_LOCUS_Append(pLocus) = *pLocusPtNew; 
		    #if __TOP_DEBUG__
		    printf( "k=%g",k);
		    #endif
		    EEM_CPLX_Print(SPKG, pLocusPtNew, "\nstore");
	    	}
	    DB.lo.x = pLocusPtNew->r - pBox->delta.x;
	    DB.hi.x = pLocusPtNew->r + pBox->delta.x;
	    DB.lo.y = pLocusPtNew->i - pBox->delta.y;
	    DB.hi.y = pLocusPtNew->i + pBox->delta.y;
	}
    }
    #if __TOP_DEBUG__
    printf("Locus required %d major iterations\n", j);
    #endif
    *pk0 = k;
}


void
eemMeetSingAndCorrect(EEMLocusPt *pLocusPtNew, int v, EEMLocus *pLocus, EEMBox *pDB, EEMComplex *pAtemp, EEMLocusPt *pLastTwo) {
    EEMComplex Ctemp; 
    double c, d, *pk;
    int q, s;
    EEMLociGrp *pLGrp=pLocus->MasterGrp;
    EEMPolyCplxRat *pRat=&pLGrp->Rat;
    EEMSilarGrp *pSing = &pLGrp->Sing;

    q = EEM_SILAR_Get(pSing, v).power;
    /*
     * The reason for using the next to last point instead
     * of the last point is because sometimes the last point
     * might be in the same box as the singular point, and
     * the last point might have overshot the singular
     * point, making the directon vector point in the
     * wrong direction
     */
    if (EEM_LOCUS_Num(pLocus) < 2)
	Ctemp = *pLastTwo;
    else 
	Ctemp = EEM_LOCUS_Get(pLocus,EEM_LOCUS_Num(pLocus)-2);
    EEM_CPLX_Print(SPKG, &Ctemp, "\n\tCtemp");
    EEM_CPLX_Print(SPKG, pLocusPtNew, "\n\tpLocusPtNew");


    /*
     * To prevent too big of a step, the direction vector
     * is divided by 2
     */
    Ctemp.r = (pLocusPtNew->r - Ctemp.r)/2.0;
    Ctemp.i = (pLocusPtNew->i - Ctemp.i)/2.0;

    EEM_CPLX_Print(SPKG, &Ctemp, "\n\tDirection Vector:");
 
    if ( q & 1) {	/* odd q */
	pLocusPtNew->r += Ctemp.r;
	pLocusPtNew->i += Ctemp.i; 
    } else {	
	/*
	 * dz = ds * exp(-jPI/q)
	 * where ds is a small vector in the
	 * tangential direction of the locus
	 * when approaching the singular
	 * point
	 * dz = ds * (cos(PI/q) -j sin(PI/q)) 
	 *    = (Ctemp.r+jCtemp.i)*(c+jd)
	 *    = c*Ctemp.r-d*Ctemp.i +
	 *	     j( c* Ctemp.i + d*Ctemp.r)
	 */
	c =  cos (M_PI/q);
	d = -sin (M_PI/q);
	pLocusPtNew->r += 2*(Ctemp.r * c - Ctemp.i * d); 
	pLocusPtNew->i += 2*(Ctemp.i * c + Ctemp.r * d); 
    } 
    /* 
     * This is to correct the jump over 
     * singularity
     */
    pk = &EEM_SILAR_Get(pSing, v).extra.val;
    for ( s = 0; TOP_ABS(pAtemp->i) >pLGrp->DiffEpsilon && s<50; s++) 
	eemNewtonIteration(pLocusPtNew, pLGrp, *pk);
    if (s >= 50) 	
	EEM_CPLX_Print(SPKG, pLocusPtNew, "\n 50th iteration of Newton:");
    EEM_SILAR_Get(pSing, v).extra.val =eemEvalK( pRat, pLocusPtNew);
}
