/* 
    Author: Mei-Tjng Huang & Kennard White

    Copyright (c) 1990-93 The Regents of the University of California.
    All rights reserved.  See the file "./copyright.h" for full copyright
    notice, limitation of liability, and disclaimer of warranty provisions.
#include "copyright.h"				COPYRIGHTENDKEY
*/
#if !defined(lint) && !defined(SABER)
static char rcsid[] = "$Header: /vol/cooley/cooley1/kennard/src/eem/RCS/pfe.c,v 1.17 93/08/27 14:29:35 kennard Exp $";
#endif


#include "topFixup.h"
#include <stdio.h>
#include <math.h>

#include "topStd.h"
#include "topVec.h"

#include "eemDflt.h"
#include "eemCplx.h"
#include "eemSilar.h" 
#include "eemPoly.h"
#include "eemPfe.h"

char _EemPfeSubPackage[] = "eem.pfe";
#define SPKG _EemPfeSubPackage


/**************************************************************************
 *
 *		 PfeTerm & PfeTermGrp funcs
 *
 **************************************************************************/

void
eemPfeGrpConstructor( EEMPfeGrp *pGrp) {
    pGrp->shift = 0;
    eemPolyCplxConstructor(&pGrp->poly);
    topVecConstructor( &pGrp->terms, sizeof(EEMPfe));
}

void
eemPfeGrpDestructor( EEMPfeGrp *pGrp) {
    eemPolyCplxDestructor(&pGrp->poly);
    topVecDestructor( &pGrp->terms);
}

void 
eemPfeGrpSetNum( EEMPfeGrp *pGrp, int n){
    topVecSetNum( &pGrp->terms, n);
}

EEMPfe*
eemPfeGrpAppend( EEMPfeGrp *pGrp, EEMComplex *pGain, EEMComplex *pPole, 
  int power) {
    int		i;
    EEMPfe	*pPfe;

    i = topVecAppendIdx( &pGrp->terms);
    pPfe = &EEM_PFE_Get(pGrp, i);
    pPfe->gain = *pGain;
    pPfe->pole = *pPole;
    pPfe->power = power;
    pPfe->numerpow = 0;
    return pPfe;
}

void 
eemPfePrint(EEMPfe *pPfe, char *title) {
    if ( title != NULL )
        printf("%s:\t", title);
    printf("{ (%g,%g) (%g,%g) %d }\n",
      pPfe->gain.r, pPfe->gain.i, pPfe->pole.r, pPfe->pole.i, pPfe->power);
}

void
eemPfeGrpPrint( EEMPfeGrp *pGrp, char *title) {
    printf("%s:\n", title);
    printf("\tShift %d\n", pGrp->shift);
    eemPolyCplxPrint( &pGrp->poly, "\tPoly ");
    EEM_PFE_LOOP_BEGIN( pGrp, pTerm) {
	eemPfePrint( pTerm, "\t");
    } EEM_PFE_LOOP_END();
}

#ifdef notdef
These are broken and the argument orders dont follow conventions
void 
eemPfeDenOneToPolyCplx(EEMPolyCplx *pD, EEMPfe *pS){
    EEM_POLYCPLX_Get(pD,0).r = -pS->Pole.r;
    EEM_POLYCPLX_Get(pD,0).i = -pS->Pole.i;
    EEM_POLYCPLX_Get(pD,1).r = 1.0;
    EEM_POLYCPLX_Get(pD,1).i = 0.0;
}

void 
eemPfeDenToPolyCplx(EEMPolyCplx *pD, EEMPfe *pS){
    EEMPolyCplx Temp;
    eemPfeDenOneToPolyCplx(&Temp, pS);
    eemPolyCplxPower2(pD, &Temp, pS->Order);
}
#endif

void
eemPfeGrpEvalImpulseLin( EEMPfeGrp *pGrp, double t0, double dt,
  EEMComplex *impulsevec, int numsamps) {
    EEMComplex	*iv;
    double	t;
    int		i,j;
    EEMComplex	val;

    for (i=0, iv=impulsevec, t=t0; i < numsamps; i++, iv++, t += dt) {
	iv->r = iv->i = 0.0;
	for (j=0; j<EEM_PFE_Num(pGrp); j++) {
       	    eemPfeToTimeS( &EEM_PFE_Get(pGrp,j), t, &val);
	    eemCplxAdd2( iv, &val);
	}
    }
}

/*
 * Take input a partial fraction expansion (pole(s) at one location)
 * and inverse laplace transform into time domain at time t0.
 *
 * XXX: This hasnt been tested very much
 * XXX: This doesnt deal with zeros/poles at the origin
 *
 */
void 
eemPfeToTimeS( EEMPfe *pPfe, double t0, EEMComplex *pVal) {
    double	temp;
    int		n = TOP_ABS(pPfe->power) - 1;

    if ( t0 <= 0 || n < 0 ) {
	EEM_CPLX_ZERO( pVal);
	return;
    }
    /* val = exp(j Im{p} * t) = cos(Im{p} * t) + j sin( Im{p} t) */
    sincos( pPfe->pole.i * t0, &pVal->i, &pVal->r);

    /* val *= exp( Re{p} * t) * t^n / n! */
    temp = exp(pPfe->pole.r * t0);
    if ( n > 0 ) {
	temp *= n==1 ? t0 : pow(t0,(double)n)/eemFactorial(n);
    }
    pVal->r *= temp;
    pVal->i *= temp;

    /* val *= gain */
    eemCplxMul2( pVal, &pPfe->gain);
}

void 
eemPfeGrpToTimeS( EEMPfeGrp *pPfeGrp, double t0, EEMComplex *pVal) {
    EEMComplex	res;

    EEM_CPLX_ZERO(&res);
    EEM_PFE_LOOP_BEGIN( pPfeGrp, pPfeTerm) {
	EEMComplex  val;
	eemPfeToTimeS( pPfeTerm, t0, &val);
	eemCplxAdd2( &res, &val);
    } EEM_PFE_LOOP_END();
    *pVal = res;
}

/******************************************************************************

		Z-domain PFE --> time

    Each pfe term consists of {gain,pole,power}.  This describes
    a Z-domain term:
	      gain
	---------------
	(z-pole)^-power
    Note that the {power} stored in the pfe is negative.
    From Z inverse tables:

    For ZT X(z), the causal ZT inverse x(n) is:
	1 / (1-p*z^-1)	  <-->	p^n u(n)			[power==1]
	1 / (1-p*z^-1)^2  <-->	(n+1) p^n u(n)			[power==2]
	1 / (1-p*z^-1)^m  <-->	A (n+1) ... (n+m-1) p^n u(n)	[power>=2]
				with  A=1/((m-1)!)

    We can then convert from z^-1 to z^+1:
	z / (z-p)	<-->	p^n u(n)			[power==1]
	z^m / (z-p)^m	<-->	A (n+1) ... (n+m-1) p^n u(n)	[power>=2]
    And using time-shift properties:
	1 / (z-p)	<-->	p^(n-1) u(n-1)
	1 / (z-p)^m	<-->	A (n-m+1) ... (n-m+m-1) p^(n-m) u(n-m)

    Note that by performing the PFE in z and not in z^-1, we get a division
    between the FIR and IIR components: The pfegrp.poly yields
    FIR components from x(0)...x(m-1), and the pfegrp.terms
    yields IIR components from x(m)...x(inf).

******************************************************************************/

/**
    Evaluate the impulse-response component to a single pfe term.
    Note that this could be optimized by supporting conjugate pfes.
    WARNING: {n} must be adjusted to allow for {pPfeGrp->shift}.
**/
void 
eemPfeToTimeZ( EEMPfe *pPfe, int n, EEMComplex *pVal) {
    int		m = - pPfe->power;

    if ( m <= 0 || n <= m-1 ) {
	EEM_CPLX_ZERO( pVal);
	return;
    }
    *pVal = pPfe->pole;
    eemCplxPow2( pVal, n-m);
    eemCplxMul2( pVal, &pPfe->gain);

    if ( m >= 2 ) {
	/* compute (n-m+1)...(n-1) */
	double	aval = 1.0;
	int	i;

	for ( i=1; i < m; i++) {
	    aval *= n-i;
	}
	aval /= eemFactorial(m-1);
	pVal->r *= aval;
	pVal->i *= aval;
    }
}

/**
    Evaluate the impulse-response of {pPfe}, using z-domain (discrete-time)
    semantics for the free variable.  {pPfe->shift} is treated as
    a time shift.
    Note that this could be optimized by supporting conjugate pfes.
    WARNING: This function ignores {pPfe->shift}.
**/
void 
eemPfeGrpToTimeZ( EEMPfeGrp *pPfeGrp, int n, EEMComplex *pVal) {
    EEMComplex	res;

    n += pPfeGrp->shift;
    if ( -n >= 0 && -n < EEM_POLYCPLX_Num(&pPfeGrp->poly) ) {
	res = EEM_POLYCPLX_Get(&pPfeGrp->poly, -n);
    } else {
	EEM_CPLX_ZERO(&res);
    }
    EEM_PFE_LOOP_BEGIN( pPfeGrp, pPfeTerm) {
	EEMComplex  val;
	eemPfeToTimeZ( pPfeTerm, n, &val);
	eemCplxAdd2( &res, &val);
    } EEM_PFE_LOOP_END();
    *pVal = res;
}

#ifdef notdef
/*
 * Take input a partial fraction expansion (conjugate poles of order 1) 
 * and inverse laplace transform into time domain at time t0.
 */
void 
eemConjPfeToTime( EEMPfe *pPfe, double t0, EEMComplex *pVal) {
    pVal->r = exp(pPfe->pole.r * t0) * sin(pPfe->pole.i *t0);
    pVal->i = 0;
}
#endif
