/* 
    Author: Mei-Tjng Huang & Kennard White

    Copyright (c) 1990-93 The Regents of the University of California.
    All rights reserved.  See the file "./copyright.h" for full copyright
    notice, limitation of liability, and disclaimer of warranty provisions.
#include "copyright.h"				COPYRIGHTENDKEY
*/
#if !defined(lint) && !defined(SABER)
static char rcsid[] = "$Header: /vol/cooley/cooley1/kennard/src/eem/RCS/polysilar.c,v 1.14 93/08/27 14:29:30 kennard Exp $";
#endif

#include "topFixup.h"
#include <stdio.h>
#include <math.h>

#include "topStd.h"
#include "topMem.h"
#include "topVec.h"

#include "eemDflt.h"
#include "eemCplx.h"
#include "eemSilar.h"
#include "eemPoly.h"

extern char _EemPolySubPackage[];
#define SPKG _EemPolySubPackage

/**
  Converts a silar to a  complex polynomial, taking into account
  the power of the  silar
**/
void 
eemSilarToPolyCplx( EEMSilar *pSilar, EEMPolyCplx *pD) {
    if ( EEM_SILAR_ConjB(pSilar) ) {
	/*
	 * (z-x)(z-x~) = z*z -2*Re{x}*z + xx~, where xx~ is real
	 * for x=(a+jb), xx~ = a*a + b*b = magsq(x)
	 */
	eemPolyCplxSetNum(pD, 3);
	EEM_CPLX_SET(&EEM_POLYCPLX_Get(pD,2), 1.0, 0.0);
	EEM_CPLX_SET(&EEM_POLYCPLX_Get(pD,1), -2 * pSilar->loc.r, 0.0);
	EEM_CPLX_SET(&EEM_POLYCPLX_Get(pD,0), 
	  EEM_CPLX_MAGSQ(&pSilar->loc), 0.0);
    } else {
	eemPolyCplxSetNum(pD, 2);
	EEM_POLYCPLX_Get(pD,0).r = - pSilar->loc.r;
	EEM_POLYCPLX_Get(pD,0).i = - pSilar->loc.i;
	EEM_CPLX_SET(&EEM_POLYCPLX_Get(pD,1), 1.0, 0.0);
    }
    eemPolyCplxPower2( pD, TOP_ABS(pSilar->power));
}


/**
    Extract either all poles or all zeros, multiply them out,
    and put the resulting expression into {pPoly}.
    The flag {sign} indicates if poles or zeros should be converted:
    any positive value will use zeros, any negative value will use poles.
    If sign==0, then the silargrp's gain will be included in the poly,
    and the zeros will be placed into the poly (.e.g, pos sign).
    More formally, consider {sign > 0}: 
	Let z_i be the zeros or poles of {pSG}.
	Let m_i be the power of z_i.  Then:
        pPoly(s) = product over {i} of (s-z_i)^m_i 
**/
void 
eemSilarGrpToPolyCplx( EEMSilarGrp *pSG, EEMPolyCplx *pPoly, int sign) {
    int		p;
    EEMPolyCplx	silpoly;

    eemPolyCplxConstructor( &silpoly);
    eemPolyCplxOne( pPoly);
    if (sign==0) {
	EEM_POLYCPLX_Get(pPoly,0) = pSG->gain;
	sign = 1;
    }
    EEM_SILAR_LOOP_BEGIN( pSG, pSilar) {
	p = pSilar->power;
	if ( (sign < 0 && p < 0) || (sign > 0 && p > 0) ) {
            eemSilarToPolyCplx( pSilar, &silpoly);
            eemPolyCplxMul2( pPoly, &silpoly);
	}
    } EEM_SILAR_LOOP_END();
    eemPolyCplxShorten(pPoly);
    eemPolyCplxDestructor(&silpoly);
}

/**
  Converts a Silar Grp of singularities to fraction of complex
  polynomials (a rational polynomial).  It's obtained by doing the expansion
     gain * Product of (s-z_i) over i
	   divided by
     Product of (s-p_i) over i
  where z_i are Silar.loc with positive power
	p_i are Silar.loc with negative power
  The resulting fraction may be an improper fraction, that is,
  the order of the numerator can be greater than that of the
  denominator.
**/
void 
eemSilarGrpToPolyCplxRat( EEMSilarGrp *pSG, EEMPolyCplxRat *pRat) {
    eemPolyCplxZero( &pRat->Whole, 0);
    eemSilarGrpToPolyCplx( pSG, &pRat->Numer, 0);
    eemSilarGrpToPolyCplx( pSG, &pRat->Denom, -1);
}

/**
    This function is a wrapper around cpoly() a complex polynomial
    factoring routine.  We massage the data between our choosen formats
    (EEMPolyCplx, EEMSilarGrp) and cpoly's formats (arrays).  The roots
    of {pPoly} are added to {pSG} as either poles or zeros according to
    {sign}.

    Notes:
    -	For a {pPoly} of size n, (n coefs), the degree is n-1,
    	and n-1 complex zeros will be found.
    -	We handle several special cases ourselves, both for efficiency and
	because cpoly() breaks on some of the simpler examples.
    -	We currently dont touch the gain of {pSG}.
**/
TOPLogical 
eemPolyCplxToSilarGrp( EEMPolyCplx *pPoly, EEMComplex *pGain,
  EEMSilarGrp *pSG, int sign) {
    int		n, k, n_in_out;
    TOPLogical	FailB;
    double	PolyR[EEM_CPOLY_SIZE], PolyI[EEM_CPOLY_SIZE];
    double	ZeroR[EEM_CPOLY_SIZE], ZeroI[EEM_CPOLY_SIZE];
    EEMComplex	loc;

    n = EEM_POLYCPLX_Num(pPoly);
    /*
     * find {k}: the idx of the lowest order non-zero coef.  Note that
     * since we dont really no the magnitude of the numbers, we require
     * exactly zero, not just eemCplxZeroB().
     */
    for ( k=0; k < n && eemCplxZeroB(&EEM_POLYCPLX_Get(pPoly,k)); k++) {
	if ( EEM_POLYCPLX_Get(pPoly,k).r != 0.0 
	  || EEM_POLYCPLX_Get(pPoly,k).i != 0.0 )
	   break;
    }
    if ( k > 0 ) {
	eemSilarGrpAppend( pSG, 0, 0.0, 0.0, k*sign);
    }
    n -= k;
    if ( n <= 0 ) {
        if ( pGain )	EEM_CPLX_ZERO(pGain);
	return TOP_TRUE;
    }
    if ( pGain ) *pGain = EEM_POLYCPLX_Get(pPoly,n-1+k);
    if ( n <= 1 )
	return TOP_TRUE;
    if ( n == 2 ) {
	loc = EEM_POLYCPLX_Get(pPoly,k);
	eemCplxDiv2( &loc, &EEM_POLYCPLX_Get(pPoly,k+1));
	eemSilarGrpAppend( pSG, 0, -loc.r, -loc.i, sign);
	return TOP_TRUE;
    }

    eemPolyCplxToArrays( pPoly, PolyR, PolyI, k);
   
    n_in_out = n-1;
    cpoly_(PolyR, PolyI, &n_in_out, ZeroR, ZeroI, &FailB);
    if (FailB) {
        eemPolyCplxPrint(pPoly, "Error: Failure in factoring polynomial");
	return FALSE;
    }
    /* n should be the #of zeros, which is the degree of the poly */
    eemArraysToSilarGrp( ZeroR, ZeroI, n_in_out, pSG, sign);
    return TRUE;
}

/**
  This function takes two arrays: pReal and pImag, and treats         
  (pReal[i], pReal[i]) as a zero (for i=0...n-1) and append it to the SilarGrp.
  If a zero appears multiple times at the same location, then the power 
  of the zero is increased, but the size of the Silar Grp is only
  increased at the first occurence of the zero
**/
void 
eemArraysToSilarGrp( double *pReal, double *pImag, int n, 
  EEMSilarGrp *pSG, int sign) {
    int		i, qidx;
    EEMSilar	*pKeySilar;
    EEMComplex	zloc;

    for ( i=0; i < n; i++, pReal++, pImag++) {
	EEM_CPLX_SET( &zloc, *pReal, *pImag);
        if ( TOP_ABS(zloc.r) < EEMParams.Reltol) 	zloc.r = 0.0;
        if ( TOP_ABS(zloc.i) < EEMParams.Reltol) 	zloc.i = 0.0;
        if (  (qidx = eemSilarGrpFindIdxByLoc(pSG, &zloc)) >= 0 ) {
	    pKeySilar = &EEM_SILAR_Get(pSG, qidx);
	    if ( EEM_SILAR_ConjB(pKeySilar) ) {
		topErrMsg("cpoly: conj/non-conj split in power compression");
	    }
            pKeySilar->power += sign;
	} else {
	    pKeySilar = &EEM_SILAR_Append(pSG);
	    pKeySilar->loc = zloc;
	    pKeySilar->power = sign;
        }
    }
}
