/* 
    Author: Kennard White

    Copyright (c) 1990-93 The Regents of the University of California.
    All rights reserved.  See the file "./copyright.h" for full copyright
    notice, limitation of liability, and disclaimer of warranty provisions.
#include "copyright.h"				COPYRIGHTENDKEY
*/
#if !defined(lint) && !defined(SABER)
static char rcsid[] = "$Header: /vol/cooley/cooley1/kennard/src/eem/RCS/silar.c,v 1.27 93/08/27 14:29:28 kennard Exp $";
#endif

#include "topFixup.h"
#include <stdio.h>
#include <math.h>

#include "topStd.h"
#include "topMem.h"
#include "topVec.h"

#include "eemDflt.h"
#include "eemCplx.h"
#include "eemSilar.h"

char _EemSilarSubPackage[] = "eem.silar";
#define SPKG _EemSilarSubPackage


void
eemSilarConstructor( EEMSilar *pSilar) {
    pSilar->flags = 0;
    pSilar->power = 0;
    pSilar->loc.r = 0.0;
    pSilar->loc.i = 0.0;
    pSilar->extra.pnt = NULL;
}

void
eemSilarDestructor( EEMSilar *pSilar) {
    ;	/* its silly to even call this */
}

void
eemSilarGrpConstructor( EEMSilarGrp *pGrp) {
    EEM_CPLX_SET( &pGrp->gain, 1.0, 0.0);
    topVecConstructor( &pGrp->silars, sizeof(EEMSilar));
}

void
eemSilarGrpDestructor( EEMSilarGrp *pGrp) {
    topVecDestructor( &pGrp->silars);
}

void
eemSilarGrpSetNum( EEMSilarGrp *pGrp, int n) {
    int i, curnum = EEM_SILAR_Num(pGrp);

    topVecSetNum( &pGrp->silars, n);
    for (i=curnum; i < n; i++) {	/* if expanding, init the new ones */
	eemSilarConstructor( &EEM_SILAR_Get(pGrp,i));
    }
}

void
eemSilarGrpInsureNum( EEMSilarGrp *pGrp, int n) {
    if ( n > EEM_SILAR_Num(pGrp) )
	eemSilarGrpSetNum( pGrp, n);
}

int
eemSilarGrpAppendIdx( EEMSilarGrp *pGrp) {
    int		i;

    i = topVecAppendIdx( &pGrp->silars);
    eemSilarConstructor( &EEM_SILAR_Get(pGrp,i));
    return i;
}

EEMSilar*
eemSilarGrpAppend( EEMSilarGrp *pGrp, TOPMask f, double lr, double li, int p) {
    int		i;
    EEMSilar	*pSilar;
    
    i = topVecAppendIdx( &pGrp->silars);
    pSilar = &EEM_SILAR_Get(pGrp,i);
    pSilar->flags = f;
    pSilar->loc.r = lr;
    pSilar->loc.i = li;
    pSilar->power = p;
    pSilar->extra.pnt = NULL;
    return pSilar;
}

/**
    Append a copy of silar {pSrc} to silargoup {pGrp}.
    Returns a pointer to the copy.
**/
EEMSilar*
eemSilarGrpAppendCopy( EEMSilarGrp *pGrp, EEMSilar *pSrc) {
    int		 i;
    EEMSilar	*pDst;

    i = topVecAppendIdx( &pGrp->silars);
    pDst = &EEM_SILAR_Get(pGrp,i);
    *pDst = *pSrc;
    return pDst;
}

void
eemSilarGrpCopy( EEMSilarGrp *pSrc, EEMSilarGrp *pDst) {
    pDst->gain = pSrc->gain;
    topVecCopy( &pDst->silars, &pSrc->silars);
}

void
eemSilarGrpRemoveIdx( EEMSilarGrp *pGrp, int idx) {
    topVecRemoveIdx( &pGrp->silars, idx, 1);
}

/**
   When a silar is killed (for whatever reason), the power is set to 0.
   This silar cant generally be re-used right away, since some idle-callback
   code might not have seen the death yet.  When an application has
   decided that a silar in the vector can be safely reused, it can
   set the EEMSilarF_Avail flag.

   This function searches for a silar with the Avail bit set
   (and turns it off), or allocates a new silar on the end
**/
int
eemSilarGrpGetAvail( EEMSilarGrp *pGrp) {
    EEM_SILAR_IDX_LOOP_FORW_BEGIN(pGrp, idx, pSilar) {
	if ( TOP_IsOn(pSilar->flags, EEMSilarF_Avail) ) {
	    pSilar->flags &= ~EEMSilarF_Avail;
	    if ( pSilar->power != 0 )
		continue;	/* its already in use by someone */
	    return idx;
	}
    } EEM_SILAR_IDX_LOOP_FORW_END();
    return eemSilarGrpAppendIdx( pGrp);
}

/**
    As described in eemSilarGrpGetAvail(), the reuse of silars
    must be delayed in some applications.  This function sets the
    F_Avail flag for all silars with power==0.
    GC = GarbageCollect
**/
void
eemSilarGrpGC( EEMSilarGrp *pGrp) {
    EEM_SILAR_IDX_LOOP_FORW_BEGIN(pGrp, idx, pSilar) {
	if ( pSilar->power == 0 )
	    pSilar->flags = EEMSilarF_Avail;
	pSilar->flags &= ~EEMSilarC_All;
    } EEM_SILAR_IDX_LOOP_FORW_END();
}

void
eemSilarGrpCompact( EEMSilarGrp *pGrp) {
    EEMSilar	*pBase, *pTgt;

    pBase = pTgt = EEM_SILAR_Array(pGrp);
    EEM_SILAR_IDX_LOOP_FORW_BEGIN(pGrp, idx, pSrc) {
	if ( pSrc->power != 0 ) {
	    if ( pTgt != pSrc ) {
	        *pTgt = *pSrc;
	        pSrc->power = 0;
	    }
	    ++pTgt;
	}
    } EEM_SILAR_IDX_LOOP_FORW_END();
    eemSilarGrpSetNum( pGrp, pTgt - pBase);
}

void 
eemSilarPrint(EEMSilar *pSilar, char *title) {
    if ( title != NULL )
        printf("%s:\t", title);
    printf("(%g,%g) %d %s%s\n", pSilar->loc.r, pSilar->loc.i,
      pSilar->power, 
      TOP_IsOn(pSilar->flags,EEMSilarF_Conj) ? "conj " : "",
      TOP_IsOn(pSilar->flags,EEMSilarF_Sel) ? "sel " : "");
}

void
eemSilarGrpPrint( EEMSilarGrp *pGrp, char *title) {
    char	buf[10];

    printf("SilarGrp %s:\n", title);
    printf("\tGain = (%g,%g)\n", pGrp->gain.r, pGrp->gain.i);
    EEM_SILAR_IDX_LOOP_FORW_BEGIN( pGrp, idx, pSilar) {
	if ( pSilar->power == 0 )
	    continue;
	sprintf( buf, "%d", idx);
	eemSilarPrint( pSilar, buf);
    } EEM_SILAR_IDX_LOOP_FORW_END();
}

int
eemSilarGrpGetPower( EEMSilarGrp *pGrp, int *pZeroPow, int *pPolePow) {
    int		p, zp = 0, pp = 0;

    EEM_SILAR_LOOP_BEGIN( pGrp, pSilar) {
	p = pSilar->power;
	if ( EEM_SILAR_ConjB(pSilar) )		p *= 2;
	if ( p > 0 )				zp += p;
	else					pp += p;
    } EEM_SILAR_LOOP_END();
    if ( pZeroPow != NULL )			*pZeroPow = zp;
    if ( pPolePow != NULL )			*pPolePow = - pp;
    return zp - pp;
}

int
eemSilarGrpExtractOrigin(EEMSilarGrp *pGrp) {
    int		cnt = 0;

    EEM_SILAR_LOOP_BEGIN( pGrp, pSilar) {
	pSilar->flags &= ~EEMSilarF_Tag2;
	if ( pSilar->power==0 )	continue;
	if ( eemCplxZeroB( &pSilar->loc) ) {
	    cnt += EEM_SILAR_TotPower(pSilar);
	    pSilar->power = 0;
	    pSilar->flags |= EEMSilarF_Tag2;
	}
    } EEM_SILAR_LOOP_END();
    return cnt;
}

/**
    Fill in a new silar grp {pTgt} that is a copy of {pSrc},
    except that each EEMSilarF_Conj are flattened (split into
    to independent vectors).  Also, zero-power silars will be dropped,
    and if {pOriginCnt} is non-NULL, silars at the origin will be
    droped and their net power stored in {*pOriginCnt}.
    
    We use to do this in-place for efficiency, however, we now copy to
    a new vec for two reasons:
    1) For greater numerical stability, we want to keep both of the split
       silars together.  With in-place, this introduces holes in the array.
    2) The vector may re-allocated when appending to the end, which leads
       to some insideous bugs.
**/
void
eemSilarGrpFlattenConj(EEMSilarGrp *pSrc, EEMSilarGrp *pTgt, int *pOriginCnt) {
    EEMSilar	*pConjSilar;

    pTgt->gain = pSrc->gain;
    topVecSetNum( &pTgt->silars, 0);
    if ( pOriginCnt )	*pOriginCnt = 0;
    EEM_SILAR_LOOP_BEGIN( pSrc, pSilar) {
	if ( pSilar->power==0 )	continue;
	if ( pOriginCnt && eemCplxZeroB( &pSilar->loc) ) {
	    *pOriginCnt += EEM_SILAR_TotPower(pSilar);
	} else if ( TOP_IsOn(pSilar->flags,EEMSilarF_Conj) ) {
	    pConjSilar = eemSilarGrpAppendCopy( pTgt, pSilar);
	    pConjSilar->flags &= ~EEMSilarF_Conj;
	    pConjSilar = eemSilarGrpAppendCopy( pTgt, pSilar);
	    pConjSilar->flags &= ~EEMSilarF_Conj;
	    pConjSilar->loc.i = - pConjSilar->loc.i;
	} else {
	    eemSilarGrpAppendCopy( pTgt, pSilar);
	}
    } EEM_SILAR_LOOP_END();
}

/**
    Observe that we can do compress in both powers and in conjugateness;
    this func doesnt conj only.
    Thus there can be undesiable interactions between compress
    powers vs compressing conjness.  It is prob. best to
    first do powers, then do conjness.
**/
void
eemSilarGrpCompressConj(EEMSilarGrp *pGrp) {
    EEM_SILAR_IDX_LOOP_FORW_BEGIN( pGrp, fi, pSilar) {
	if ( pSilar->power==0 || EEM_SILAR_ConjB(pSilar) )
	    continue;
    	EEM_SILAR_IDX_LOOP_BACK_BEGIN( pGrp, bi, pConj) {
	    if ( bi <= fi )
		break;
	    if ( pConj->power!=pSilar->power || EEM_SILAR_ConjB(pConj) )
	        continue;
	    if ( eemCplxConjSameB( &pConj->loc, &pSilar->loc) ) {
		pSilar->flags |= EEMSilarF_Conj;
		pConj->power = 0;
		break;
	    }
        } EEM_SILAR_IDX_LOOP_BACK_END();
    } EEM_SILAR_IDX_LOOP_FORW_END();
}

/**
If two or more singular points are very close and have the
same silar.extra.val, then one of them is deleted
**/
void
eemDeleteMultSing(EEMSilarGrp *pSing ) {
    int i, q, w, n, c, d, m;
    EEMComplex temp1, temp2;
    m = EEM_SILAR_Num(pSing);
    if (m < 1)
	return;
    for (i=0; i<m ; i++) {
        q=EEM_SILAR_Get(pSing, i).power;
        if (q==0) 
	    continue;
	temp1=EEM_SILAR_Get(pSing,i).loc;
	/*
	 * w keeps track of the number of singular
	 * points deleted.  w should be less than q-2
	 */
        for (w=0, n=i+1; q>2 && w<q-2 && n<EEM_SILAR_Num(pSing); n++){
	    if (EEM_SILAR_Get(pSing, n).power==0)
		continue;
	    temp2=EEM_SILAR_Get(pSing,n).loc;
	    /*
	     * Two points are close if |r_1-r_2|^2 / ((r_1)^2 + (r_2)^2)
	     * is less than .001
	     */
	    d=EEM_CPLX_MAGSQ(&temp1)+EEM_CPLX_MAGSQ(&temp2);
	    eemCplxSub2(&temp1, &temp2);
	    d=EEM_CPLX_MAGSQ(&temp1)/d;
	    if (d < .01) {
		c = EEM_SILAR_Get(pSing, i).extra.val;
		if (EEM_SILAR_Get(pSing, n).power == q &&
		  TOP_ABS(EEM_SILAR_Get(pSing, n).extra.val-c)<=.00001*c){ 
		    EEM_SILAR_Get(pSing,n).power=0;
		    w++;
		}
	    } 
	}	
        if (q >2 && w != q-2) {    	     
	    printf("\n\nERROR: q-2 does not match the number of singular points deleted"); 
	    break;
	}
    }
    eemSilarGrpCompact(pSing);
}

/**
  Extracts either the zeros (sign>=0) or the poles (sign<0) of {pGrp},
  considers them as a complex polynomial is factored form, evaluates
  them at {pIn}, and stores the result in {pOut}.  The number of
  poles or zeros found is returned.
**/
int
eemSilarGrpEvalPart(EEMSilarGrp *pGrp, EEMComplex *pIn, EEMComplex *pOut,
  int sign) {
    int		power, abspow, count = 0;
    double	sr, si;
    EEMComplex	in, temp, insqr;

    in = *pIn;
    EEM_CPLX_SET(pOut, 1.0, 0.0);
    eemCplxSqr2( &insqr, &in);
    EEM_SILAR_IDX_LOOP_FORW_BEGIN(pGrp,idx,pSilar) {
	power = pSilar->power;
	if ( power==0 || ((sign>=0) ^ (power>=0)) )
	    continue;
	++count;
	sr = pSilar->loc.r;
	si = pSilar->loc.i;
	if ( TOP_IsOn(pSilar->flags,EEMSilarF_Conj) ) {
	    /*
	     * z=pIn, x= pSilar->loc
	     * (z-x)(z-x~) = z*z -2*Re{x}*z + xx~, where xx~ is real
	     * for x=a+jb, xx~ = (a+jb)(a-jb) = a^2 + b^2
	     */
	    temp.r = insqr.r - 2*sr*in.r + sr*sr + si*si;
	    temp.i = insqr.i - 2*sr*in.i;
	} else {
	    temp.r = in.r - sr;
	    temp.i = in.i - si;
	}

	abspow = TOP_ABS(power);
	if ( abspow > 1 )
	    eemCplxPow2(&temp, abspow);
	eemCplxMul2(pOut, &temp);
    } EEM_SILAR_IDX_LOOP_FORW_END();
    return count;
}

/**
  Treats the silar group as a fraction of complex polynomials in factored
  form in the s domain, and evaluated the fraction at s=pVal, and stores
  the output at pOut.
**/
void
eemSilarGrpEval(EEMSilarGrp *pGrp, EEMComplex *pIn, EEMComplex *pOut) {
    EEMComplex	temp;

    *pOut = pGrp->gain;
    if ( eemSilarGrpEvalPart( pGrp, pIn, &temp,  1) > 0 )
	eemCplxMul2( pOut, &temp);
    if ( eemSilarGrpEvalPart( pGrp, pIn, &temp, -1) > 0 )
	eemCplxDiv2( pOut, &temp);
}


/**
  Treats the silar group as a fraction of complex polynomials in factored
  form in the s domain; this function evaluates the fraction at s=pval
  with the ith silar removed.  The silar removal is done by setting the
  power of the ith silar to 0, then returning the power of the ith silar
  to its original multiplicity after the computation.  
**/
void
eemSilarGrpEvalA0(EEMSilarGrp *pGrp, EEMSilar *pSilar, EEMComplex *pOut) {
    int		p;

    p = pSilar->power;
    pSilar->power = 0;
    eemSilarGrpEval(pGrp, &pSilar->loc, pOut);
    pSilar->power = p;
}

/**
    Evaluates the absolute (unwrapped) phase of pGrp wrt {pIn} in the plane.
    This is similar to eemSilarGrpEval() and taking the phase (atan2())
    of the result.  But that returns the wrapped phase, while this
    returns the absolute phase (e.g., without any k*2*PI ambiguity).
**/
double
eemSilarGrpEvalAbsPhase(EEMSilarGrp *pGrp, EEMComplex *pIn) {
    int		power;
    EEMComplex	insqr, in, temp;
    double	absph, sr, si;

    absph = eemCplxPhase( pGrp->gain.r, pGrp->gain.i);
    in = *pIn;
    eemCplxSqr2( &insqr, &in);
    EEM_SILAR_LOOP_FORW_BEGIN(pGrp,pSilar) {
	if ( (power = pSilar->power) == 0 )	continue;
	sr = pSilar->loc.r;
	si = pSilar->loc.i;
	if ( TOP_IsOn(pSilar->flags,EEMSilarF_Conj) ) {
	    /* see eemSilarGrpEvalPart() */
	    temp.r = insqr.r - 2*sr*in.r + sr*sr + si*si;
	    temp.i = insqr.i - 2*sr*in.i;
	} else {
	    temp.r = in.r - sr;
	    temp.i = in.i - si;
	}
	absph += power * EEM_CPLX_PHASE(temp.r, temp.i);
    } EEM_SILAR_LOOP_FORW_END();
    return absph;
}

int
eemSilarGrpBB( EEMSilarGrp *pGrp, EEMComplex *pLo, EEMComplex *pHi) {
    int		cnt = 0;
    EEMComplex	lo, hi;

    EEM_CPLX_SET(&lo,0.0,0.0);
    EEM_CPLX_SET(&hi,0.0,0.0);
    EEM_SILAR_IDX_LOOP_FORW_BEGIN( pGrp, idx, pSilar) {
	if ( pSilar->power == 0 )
	    continue;
	if ( cnt++ == 0 ) {
	    lo = hi = pSilar->loc;
	    continue;
	}
	if ( pSilar->loc.r < lo.r )		lo.r = pSilar->loc.r;
	if ( pSilar->loc.r > hi.r )		hi.r = pSilar->loc.r;
	if ( pSilar->loc.i < lo.i )		lo.i = pSilar->loc.i;
	if ( pSilar->loc.i > hi.i )		hi.i = pSilar->loc.i;
	if ( EEM_SILAR_ConjB(pSilar) ) {
	    if ( -pSilar->loc.i < lo.i )	lo.i = -pSilar->loc.i;
	    if ( -pSilar->loc.i > hi.i )	hi.i = -pSilar->loc.i;
	}
    } EEM_SILAR_IDX_LOOP_FORW_END();
    *pLo = lo;
    *pHi = hi;
    return cnt;
}

double
eemSilarGrpBbMag( EEMComplex *pLo, EEMComplex *pHi, double minval) {
    double	l1 = TOP_ABS(pLo->r);
    double	l2 = TOP_ABS(pLo->i);
    double	h1 = TOP_ABS(pHi->r);
    double	h2 = TOP_ABS(pHi->i);
    double	m1, m2, m;

    m1 = TOP_MAX(l1,l2);
    m2 = TOP_MAX(h1,h2);
    m = TOP_MAX(m1,m2);
    return TOP_MAX(m,minval);
}

/**
  This function checks to see if a silar is in the silar group.
  A silar is considered to be in the silar group if the two silar.loc
  are the same.  It doesn't care about silar.power.
  If a silar is in the silar group, then the function stops the for
  loop and return the index which indicates where the silar is in the
  silar group.  If the silar is not in the silar group, then the 
  function returns -1
**/
int 
eemSilarGrpFindIdxByLoc(EEMSilarGrp *pSG, EEMComplex *pLoc) {
    EEM_SILAR_IDX_LOOP_FORW_BEGIN( pSG, idx, pSilar) {
        if ( eemCplxSameB( &pSilar->loc, pLoc) )
	    return idx;
    } EEM_SILAR_IDX_LOOP_FORW_END();
    return -1;
}

/**
	Inputs a location, this function finds the point in the silar
	group closest to the point pS2.  The function returns the index
**/
int 
eemCloseToNthSilarInSilarGrp(EEMSilarGrp *pSG, EEMComplex *pLoc) {
    int		minIdx = -1;
    EEMComplex	delta;
    double	distSq, minDistSq = 0;

    EEM_SILAR_IDX_LOOP_FORW_BEGIN( pSG, idx, pSilar) {
	if ( pSilar->power == 0 )
	    continue;
	delta.r = pSilar->loc.r - pLoc->r;
	delta.i = pSilar->loc.i - pLoc->i;
        distSq = EEM_CPLX_MAGSQ(&delta);
	if ( minIdx < 0 || distSq < minDistSq ) {
	    minIdx = idx; 
            minDistSq = distSq;
        }
    } EEM_SILAR_IDX_LOOP_FORW_END();
    return minIdx;
}

/**
  Checks to see of the two silars have the same location.
  It does not check to see if the multiplicities (power) of the silars
  are the same.
**/
TOPLogical 
eemSilarSameLocB(EEMSilar *pS1, EEMSilar *pS2) {
    return eemCplxSameB( &pS1->loc, &pS2->loc);
}

TOPLogical 
eemSilarSameB(EEMSilar *pS1, EEMSilar *pS2) {
    return pS1->power == pS2->power 
      && ((pS1->flags ^ pS2->flags) & EEMSilarF_Conj) == 0
      && eemCplxSameB( &pS1->loc, &pS2->loc);
}

void
eemSilarGrpInvertVar( EEMSilarGrp *pGrp) {
    EEMComplex	sp;
    double	spr;
    int		orgCnt;

    orgCnt = - eemSilarGrpExtractOrigin(pGrp);
    EEM_SILAR_LOOP_BEGIN( pGrp, pSilar) {
	if ( pSilar->power == 0 ) continue;
	if ( EEM_SILAR_ConjB(pSilar) ) {
	    spr = EEM_CPLX_MAGSQ(&pSilar->loc);
	    spr = pow( spr, (double) pSilar->power);
	    pGrp->gain.r *= spr; pGrp->gain.i *= spr;
	} else {
	    eemCplxPow3( &pSilar->loc, pSilar->power, &sp);
	    eemCplxMul2( &pGrp->gain, &sp);
	}
	if ( TOP_ABS(pSilar->power) & 1 ) {	/* abspow is odd */
	    eemCplxNeg1( &pGrp->gain);
	}
	orgCnt -= EEM_SILAR_TotPower(pSilar);
	eemCplxInv1( &pSilar->loc);
    } EEM_SILAR_LOOP_END();
    if ( orgCnt != 0 )
        eemSilarGrpAppend( pGrp, 0, 0.0, 0.0, orgCnt);
}
