proc tcpServer { {servername "Unknown"} {port -2323} } {
    set retry 0
    if { $port < 0 } {
	set retry 1
	set port [expr { - $port }]
	while 1 {
            if [catch {tcp server -port $port -command tcpLoginInit} server] {
		incr port
	    } else {
		break
	    }
	}
    } else {
        set server [tcp server -port $port -command tcpLoginInit]
    }
puts stdout "Bound server $server to port [$server conf -port]"
    upvar #0 top-$server svars
    set svars(servername) $servername
    $server start
    return $server
}

proc tcpLoginInit { client } {
    set server [$client server]
puts stdout "tcpLoginInit: $client of $server"

    upvar #0 top-$client cvars top-$server svars

    set cvars(state) start
    set cvars(servername) $svars(servername)
    set cvars(action) ""
    $client command "tcpLoginFsm"
    return [list $cvars(servername) "Send IDENT"]
}

proc tcpFsmError { client message } {
    upvar #0 top-$client cvars

    puts stderr "tcpLoginFsm: client $client, state $cvars(state), lastcomm $cvars(commword): $message."
    $client close
}

proc tcpLoginFsm { client command } {
    upvar #0 top-$client cvars

puts stdout "tcpLoginFsm: $client $command"
    if { ! [info exist cvars(state)] } {
	$client close
	puts stderr "tcpLoginFsm: $client: missing state"
    }
    set comword [lindex $command 0]
    set cvars(commword) $comword

    case $cvars(state) {
      start {
	case "$comword" {
	  IDENT {
	    set cvars(state) authwait
	    tcpLoginWindow $client $command
	    return WAIT
	  }
	  default {
	    tcpFsmError $client "No IDENT"
	  }
	}
      }
      authwait {
	# The server still hasnt decided what to do with the connection yet.
	# Just spin.
	case "$comword" {
	  POLL {
	    return WAIT
	  }
	  default {
	    tcpFsmError $client "Command before ready"
	  }
	}
      }
      authreply {
	# The server has decided what to do with the connection.  Inform
	# the client, and then take appropriate action.
	case "$comword" {
	  POLL {
	    case $cvars(action) {
	      accept {
	        set cvars(state) active
		set cvars(action) ""
	        return "ACCEPT"
	      }
	      refuse {
	        set cvars(state) closing
		set cvars(action) ""
	        return "REFUSE"
	      }
	      default {
	        tcpFsmError $client "Bad action"
	      }
	    }
	  }
	  default {
	    $client close
	    tcmFsmError $client "Command before ready"
	  }
        }
      }
      active {
	# This is the normal mode.  We just run the commands as given.
	eval $command
      }
      closing {
	# For whatever reason, we have decided closed this connection and will
	# not process commands on it.  The close may not have been issued yet,
	# so just error on the messages.
	$client close
	puts stderr "tcpLoginFsm: $client: command while closing"
	return [list CLOSED "Connection has been closed."]
      }
      default {
	$client close
	puts stderr "tcpLoginFsm: $client: bad state"
      }
    }
}

proc tcpLoginAction { client action } {
    upvar #0 top-$client cvars

    if { [info exist cvars(loginwindow)] } {
	topWin.SafeDestroy $cvars(loginwindow)
	set cvars(loginwindow) ""
    }

    set cvars(state) authreply
    set cvars(action) $action
}

proc tcpLoginWindow { client command } {
    upvar #0 top-$client cvars

    set suf [topWin.GenName .tcp-server]
    set w [lindex $suf 0]
    set idx [lindex $suf 1]
    set cvars(loginwindow) $w

    toplevel $w
    wm title $w "$cvars(servername) Request $idx"

    pack append $w [label $w.title \
      -text "Server $cvars(servername) Connection Request"] { top fillx }
    set from [lindex $command 1]
    pack append $w [label $w.from -text "$from"] { top fillx }

    set dd $w.buts; pack append $w [frame $dd] { top fillx }
    pack append $dd [button $dd.accept -text Accept \
      -com "tcpLoginAction $client accept"]  { left fill expand }
    pack append $dd [button $dd.refuse -text Refuse \
      -com "tcpLoginAction $client refuse"] { left fill expand }
}

proc tcpConn { clientname serverport serverhost } {
    set conn [tcp connect $serverhost $serverport]
puts stdout "Created tcp connection $conn"
    upvar #0 top-$conn nvars
    set nvars(state) start
    set nvars(clientname) $clientname

    $conn send IDENT $nvars(clientname)
    tcpConnBgPoll $conn
    return $conn
}

proc tcpConnError { conn message } {
    upvar #0 top-$conn nvars

    puts stderr "tcpConnError: conn $conn, state $nvars(state): $message"
    $conn close
}

proc tcpConnPollCB { conn } {
    upvar #0 top-$conn nvars

    if { ! [info exist nvars(state)] } {
	return
    }
    if [catch {$conn send POLL} result] {
	tcpConnError $conn "POLL error: ``$result''"
    }
    case [lindex $result 0] {
      WAIT {
	tcpConnBgPoll $conn
      }
      ACCEPT {
	set nvars(state) active
      }
      REFUSE {
	set nvars(state) closing
      }
      default {
	tcpConnError $conn "Bad POLL response: ``$result''"
      }
    }
}

proc tcpConnBgPoll { conn } {
    after 1000 tcpConnPollCB $conn
}

proc tcpConnCmd { conn command } {
    upvar #0 top-$conn nvars

    if { ! [info exist nvars(state)] } {
	error "$conn no longer exists"
    }
    case $nvars(state) {
      active {
	eval {$conn} send $command
      }
      default {
	error "$conn is not active"
      }
    }
}
