#
#
#

proc page.tcl { } { }

proc xpPage.Create { win { width 80 } { height 24 } } {
    upvar #0 $win wvars
    # text widet uses: "*-Courier-Medium-R-Normal-*-120-*"
    # canvas default : "-Adobe-Helvetica-Bold-R-Normal-*-120-*"
    set wvars(font)		"*-Courier-Bold-R-Normal-*-120-*"
    set wvars(normcolor)	"#000"
    set wvars(newcolor)		"#f00"
    set wvars(oldcolor)		"#c10"
    canvas $win -bd 2
    bind $win <Destroy> "topWin.Unset %W"
    set zeros [format "%0100d\n" 0]
    for {set i 0} {$i < 100} {incr i} {
	append bigzeros $zeros
    }
    set id [$win create text 0 0 -text $bigzeros -font $wvars(font) -anc nw]
    set bbox [$win bbox $id]
    set wvars(charwidth)	[expr {[lindex $bbox 2]/100}]
    set wvars(linespace)	[expr {[lindex $bbox 3]/100}]
#puts stdout "xpPage.Create $bbox cw $wvars(charwidth) ls $wvars(linespace)"
    $win delete $id
    $win conf \
      -width  [expr {$width * $wvars(charwidth)}] \
      -height [expr {$height * $wvars(linespace)}]
    return $win
}

proc xpPage.Destroy { win } {
    upvar #0 $win wvars
    destroy $win
    unset wvars
}

proc xpPage.Insert { win iteminfos {key ""} } {
    upvar #0 $win wvars
    set yloc [lindex [$win bbox all] 3]
    if { "$yloc" == "" } {
	set yloc 0
    }
    set xloc -1

    foreach iteminfo $iteminfos {
	set newx [expr { [lindex $iteminfo 1] * $wvars(charwidth) }]
	if { $newx <= $xloc } {
	    set yloc [expr { $yloc + $wvars(linespace) }]
	}
	set xloc $newx
        $win create text $xloc $yloc -text [lindex $iteminfo 2] \
	  -tags "key${key} ${key}[lindex $iteminfo 0]" \
	  -font $wvars(font) -anc nw
    }
    set wvars(changeset-last-${key})	"x"
    set wvars(oldvals-${key})		""
}

proc xpPage.ClearHilight { win {key ""} {sn ""} } {
    upvar #0 $win wvars

    set cs [expr { "$sn" == "" 
      ? "$wvars(changeset-last-${key})" : "changeset-${key}-${sn}" } ]
    $win itemconf $cs -fill $wvars(normcolor)
    $win dtag $cs
}

proc xpPage.Delete { win iteminfos {key ""} } {
    $win delete "key${key}"
}

proc xpPage.Update { win iteminfos itemvals {key ""} {sn ""} } {
    upvar #0 $win wvars
    set updateB 1

    if { "$key" != "" && "[$win gettags key${key}]" == "" } {
	# We want the insert to take place at "current" y pos, so kill
	# off everything not yet updated.
	xpPage.UpdateEnd $win 0
	xpPage.Insert $win $iteminfos $key
    }
    if { "$key"!="" } {
        lappend wvars(keys-cur) $key
    }
    if { "$sn"=="" } {
	# expr { [info exist wvars(sn-${key})] ? [incr wvars(sn-${key})] : [set wvars(sn-${key}) 1] }
	# set sn $wvars(sn-${key})
	expr { [info exist wvars(sn)] ? [incr wvars(sn)] : [set wvars(sn) 1] }
	set sn $wvars(sn)
    }

    #
    # Change everything in last update to oldcolor
    #
    set lastcs $wvars(changeset-last-${key})
    $win itemconf $lastcs -fill $wvars(oldcolor)

    #
    # Now update everything that has changed
    #
    set oldvals $wvars(oldvals-${key})
    if { "$oldvals" == "$itemvals" } {
	set wvars(changeset-last-${key}) x
    } else {
	set cs changeset-${key}-${sn}
	set wvars(changeset-last-${key}) $cs
	set idx 0
	foreach iteminfo $iteminfos {
	    if { "[lindex $iteminfo 2]"!="" } {
		continue
	    }
	    set newval [lindex $itemvals $idx]
	    if { $updateB && "[lindex $oldvals $idx]" == "$newval" } {
		incr idx
		continue
	    }
	    set loctag "${key}[lindex $iteminfo 0]"
	    # Must remove all previous changeset tags from the item
	    $win itemconf $loctag -text $newval -fill $wvars(newcolor) \
	      -tags "key${key} $loctag $cs"
	    # $win addtag $cs withtag $loctag
	    incr idx
	}
	set wvars(oldvals-${key}) $itemvals
	after 5000 xpPage.ClearHilight $win $key $sn
    }
}

proc xpPage.UpdateEnd { win {lastB 1} } {
    upvar #0 $win wvars

    if { ! [info exist wvars(keys-cur)] } {
	set wvars(keys-cur) ""
    }
    if { [info exist wvars(keys-prev)] } {
#puts stdout "Prev: $wvars(keys-prev)\nCur: $wvars(keys-cur)"
	foreach key $wvars(keys-prev) {
	    if { [lsearch $wvars(keys-cur) $key] < 0 } {
		$win delete key${key}
	    }
	}
    }
    if { $lastB } {
        set wvars(keys-prev) $wvars(keys-cur)
        set wvars(keys-cur) ""
    }
}
