#
# silargraph.tcl :: XPole
#
# This is a Tk script to implement the windows and user interface
# for XPole's "plane window".  This window allows input/output of
# a filter's singularities.  It uses the "xpplanegraph" widget
# implemented by "planeTk.c" as the core of its work...this simply
# provides the "fluff" around it.
#
# Most of code for this is implemented as part of the generic
# "xpSimplePlaneWindow".  This just provides the special case UI
# for XPole.
#

proc silargraph.tcl { } { }

proc xpSPG.InsSilarSet { graph viewspec } {
    upvar #0 $graph.draw pvars

    set viewinfo [view get $viewspec]
    set fltrinfo [filter info [lindex $viewinfo 0]]
    set fltrdom [lindex $fltrinfo 1]
    if { "$pvars(domain)"=="" } {
	set pvars(domain) $fltrdom
	$graph.draw conf -unitCircle [expr { "$pvars(domain)" == "z" } ]
    } else {
	if { "$pvars(domain)" != "$fltrdom" } {
	    error "Cant mix s/z domains"
	}
    }
    $graph.draw conf -datasrc xpole -dataspec $viewspec
    set pvars(defstatetext) [lindex $fltrinfo 2].[lindex $viewinfo 1]
    xpPlane.setState $graph.draw $pvars(state)
}

#
# This is menu service func.  It displays data dervied from this silar set.
#
proc _xpSPG.MainMenuCB { graph type } {
    upvar #0 $graph.draw pvars

    set viewspec [lindex [$graph.draw conf -dataspec] 4]
    set viewinfo [view info $viewspec]
    set viewspec [lindex $viewinfo 0].[lindex $viewinfo 1]

    case $type {
      {Impulse Fourier} {
        xpView.DisplayByClass $viewspec CplxSeq$type
      }
      StdDsgn {
	xpStdFltrDlgPopup $viewspec
      }
      Export {
	xpExportFltrDlgPopup $viewspec
      }
    }
}

#
# Encapsulation window.  The {w} passed in should be of class
# XPPlaneWindow for this to work.
#
proc xpSilarPlaneGraph { w } {
    upvar #0 $w.draw pvars
    set wd $w.draw
    set pvars(graph) $w
    xpPlane.InitVars $wd

    xpStdTitleBar $w.topbar "Pole-Zero Plane" "(xpole)PoleZero Win"
    pack append $w $w.topbar { top fillx }
    set menu $w.topbar.title.m; menu $menu
    xpPlane.addmenu $w $menu
    $menu add sep
    $menu add com -lab "Display Impulse" -com "_xpSPG.MainMenuCB $w Impulse"
    $menu add com -lab "Display Fourier" -com "_xpSPG.MainMenuCB $w Fourier"
    $menu add com -lab "Standard Design" -com "_xpSPG.MainMenuCB $w StdDsgn"
    $menu add com -lab "Export" -com "_xpSPG.MainMenuCB $w Export"
    $menu add com -lab "Zap Window" -acc "^z" -com "xpPlane.SafeDestroy $w"

    bind $menu <Control-z> [list $menu invoke "Zap Window"]

    set pvars(locfmt) "%.3E"
    xpPlane.MakeLocWin $w

    xpAxisCreate $w.lAxis -side right -flip true -label Imag
    pack append $w $w.lAxis {left filly}

    xpAxisCreate $w.bAxis -side top -label Real
    pack append $w $w.bAxis {bottom fillx}

    xpplanewidget $w.draw -xaxis $w.bAxis -yaxis $w.lAxis -notify xpPlane.notifyCB
    pack append $w $w.draw {top expand fill}

    $w.lAxis conf -datawindow $w.draw -command "$w.draw axisxf"
    $w.bAxis conf -datawindow $w.draw -command "$w.draw axisxf"
    xpAxis.AspectOn $w.lAxis $w.bAxis

    xpPlane.addbind $w $menu
    rubber $w.draw.rubber $w.draw -command "xpPlane.rubberFunc $w.draw"
    trace var pvars u "xpSPG.TraceDestroy $w"
    after 1 xpPlane.grow $w.draw
}
