#
# stdmenus.tcl
#

#option add *topbar.title.Font 9x15bold
option add *topbar.title.Font "*-bold-r-*-18-*-iso8859-1"

proc help { args } {
    if { "$args"=="" } {
	set node (xpole)Top
    } else {
        set node [eval concat $args]
    }
#    tkiInfoWindow $node
    tkiGetAndDisplay $node xpole
    return ""
}

#
# Create an instance of the XPole main menu called {w}.
# This menu is normally in the upper left corner of all XPole windows.
# The caller must arrange for a geometry manager for {w}.
#
proc xpStdMainMenu { w } {
    menubutton $w -menu $w.m -text "XPole"
    menu $w.m
    $w.m add com -label "About"		-command "help (xpole)about"
    $w.m add com -label "Help"		-command "help (xpole)"
    $w.m add com -label "Filters/Views..." -command "xpViewSelDlg"
    $w.m add com -label "Agents..."	-command "xpAgentDlg"
    $w.m add com -label "Std Filters..." -command "xpStdFltrDlgPopup"
    $w.m add com -label "Export..."	-command "xpExportFltrDlgPopup"
    $w.m add com -label "hierQuery"	-command "hq.setupquery [winfo name .]"
    $w.m add com -label "Quit"		-command "xpole quit"
    
    return $w
}

#
# Create an instance of the standard XPole title bar called {w}.
# This title bar normally appears at the top of all toplevel XPole windows.
# The title is actually a menu button with title {title}; the
# caller may create a menu for it.
# The caller must arrange for a geometry manager for {w}.
# 
proc xpStdTitleBar { w title helpTag } {
    upvar #0 $w wvars
    frame $w -relief raised
    xpStdMainMenu $w.xpole
    menubutton $w.title -text $title -menu $w.title.m

    button $w.help -text "Help" -command "help $helpTag"

    pack append $w $w.xpole { left frame w }
    pack append $w $w.help  { right frame e }
    pack append $w $w.title { top expand }

    # Setup <Help> bindings.  We do the actual bindings here
    # (in addtion to the global binding) in-case of key-propagation mode.
    set wvars(infonode) $helpTag
    bind $w <Key-Help> "tkiContextHelp %W xpole"
    set pntw [winfo toplevel $w]
    if { "$pntw"!="" && "$pntw"!="$w" } {
	upvar #0 $pntw pntwvars
	if ![info exist pntwvars(infonode)] {
            set pntwvars(infonode) $helpTag
     	    bind $w <Key-Help> "tkiContextHelp %W xpole"
	}
    }

    return $w
}

#
# Create a toplevel window.  {dlg} specifies the path name, and
# {win_title} is the passed to the window manager.  Either or both
# {dlg} and {win_title} may be empty.  {menu_title} is the label of
# the pull-down menu in the top center of the window.  The caller
# must create a menu "$dlg.topbar.title.m".
# A help button will be created in the top right corner; it will bring
# up help text using {helpTag}.
#
proc xpStdDlgPopup { dlg win_title menu_title helpTag } {
    if { "$win_title"=="" } {
        set win_title $menu_title
    }
    if { "$dlg"=="" } {
	set dlg $win_title
	regsub -all ".\t " $dlg {-} dlg
	set dlg ".$dlg"
    }
    set suf [topWin.GenName $dlg]
    set dlg [lindex $suf 0]
    set num [lindex $suf 1]
    topWin.GetTopLevel $dlg
    wm minsize $dlg 20 20
    wm title    $dlg "XPole $win_title $num"
    wm iconname $dlg "XPole $win_title $num"
    pack append $dlg [xpStdTitleBar $dlg.topbar $menu_title $helpTag] \
      { top fillx }
    return $dlg
}

#############################################################################
# 
#	Standard Error Handling (user interface)
#
# Errors may be displayed in one of two ways:
#    popup	Each time an error occurs, a popup window is created to
#		contain the error message.
#    wdg	The application has a built-in status/error window
#		that can be used to display the error message.  The details
#		of the error message are stilled displayed in a popup window.
#
#############################################################################

proc xpStd.DlgErrClear { key } {
    upvar #0 $key fvars

    if { [info exist fvars(error-wdg)] } {
	$fvars(error-wdg).label conf -text "Status:"
	$fvars(error-wdg).msg conf -text $fvars(error-msg)
	pack unpack $fvars(error-wdg).more
    }
    set fvars(error-detail) "No Error Recorded"
    set fvars(error-activeB) 0
}

proc xpStd.DlgErrSetStatus { key {msg "Ok"} } {
    upvar #0 $key fvars

    if { [info exist fvars(error-msg)] } {
	set fvars(error-msg) $msg
	if { ! $fvars(error-activeB) } {
	    xpStd.DlgErrClear $key
	}
    }
}

proc xpStd.DlgErrMore { key {sn ""} } {
    upvar #0 $key fvars

    if { "$sn"=="" && [info exist fvars(error-lastsn)] } {
	set sn $fvars(error-lastsn)
    }
    case $fvars(error-config) {
      popup {
	set w $fvars(error-window-$sn)
	text $w.detail
	$w.detail insert end $fvars(error-detail-$sn)
	$w.detail conf -state disabled
	pack before $w.buts $w.detail { top fill }
	pack unpack $w.buts.more
      }
      wdg {
	set w .w[topGenSN]
	toplevel $w
	wm minsize $w 20 20
	set dd $w.detail; pack append $w [frame $dd] {top fill expand}
	pack append $dd [text $dd.text] {left fill expand}
	pack append $dd [scrollbar $dd.sy -command "$dd.text yview"] {left fill}
	$dd.text insert end $fvars(error-detail-$sn)
	$dd.text conf -state disabled -yscroll "$dd.sy set"
	pack append $w [button $w.dismiss -text "Dismiss" \
	  -com "topWin.IdleDestroy $w"] {top fillx}
	xpStd.DlgErrClear $key
      }
    }
}

proc xpStd.DlgErrCleanup { key sn w } {
    catch {unset fvars(error-error-$sn)}
    catch {unset fvars(error-context-$sn)}
    catch {unset fvars(error-detail-$sn)}
    topWin.IdleDestroy $w
}


#
# Display an error message to the user.  {key} may be any abritrary text
# (or empty); it is used for more sophisticated dialog systems where
# there is a status/error field.
#
proc xpStd.DlgErr { key context error {detailErr ""} {type "Error"} } {
    if { "$key"=="" } {
	set key errs
    }
    upvar #0 $key fvars

    if { ! [info exist fvars(error-config)] } {
	set fvars(error-config) popup
    }
    if { ! [info exist fvars(error-sn)] } {
	set fvars(error-sn) 1
    }
    set sn $fvars(error-sn)
    set fvars(error-error-$sn) $error
    set fvars(error-context-$sn) $context
    set fvars(error-detail-$sn) $detailErr

    case $fvars(error-config) {
      popup {
        incr fvars(error-sn)
	set ddinfo [topWin.GenName .error]
	set dd [lindex $ddinfo 0]
	set wintitle "XPole $type [lindex $ddinfo 1]"
	toplevel $dd
	wm title $dd $wintitle
	wm iconname $dd $wintitle
	set fvars(error-window-$sn) $dd

	pack append $dd [text $dd.m -width [string length $error] -height 1] \
	  { top fill expand }
	$dd.m insert end $error
	$dd.m conf -state disabled
	pack append $dd [frame $dd.buts] { top fillx }
	pack append $dd.buts [button $dd.buts.dismiss -text "Dismiss" \
	  -com "topWin.IdleDestroy $dd"] { left fill expand }
	pack append $dd.buts [button $dd.buts.more -text "More" \
	  -com "xpStd.DlgErrMore $key $sn"] { left fill expand }
	bind $dd <Destroy> "xpStd.DlgErrCleanup $key $sn $dd"
      }
      wdg {
	$fvars(error-wdg).label conf -text "$type:"
	$fvars(error-wdg).msg conf -text $error
	if { "$detailErr"!="" } {
	    pack append $fvars(error-wdg) $fvars(error-wdg).more { left }
	}
	set fvars(error-activeB) 1
	set fvars(error-lastsn) $sn
      }
   }
}

#
# Create a "wdg" style status/error message window (e.g., a built-in display,
# not a popup).  {key} must be non-null and unique, {w} must be the
# name of a window which doesn't yet exist but whose parent does.
# {msg} is the default message to display where there are no errors.
#
proc xpStd.DlgErrCreate { key w {msg "Ok"} } {
    upvar #0 $key fvars
    set fvars(error-config) wdg
    set fvars(error-wdg) $w
    set fvars(error-msg) $msg
    frame $w
    pack append $w [label $w.label -text "Status:" -width 8] { left }
    pack append $w [button $w.msg -rel flat -anc w \
      -text "" -com "xpStd.DlgErrClear $key"] { left expand fill }
    button $w.more -text "More" -com "xpStd.DlgErrMore $key"
    $w.msg invoke
    return $w
}
