#
# tkpole startup script
#
puts stdout "Welcome to XPole ([xpole version]).  Initializing ..."

proc tkerror { message } {
    global errorInfo
    puts stdout [format "TKError: %s" $message]
    puts stdout $errorInfo
}

proc toperror { message } {
    puts stdout [format "TOPError: %s" $message]
}


proc xpole_init_env { } {
    global tcl_library tk_library auto_path app_library

    global PI; set PI 3.14159265358979323846

    #
    # Source the standard tcl & tk scripts.  We MUST do this first because
    # tk.tcl wipes out the auto_load path.  Also, both scripts assume
    # they are being run at the global-level, so we uplevel them.
    #
    if { ! [info exist tcl_library] } { set tcl_library [info library] }
    uplevel #0 source $tcl_library/init.tcl
    uplevel #0 source $tk_library/tk.tcl

    global tkiEmbed; set tkiEmbed 1
    global hqEmbed; set hqEmbed 1
    if { [file exist "tkpole.c"] } {
	set app_library ./lib
	lappend auto_path $app_library ../tkgraph/lib
        tkiAddInfoPaths ./doc/user
    } else {
	lappend auto_path $app_library
        tkiAddInfoPaths $app_library
    }

#    if [info exists geometry] {
#	wm geometry . $geometry
#    }
}

xpole_init_env

proc xpole_init_packages {} {
    topmisc.tcl
    xpdefvals.tcl
    xpaxis.tcl
    xpplanegraph.tcl
    xpseqgraph.tcl

    viewsel.tcl
    silargraph.tcl
}

proc xpole_init_session { } {
    global argv xpole

    set sFlag		0
    set zFlag		0
    set fltrList	""
    set sfList		""
    set cmdList		""
    set sendB		0
    set whiteB		0
    set blackB		0

    set optlist { 
      {s sFlag bool} {z zFlag bool} 
      {filter fltrList append} {sf sfList append}
      {send sendB bool}
      {command cmdList append}
      {white whiteB bool} {black blackB bool}
    }
    set rest_args [topgetopt $optlist $argv]

    # This fixed a bug in Ultrix's Borne shell's "$@" notation
    if { [llength $rest_args]>0 && "[lindex $rest_args 0]"=="" } {
	set rest_args [lreplace $rest_args 0 0]
    }
    if { "$rest_args"!="" } {
	error "Extra command line arguments not allowed: ``$rest_args''"
    }
    
    # By default (or with "+send" option), kill off the send command
    # This also unregisters us, so that we wont respond to send commands
    # from other apps.
    if { ! $sendB } {
        rename send ""
	proc send { interp args } {
	    if { "$interp"!="[winfo name .]" } {
		error "The send command is not enabled: use -send when starting XPole."
	    } 
	    eval $args
	}
    }

    if { $blackB } {
	global TopColorScreenB TopColorReverseB
	set TopColorScreenB 0
	set TopColorReverseB 0
    }
    if { $whiteB } {
	global TopColorScreenB TopColorReverseB
	set TopColorScreenB 0
	set TopColorReverseB 1
    }

    if { $sFlag } { set xpole(xfdomain) "s" }
    if { $zFlag } { set xpole(xfdomain) "z" }

    if { "$fltrList"=="" && "$sfList"=="" } {
	set fltrList H1
	set sfList H1
    }
    foreach f $fltrList {
        filter create $f
    }

    xpole_init_packages

    foreach sf $sfList {
        xpView.DisplayByClass $sf SilarRep
        xpView.DisplayByClass $sf CplxSeqFreq
    }

    # filter create H2(s)
    # xpStdFltrDlgPopup H1.polezero
    # xpView.DisplayByClass H2 SilarRep
    # xpExportFltrDlgPopup H1.polezero
    xpView.DisplayByClass h1.polezero CplxSeqTime
#    xpView.DisplayByClass h1.polezero-impulse CplxSeqFourier

    foreach c $cmdList {
	eval $c
    }

    # We do global "all" binding here, but also do a binding for each
    # main window for key-propgation mode
    bind all <Key-Help> "tkiContextHelp %W xpole"
    bind all <Key-F9> "hq.configinfowin [winfo name .] %W"
}
xpole_init_session

puts stdout "... Initialization done."
