/* 
    Author: Kennard White

    Copyright (c) 1990-93 The Regents of the University of California.
    All rights reserved.  See the file "./copyright.h" for full copyright
    notice, limitation of liability, and disclaimer of warranty provisions.
#include "copyright.h"				COPYRIGHTENDKEY
*/
#if !defined(lint) && !defined(SABER)
static char rcsid[] = "$Header: /vol/cooley/cooley1/kennard/src/xpole/RCS/pipe.c,v 1.5 93/08/27 14:28:18 kennard Exp $";
#endif

#include "topFixup.h"
#include <tcl.h>

#include "topStd.h"
#include "eemDft.h"
#include "eemPfe.h"

#include "pole.h"
#include "filter.h"
#include "cplxseq.h"
#include "silar.h"
#include "pipe.h"

char _XpPipeSubPackage[] = "xp.pipe";
#define SPKG _XpPipeSubPackage


/******************************************************************************
 *
 *				Polezero Impulse
 *
 *	Compute the Impulse response from the singularity set.  We PFE
 *	the H(s/z), and then inverse Laplace/Z each term and evaluate.
 *
 *****************************************************************************/

static void
_xpPipeSilarImpulseCalcZ( XPSilarSnap *pSR, EEMPfeGrp *pPfe, XPCplxSeqSnap *pSeq) {
    double	T = pSR->PntView->DM.SampleTime;
    double	Tinv = 1.0/T;
    int		thisidx, lastidx = -1;

    eemPolyCplxSetNum( &pSeq->valRI, XP_CPLXSEQ_NumIndex(pSeq));
    XPCPLXSEQ_I_IDX_LOOP_FORW_BEGIN( pSeq, ptIdx, pIdxVal) {
	EEMComplex *pRIval = &EEM_POLYCPLX_Get(&(pSeq)->valRI,ptIdx);
	EEM_CPLX_ZERO( pRIval);
	thisidx = (int) floor(*pIdxVal * Tinv);
	if ( ptIdx!=0 && thisidx==lastidx )	/* always do first pt */
	    continue;
	lastidx = thisidx;
	eemPfeGrpToTimeZ( pPfe, thisidx, pRIval);
    } XPCPLXSEQ_I_IDX_LOOP_FORW_END();

}

static void
_xpPipeSilarImpulseCalcS( XPSilarSnap *pSR, EEMPfeGrp *pPfe, XPCplxSeqSnap *pSeq) {
    eemPolyCplxSetNum( &pSeq->valRI, XP_CPLXSEQ_NumIndex(pSeq));
    XPCPLXSEQ_I_IDX_LOOP_FORW_BEGIN( pSeq, ptIdx, pIdxVal) {
	EEMComplex *pRIval = &EEM_POLYCPLX_Get(&(pSeq)->valRI,ptIdx);
	eemPfeGrpToTimeS( pPfe, *pIdxVal, pRIval);
    } XPCPLXSEQ_I_IDX_LOOP_FORW_END();

}

static void
_xpPipeSilarImpulseCalc( XPSilarSnap *pSR, EEMPfeGrp *pPfe, XPCplxSeqSnap *pSeq) {
    if ( pSR->PntView->DM.XfDomain == XP_XFDOM_S ) {
	_xpPipeSilarImpulseCalcS( pSR, pPfe, pSeq);
    } else {
	_xpPipeSilarImpulseCalcZ( pSR, pPfe, pSeq);
    }
    pSeq->ValidMask = XPCSeqVals_RI 
      | (pSeq->ValidMask & (XPCSeqVals_Index | XPCSeqVals_FreqLoc));
}

typedef struct _XPPipeSilarImpulse XPPipeSilarImpulse;
struct _XPPipeSilarImpulse {
    EEMPfeGrp		pfe;
};

static void
_xpPipeSilarImpulseNotify( XPPipeMod *pPipe, XPViewEvent ev,
  XPViewSnap *pSrcSnap, XPAgent *pTgtAg) {
    XPSilarSnap		*pSR = (XPSilarSnap*) pSrcSnap;
    XPPipeSilarImpulse	*pPipeExt = (XPPipeSilarImpulse*) pPipe->cSide.Ext;
    double		T = pSR->PntView->DM.SampleTime;

    switch ( ev ) {
    case XPVE_Create:
	pPipeExt = MEM_ALLOC(XPPipeSilarImpulse);
	pPipe->cSide.Ext = (TOPPtr) pPipeExt;
	eemPfeGrpConstructor( &pPipeExt->pfe);
	pPipe->Flags |= XPPipeF_Update1;
	break;
    case XPVE_Destroy:
	eemPfeGrpDestructor( &pPipeExt->pfe);
	MEM_FREE(pPipeExt);
	break;
    case XPVE_Modify:
	if ( TOP_IsOff(pSR->ChangeMask,EEMSilarC_Value) )
	    break;
    case XPVE_Switch:
	pPipe->Flags |= XPPipeF_Update1;
	XPVIEWPART_LOOP_FORW_BEGIN( pTgtAg->CurView, pVP) {
	    XPCplxSeqSnap *pSeq = (XPCplxSeqSnap*)
	      xpViewPartGetSnap( pVP, XPGS_AsIs);
	    pVP->Flags &= ~XPViewF_Valid;
    	    pSeq->DomStep = pSR->PntView->DM.SampleTime;
    	    if ( pSR->PntView->DM.XfDomain == XP_XFDOM_S ) {
		double r = eemSilarGrpBbMag( &pSR->bbLo, &pSR->bbHi, 0.0);
		if ( r == 0.0 )
		    r = 2;
		pSeq->SigDomMin = 0.0;
		pSeq->SigDomMax =  1/r * 5.0;	/* XXX: 5 is abitrary */
	    } else {
    		pSeq->SigDomMin = -10 * T;
    		pSeq->SigDomMax = 40 * T;
	    }
	    xpViewPartNotify( pVP, XPNW_Now, XPVE_Modify, pTgtAg);
	} XPVIEWPART_LOOP_FORW_END();
    default: ;
    }
}

static TOPLogical
_xpPipeSilarImpulseUpdate( XPPipeMod *pPipe,
  XPViewSnap *pSrcSnap, XPViewSnap *pTgtSnap) {
    XPPipeSilarImpulse	*pPipeExt = (XPPipeSilarImpulse*) pPipe->cSide.Ext;
    XPSilarSnap		*pSR = (XPSilarSnap*) pSrcSnap;
    XPCplxSeqSnap	*pSeq = (XPCplxSeqSnap*) pTgtSnap;

    if ( TOP_IsOn(pPipe->Flags,XPPipeF_Update1) ) {
	eemSilarGrpToPfeGrp( XP_SILSET_Grp(pSR), &pPipeExt->pfe,
	  pSR->PntView->DM.XfDomain == XP_XFDOM_S ? 0 : EEM_PFE_ExtractOrigin);
	pPipe->Flags &= ~XPPipeF_Update1;
    }
    _xpPipeSilarImpulseCalc( pSR, &pPipeExt->pfe, pSeq);
    return TOP_TRUE;
}

static XPPipeClass _xpPipeSilarImpulseCls = {
    "PolezeroImpulse", XPPipeF_ProdParams,
    { "SilarRep", XPPipeF_MultiPart, 
      _xpPipeSilarImpulseNotify, _xpPipeSilarImpulseUpdate },
    { "CplxSeqTime", 0, NULL, NULL }
};

/******************************************************************************
 *
 *				Silar Fourier
 *
 *	Compute the fourier transform from the singularity set.  We do
 *	this by direct evaluation of H(s/z).  The routine below to calculate
 *	the foureir points could be greatly optimized for the equally spaced
 *	case (somewhat like FFT is optimized form of DFT).
 *
 *****************************************************************************/

/**
    Calc fourier pts: s-domain jw or z-domain ejwT, where w is the seq
    index value.  These points are used as s/z in evaluating H(s/z).
    Note that to calculate z-transform directly from impulse, the
    reciprocol of these pointes are required: H(z) = sum_k{ h(k) z^-k }.
**/
static void _xpPipeCalcFourierPt( XPCplxSeqSnap *pSeq,
  double w, EEMComplex *pFL) {
    double	T = pSeq->PntView->DM.SampleTime;

    if ( pSeq->PntView->DM.XfDomain == XP_XFDOM_S ) {
	EEM_CPLX_SET( pFL, 0, w);
     } else {
	sincos( w * T, &pFL->i, &pFL->r);
    }
}

/**
    {dir}, the direction, should be +1 or -1.
**/
static void _xpPipeCalcFourierPts( XPCplxSeqSnap *pSeq, int dir) {
    double	T = pSeq->PntView->DM.SampleTime * dir;
    eemPolyCplxSetNum( &pSeq->valFreqLoc, XP_CPLXSEQ_NumIndex(pSeq));
    if ( pSeq->PntView->DM.XfDomain == XP_XFDOM_S ) {
        XPCPLXSEQ_I_IDX_LOOP_FORW_BEGIN( pSeq, ptIdx, pIdxVal) {
	    EEMComplex *pFLval = &EEM_POLYCPLX_Get(&(pSeq)->valFreqLoc,ptIdx);
	    EEM_CPLX_SET( pFLval, 0, *pIdxVal);
        } XPCPLXSEQ_I_IDX_LOOP_FORW_END();
     } else {
        XPCPLXSEQ_I_IDX_LOOP_FORW_BEGIN( pSeq, ptIdx, pIdxVal) {
	    EEMComplex *pFLval = &EEM_POLYCPLX_Get(&(pSeq)->valFreqLoc,ptIdx);
	    sincos( *pIdxVal * T, &pFLval->i, &pFLval->r);
        } XPCPLXSEQ_I_IDX_LOOP_FORW_END();
    }
    pSeq->ValidMask = XPCSeqVals_FreqLoc | 
      (pSeq->ValidMask & XPCSeqVals_Index);
}



static void	/* Fourier */
_xpPipeSilarFourierNotify( XPPipeMod *pPipe, XPViewEvent ev,
  XPViewSnap *pSrcSnap, XPAgent *pTgtAg) {
    XPSilarSnap		*pSR = (XPSilarSnap*) pSrcSnap;

    switch ( ev ) {
    case XPVE_Modify:
	if ( TOP_IsOff(pSR->ChangeMask,EEMSilarC_Value) )
	    break;
    case XPVE_Switch:
	XPVIEWPART_LOOP_FORW_BEGIN( pTgtAg->CurView, pVP) {
	    XPCplxSeqSnap *pSeq = (XPCplxSeqSnap*)
	      xpViewPartGetSnap( pVP, XPGS_AsIs);
	    if ( pSR->PntView->DM.XfDomain == XP_XFDOM_S ) {
		double r = eemSilarGrpBbMag( &pSR->bbLo, &pSR->bbHi, 0.0);
		if ( r == 0.0 )
		    r = 2;
		pSeq->SigDomMin = - r * 2;
		pSeq->SigDomMax =   r * 2;
	    } else {
		pSeq->SigDomMin = -M_PI;
		pSeq->SigDomMax =  M_PI;
	    }
	    pVP->Flags &= ~XPViewF_Valid;
	    xpViewPartNotify( pVP, XPNW_Now, XPVE_Modify, pTgtAg);
	} XPVIEWPART_LOOP_FORW_END();
	break;
    default:
	;
    }
}

static double
_xpPipeSilarFourierAbsPhaseCB( XPCplxSeqSnap *pSeq, int idx, TOPPtr cld) {
    XPSilarSnap		*pSR = (XPSilarSnap*) cld;
    return eemSilarGrpEvalAbsPhase( XP_SILSET_Grp(pSR), 
      &EEM_POLYCPLX_Get(&pSeq->valFreqLoc,idx));
}

static double
_xpPipeSilarFourierPhaseCB( XPCplxSeqSnap *pSeq, double freq, TOPPtr cld) {
    XPSilarSnap		*pSR = (XPSilarSnap*) cld;
    EEMComplex	fl;
    _xpPipeCalcFourierPt( pSeq, freq, &fl);
    return eemSilarGrpEvalAbsPhase( XP_SILSET_Grp(pSR), &fl);
}

static TOPLogical
_xpPipeSilarFourierUpdate( XPPipeMod *pPipe,
  XPViewSnap *pSrcSnap, XPViewSnap *pTgtSnap) {
    XPSilarSnap		*pSR = (XPSilarSnap*) pSrcSnap;
    XPCplxSeqSnap	*pSeq = (XPCplxSeqSnap*) pTgtSnap;
    /*double		T = pSR->PntView->DM.SampleTime;*/

    if ( TOP_IsOff(pSeq->ValidMask,XPCSeqVals_FreqLoc) )
	_xpPipeCalcFourierPts( pSeq, 1);

    eemPolyCplxSetNum( &pSeq->valRI, EEM_POLYCPLX_Num(&pSeq->valFreqLoc));
    EEM_POLYCPLX_IDX_LOOP_FORW_BEGIN( &pSeq->valFreqLoc, ptIdx, pFLval) {
	EEMComplex *pRIval = &EEM_POLYCPLX_Get(&pSeq->valRI,ptIdx);
	xpSilarSnapFreqResp( pSR, pFLval, pRIval);
    } EEM_POLYCPLX_IDX_LOOP_FORW_END();
    pSeq->ValidMask = XPCSeqVals_RI 
      | (pSeq->ValidMask & (XPCSeqVals_Index | XPCSeqVals_FreqLoc));
    pSeq->PntView->PhaseFunc = _xpPipeSilarFourierPhaseCB;
    pSeq->PntView->PhaseCld = (TOPPtr) pSR;
    pSeq->PntView->AbsPhaseFunc = _xpPipeSilarFourierAbsPhaseCB;
    pSeq->PntView->AbsPhaseCld = (TOPPtr) pSR;
    return TOP_TRUE;
}

static XPPipeClass _xpPipeSilarFourierCls = {
    "PolezeroFourier", XPPipeF_ProdParams,
    { "SilarRep", XPPipeF_MultiPart,
      _xpPipeSilarFourierNotify, _xpPipeSilarFourierUpdate },
    { "CplxSeqFreq", 0, NULL, NULL },
};

/******************************************************************************
 *
 *			Impulse Fourier
 *
 *   Compute the Fourier response given the Impulse response.  In both
 *   cases we are dealing with samples only.  For FIR, we can do this
 *   exactly, but for IIR we must window first.  Then given that
 *   finite set of samples, we must compute either the DFT or FFT.
 *   For now we always use the DFT becuase its simpler.  In future,
 *   use FFT when samples are linearly spaced.
 *
 *****************************************************************************/

typedef struct _XPPipeImpulseFourier XPPipeImpulseFourier;
struct _XPPipeImpulseFourier {
    int		winSize;
    TOPVector	winVec;		/* array of doubles */
};

static void
_xpPipeImpulseFourierNotify( XPPipeMod *pPipe, XPViewEvent ev,
  XPViewSnap *pSrcSnap, XPAgent *pTgtAg) {
    Tcl_Interp		*ip = XPTheMachine.ip;
    XPCplxSeqSnap	*pImpulse = (XPCplxSeqSnap*) pSrcSnap;
    char		*winSizeStr = NULL, *winTypeStr = NULL;
    int			winSizeNew;
    EEMWindowType	winType;
    double		T = pImpulse->PntView->DM.SampleTime;
    XPAgent		*pSrcAg = pTgtAg == &pPipe->pSide.Ag ? &pPipe->cSide.Ag
			  : &pPipe->pSide.Ag;
    XPPipeImpulseFourier *pExt = (XPPipeImpulseFourier*) pPipe->cSide.Ext;
    char		*n1 = pPipe->Lbl.Name;
    char		buf[100];

    switch ( ev ) {
    case XPVE_Create:
	pExt = MEM_ALLOC(XPPipeImpulseFourier);
	pExt->winSize = 256;
	sprintf(buf,"%d",pExt->winSize);
	Tcl_SetVar2(ip,n1,"windowsize",buf, TCL_GLOBAL_ONLY);
	topVecConstructor( &pExt->winVec, sizeof(double));
	Tcl_SetVar2(ip,n1,"windowtype","rectangle", TCL_GLOBAL_ONLY);
	pPipe->cSide.Ext = (TOPPtr) pExt;
	/*FALLTHROUGH*/
    case XPVE_Param:
	winSizeStr = Tcl_GetVar2(ip, n1, "windowsize", TCL_GLOBAL_ONLY);
	if ( winSizeStr != NULL ) {
	    if ( Tcl_GetInt(ip, winSizeStr, &winSizeNew) == TCL_OK 
	      && winSizeNew > 0 )
		pExt->winSize = winSizeNew;
	}
	if ( pSrcSnap->PntVP->RangeOwner == pSrcAg ) {
	    xpCplxSeqSetIndexLin( pImpulse, pExt->winSize, 0.0, T);
	}
	winType = EEMWinType_Rectangle;
	winTypeStr = Tcl_GetVar2(ip, n1, "windowtype", TCL_GLOBAL_ONLY);
	if ( winTypeStr != NULL 
	  && (winType = eemWindowNameToType( winTypeStr)) != EEMWinType_Null ) {
	    topVecSetNum( &pExt->winVec, pExt->winSize);
	    if ( ! eemWindowCoefVec( winType, &pExt->winVec) )
		topVecSetNum( &pExt->winVec, 0);
	} else {
	    topVecSetNum( &pExt->winVec, 0);
	}
	/*FALLTHROUGH*/
    case XPVE_Switch:
    case XPVE_Modify:
	XPVIEWPART_LOOP_FORW_BEGIN( pTgtAg->CurView, pTgtVP) {
	    XPCplxSeqSnap *pSeq = (XPCplxSeqSnap*)
	      xpViewPartGetSnap( pTgtVP, XPGS_AsIs);
	    pSeq->SigDomMin = -M_PI;
	    pSeq->SigDomMax =  M_PI;
	    pTgtVP->Flags &= ~XPViewF_Valid;
	    xpViewPartNotify( pTgtVP, XPNW_Now, XPVE_Modify, pTgtAg);
	} XPVIEWPART_LOOP_FORW_END();
    default: ;
    }
}


static TOPLogical
_xpPipeImpulseFourierUpdate( XPPipeMod *pPipe,
  XPViewSnap *pSrcSnap, XPViewSnap *pTgtSnap) {
    XPCplxSeqSnap	*pImpulse = (XPCplxSeqSnap*) pSrcSnap;
    XPCplxSeqSnap	*pFourier = (XPCplxSeqSnap*) pTgtSnap;
    EEMComplex		*pFL;
    double		T = pImpulse->PntView->DM.SampleTime;
    int			srcN;
    XPPipeImpulseFourier *pExt = (XPPipeImpulseFourier*) pPipe->cSide.Ext;
    EEMPolyCplx		imp;

    xpCplxSeqCalc( pImpulse, XPCSeqVals_RI);
    srcN = EEM_POLYCPLX_Num(&pImpulse->valRI);
    if ( TOP_IsOff(pImpulse->Flags,XPCSeqF_Discrete)
      || srcN == 0 || pImpulse->DomOfs != 0.0 || pImpulse->DomStep != T ) {
	topWarnMsg("%s %s: invalid impulse",SPKG,pPipe->Lbl.Tag);
	return TOP_FALSE;
    }
    /*
     * At this point we can be assured that Impulse is linear,
     * 0-origin and SampleTime spaced.
     */
    eemPolyCplxConstructor( &imp);
    eemPolyCplxDotVec3( &pImpulse->valRI, &pExt->winVec, &imp);

    /* How we compute the Fourier depends on the scale of the Fourier.  */
#ifdef notdef
    if ( pFourier->domStep > 0 ) {
	/* Yeah: can use DFT/FFT */
    } else
#endif
    {
	/* Nope: must compute DTFT at each point of interest */
	if ( TOP_IsOff(pFourier->ValidMask,XPCSeqVals_FreqLoc) )
	    _xpPipeCalcFourierPts( pFourier, -1);
	eemPolyCplxSetNum( &pFourier->valRI, XP_CPLXSEQ_NumIndex(pFourier));
	EEM_POLYCPLX_IDX_LOOP_FORW_BEGIN( &pFourier->valRI, idx, pRI) {
	    pFL = &EEM_POLYCPLX_Get( &pFourier->valFreqLoc, idx);
	    eemPolyCplxEval( &imp, pFL, pRI);
	} EEM_POLYCPLX_IDX_LOOP_FORW_END();
    }
    pFourier->ValidMask = XPCSeqVals_RI 
      | (pFourier->ValidMask & (XPCSeqVals_Index | XPCSeqVals_FreqLoc));
    eemPolyCplxDestructor( &imp);
    return TOP_TRUE;
}

static XPPipeClass _xpPipeImpulseFourierCls = {
    "ImpulseFourier", XPPipeF_ProdParams,
    { "CplxSeqTime", XPPipeF_MultiPart,
      _xpPipeImpulseFourierNotify, _xpPipeImpulseFourierUpdate },
    { "CplxSeqFreq", 0, NULL, NULL },
};




/*****************************************************************************
 *****************************************************************************/

#ifdef notdef

/**
    Given finite samples of a frequency response H'(w), compute
    an FIR h(n).
**/
static void
_xpPipeFsampImpulseNotify( XPPipeMod *pPipe, TOPLogical pmB,
  XPViewEvent t, XPViewSnap *pSrcSnap, XPViewSnap *pTgtSnap) {
    XPCplxSeqSnap	*pFsamp = (XPCplxSeqSnap*) pSrcSnap;
    XPCplxSeqSnap	*pImpulse = (XPCplxSeqSnap*) pTgtSnap;
    int			nSamps = 256;
    int			winSize = 256;
    double		T = pImpulse->PntView->DM.SampleTime;

    switch ( t ) {
    case XPVE_Create:
        pFsamp->Flags = XPCSeqF_Master | XPCSeqF_Discrete;
        xpCplxSeqSetIndexLin( pImpulse, winSize, 0.0,
	  M_PI/T/nSamps);
	eemPolyCplxSetNum( &pFsamp->valRI, winSize);
	pFsamp->ValidMask = XPCSeqVals_Index | XPCSeqVals_R;
	/*FALLTHROUGH*/
    case XPVE_Switch:
	/*FALLTHROUGH*/
    case XPVE_Modify:
        pImpulse->Flags = XPCSeqF_Master | XPCSeqF_Discrete;
        xpCplxSeqSetIndexLin( pImpulse, winSize, 0.0, T);
	eemDftCplxVec( -1, &pFsamp->valRI.Coef, &pImpulse->valRI.Coef);
	pImpulse->ValidMask |= XPCSeqVals_RI;
	xpAgentNotifyFrom( &pPipe->PM, XPNotifySN_Unspec, XPVE_Modify);
	break;
    default:
	;
    }
}

#endif

/******************************************************************************
 *
 *				Silar Map
 *
 *	Compute various silar-to-silar mappings.  Currently, there is
 *	an problem in that the target silar notify is sent out, but
 *	the silars havnt been updated and thus the bb and other stats
 *	of the tgt silar is not known.
 *
 *****************************************************************************/

/* s --> z */
static void
_xpPipeSilarBilinearCalc( XPSilarSnap *pSrcSR, XPSilarSnap *pDstSR, double w0) {
    int			numSils = xpSilarSnapGetNum( pSrcSR);
    int			totPower = 0;
    XPSilar		*pDstSilar;
    EEMComplex		srcLoc, dstLoc, dstTop, dstBot;
    double		T = pDstSR->PntView->DM.SampleTime;
    EEMComplex		sLoc, zLoc, srcGain, dstGain;

    xpSilarSnapSetNum( pDstSR, numSils+1);
    XP_SILSET_IDX_LOOP_BEGIN( pSrcSR, silIdx, pSrcSilar) {
	totPower += EEM_SILAR_TotPower(pSrcSilar);
	srcLoc = pSrcSilar->loc;
	EEM_CPLX_SET( &dstTop, 2 + T*srcLoc.r, srcLoc.i);
	EEM_CPLX_SET( &dstBot, 2 - T*srcLoc.r, -srcLoc.i);
	eemCplxDiv3( &dstTop, &dstBot, &dstLoc);
	pDstSilar = xpSilarGetOrCreate( pDstSR, silIdx);
	xpSilarConfig( pDstSilar, &dstLoc, pSrcSilar->power, pSrcSilar->flags);
    } XP_SILSET_IDX_LOOP_END();
    pDstSilar = xpSilarGetOrCreate( pDstSR, numSils);
    pDstSilar->flags &= ~EEMSilarF_Conj;
    xpSilarSetLocPower( pDstSilar, -1.0, 0.0, -totPower);
    xpSilarSnapSetGain( pDstSR, 1.0, 0.0);

    EEM_CPLX_SET(&sLoc,0.0,tan(w0*T/2.0)*2.0/T);
    eemSilarGrpEval( &pSrcSR->Grp, &sLoc, &srcGain);
    EEM_CPLX_SET(&zLoc,cos(w0*T),sin(w0*T));
    eemSilarGrpEval( &pDstSR->Grp, &zLoc, &dstGain);
    if ( !eemCplxZeroB(&srcGain) && !eemCplxZeroB(&dstGain) ) {
	eemCplxDiv3( &srcGain, &dstGain, &XP_SILSET_Gain(pDstSR));
    }
}

/* z --> s */
static void
_xpPipeSilarBilinearInvCalc( XPSilarSnap *pSrcSR, XPSilarSnap *pDstSR,
  double w0) {
    int			numSils = xpSilarSnapGetNum( pSrcSR);
    XPSilar		*pDstSilar;
    EEMComplex		srcLoc, dstLoc, dstTop, dstBot;
    double		T = pSrcSR->PntView->DM.SampleTime;
    EEMComplex		sLoc, zLoc, srcGain, dstGain;

    xpSilarSnapSetNum( pDstSR, numSils);
    XP_SILSET_IDX_LOOP_BEGIN( pSrcSR, silIdx, pSrcSilar) {
	srcLoc = pSrcSilar->loc;
	pDstSilar = xpSilarGetOrCreate( pDstSR, silIdx);
	if ( eemDfltSameB(srcLoc.r,-1.0) && eemDfltSameB(srcLoc.i,0.0) ) {
	    xpSilarConfig( pDstSilar, NULL, 0, 0);
	    continue;
	}
	EEM_CPLX_SET( &dstTop, srcLoc.r - 1, srcLoc.i);
	EEM_CPLX_SET( &dstBot, srcLoc.r + 1, srcLoc.i);
	eemCplxDiv3( &dstTop, &dstBot, &dstLoc);
	dstTop.r *= 2 / T; dstTop.i *= 2 / T;
	xpSilarConfig( pDstSilar, &dstLoc, pSrcSilar->power, pSrcSilar->flags);
    } XP_SILSET_IDX_LOOP_END();

    xpSilarSnapSetGain( pDstSR, 1.0, 0.0);
    EEM_CPLX_SET(&zLoc,cos(w0*T),sin(w0*T));
    eemSilarGrpEval( &pSrcSR->Grp, &zLoc, &srcGain);
    EEM_CPLX_SET(&sLoc,0.0,tan(w0*T/2.0)*2.0/T);
    eemSilarGrpEval( &pDstSR->Grp, &sLoc, &dstGain);
    if ( !eemCplxZeroB(&srcGain) && !eemCplxZeroB(&dstGain) ) {
	eemCplxDiv3( &srcGain, &dstGain, &XP_SILSET_Gain(pDstSR));
    }
}

static void
_xpPipeSilarMatchedZCalc( XPSilarSnap *pSrcSR, XPSilarSnap *pDstSR) {
    int			numSils = xpSilarSnapGetNum( pSrcSR);
    XPSilar		*pDstSilar;
    EEMComplex		dstLoc;
    double		rexp, T = pDstSR->PntView->DM.SampleTime;

    xpSilarSnapSetNum( pDstSR, numSils);
    XP_SILSET_IDX_LOOP_BEGIN( pSrcSR, silIdx, pSrcSilar) {
	/*
	 * z = exp(sT); for s=a+jb: z = exp(aT) * ( cos(bT) + j sin(bT) )
	 * 
	 */
	sincos( pSrcSilar->loc.i * T, &dstLoc.i, &dstLoc.r);
	rexp = exp( pSrcSilar->loc.r * T);
	dstLoc.r *= rexp;
	dstLoc.i *= rexp;
	pDstSilar = xpSilarGetOrCreate( pDstSR, silIdx);
	xpSilarConfig( pDstSilar, &dstLoc, pSrcSilar->power, pSrcSilar->flags);
    } XP_SILSET_IDX_LOOP_END();
    /*
     * I dont know what to do with poles/zeros at infinity.
     * pDstSilar = xpSilarGetOrCreate( pDstSR, numSils);
     * pDstSilar->flags &= ~EEMSilarF_Conj;
     * xpSilarSetLocPower( pDstSilar, -1.0, 0.0, -totPower);
     */
    xpSilarSnapSetGain( pDstSR,
      XP_SILSET_Gain(pSrcSR).r, XP_SILSET_Gain(pSrcSR).i);
}

/* impulse invarient transform s->z */
static void
_xpPipeSilarImpulseXfCalc( XPSilarSnap *pSrcSR, XPSilarSnap *pDstSR,
  int numzeros) {
    EEMPfeGrp		pfe;
    EEMComplex		dstLoc;
    int			m;
    double		rexp, T = pDstSR->PntView->DM.SampleTime;

    eemPfeGrpConstructor( &pfe);
    eemSilarGrpToPfeGrp( XP_SILSET_Grp(pSrcSR), &pfe, 0);
    EEM_PFE_LOOP_BEGIN( &pfe, pTerm) {
	if ( (m = - pTerm->power) <= 0 )	continue;
	eemCplxExp( T, &pTerm->pole, &dstLoc);
	pTerm->pole = dstLoc;
	/* gain *= T^m */
	rexp = m==1 ? T : pow( T, (double) m);
	pTerm->gain.r *= rexp; pTerm->gain.i *= rexp;
	pTerm->numerpow = m;
    } EEM_PFE_LOOP_END();
    eemPfeGrpToSilarGrp( &pfe, XP_SILSET_Grp(pDstSR), numzeros);
    xpSilarSnapSetPnts( pDstSR);
    eemPfeGrpDestructor( &pfe);
}

typedef enum _XPPipeSilarMapType XPPipeSilarMapType;
enum _XPPipeSilarMapType {
    XPPSMT_Null, XPPSMT_Bilinear, XPPSMT_MatchedZ, XPPSMT_ImpulseInv,
};

typedef struct _XPPipeSilarMap XPPipeSilarMap;
struct _XPPipeSilarMap {
    XPPipeSilarMapType	type;
    int			numZeros;
};

static void
_xpPipeSilarMapNotify( XPPipeMod *pPipe, XPViewEvent ev,
  XPViewSnap *pSrcSnap, XPAgent *pTgtAg) {
    /* XPSilarSnap	*pSrcSR = (XPSilarSnap*) pSrcSnap; */
    XPPipeSilarMap	*pExt = (XPPipeSilarMap*) pPipe->cSide.Ext;
    Tcl_Interp		*ip = XPTheMachine.ip;
    char		*n1 = pPipe->Lbl.Name, *mapStr, *nzStr;

    switch ( ev ) {
    case XPVE_Create:
	pExt = MEM_ALLOC(XPPipeSilarMap);
	pExt->type = XPPSMT_Bilinear;
	if ( Tcl_GetVar2(ip,n1,"maptype",TCL_GLOBAL_ONLY) == NULL )
	    Tcl_SetVar2(ip,n1,"maptype","bilinear",TCL_GLOBAL_ONLY);
	pExt->numZeros = -1;
	if ( Tcl_GetVar2(ip,n1,"numzeros",TCL_GLOBAL_ONLY) == NULL )
	    Tcl_SetVar2(ip,n1,"numzeros","-1",TCL_GLOBAL_ONLY);
	pPipe->cSide.Ext = (TOPPtr) pExt;
	/*FALLTHROUGH*/
    case XPVE_Param:
	mapStr = Tcl_GetVar2(ip,n1,"maptype",TCL_GLOBAL_ONLY);
	/*IF*/ if ( mapStr==NULL || strcmp(mapStr,"bilinear")==0 ) {
	    pExt->type = XPPSMT_Bilinear;
	} else if ( strcmp(mapStr,"matchedz")==0 ) {
	    pExt->type = XPPSMT_MatchedZ;
	} else if ( strcmp(mapStr,"impulse")==0 ) {
	    pExt->type = XPPSMT_ImpulseInv;
	}
	nzStr = Tcl_GetVar2(ip,n1,"numzeros",TCL_GLOBAL_ONLY);
	if ( nzStr != NULL ) {
	    pExt->numZeros = atoi(nzStr);
	}
	/*FALLTHROUGH*/
    case XPVE_Switch:
    case XPVE_Modify:
        XPVIEWPART_LOOP_FORW_BEGIN( pTgtAg->CurView, pVP) {
	    pVP->Flags &= ~XPViewF_Valid;
            xpViewPartNotify( pVP, XPNW_Now, XPVE_Switch, pTgtAg);
        } XPVIEWPART_LOOP_FORW_END();
	break;
    default: ;
    }
}


static TOPLogical
_xpPipeSilarMapUpdate( XPPipeMod *pPipe,
  XPViewSnap *pSrcSnap, XPViewSnap *pTgtSnap) {
    XPSilarSnap		*pSrcSR = (XPSilarSnap*) pSrcSnap;
    XPSilarSnap		*pTgtSR = (XPSilarSnap*) pTgtSnap;
    XPPipeSilarMap	*pExt = (XPPipeSilarMap*) pPipe->cSide.Ext;

    switch ( pExt->type ) {
    case XPPSMT_Bilinear:
	if ( pSrcSR->PntView->DM.XfDomain == XP_XFDOM_S ) {
	    _xpPipeSilarBilinearCalc( pSrcSR, pTgtSR, 0.0);
	} else {
	    _xpPipeSilarBilinearInvCalc( pSrcSR, pTgtSR, 0.0);
	}
	break;
    case XPPSMT_MatchedZ:
	if ( pSrcSR->PntView->DM.XfDomain == XP_XFDOM_S ) {
	    _xpPipeSilarMatchedZCalc( pSrcSR, pTgtSR);
	}
	break;
    case XPPSMT_ImpulseInv:
	if ( pSrcSR->PntView->DM.XfDomain == XP_XFDOM_S ) {
	    _xpPipeSilarImpulseXfCalc( pSrcSR, pTgtSR, pExt->numZeros);
	}
	break;
    default:
	return TOP_FALSE;
    }
    return TOP_TRUE;
}

static XPPipeClass _xpPipeSilarMapCls = {
    "PolezeroMap", 0,
    { "SilarRep", 0, _xpPipeSilarMapNotify, _xpPipeSilarMapUpdate, },
    { "SilarRep", 0, _xpPipeSilarMapNotify, _xpPipeSilarMapUpdate }
};

#ifdef notdef

/******************************************************************************
 *
 *
 *****************************************************************************/

#define MAX_COOEF	100

extern void rpoly_();

void
jtFactor( double *cooefs, int *degree, double *zeroR, double *zeroI,
  int *failB) {
#ifdef notdef
    rpoly_( cooefs, degree, zeroR, zeroI, failB);
#else
    topProgMsg( "Cant factor poly: not linked in");
#endif
}


static void
_xpPipeImpuseSilarNotify( XPPipeMod *pPipe, TOPLogical pmB, XPViewEvent t,
  XPViewSnap *pSrcSnap, XPViewSnap *pTgtSnap) {
#ifdef notdef
    XPSilar		*pSilar;
    XPSilarSnap		*pSR;
    XPSilarView	*pSdm;
    XPCplxSeqSnap		*pSeq;
    int			i, deg, failB;
    double		Cooefs[MAX_COOEF];
    double		zeroR[MAX_COOEF], zeroI[MAX_COOEF];
    EEMComplex		zeroLoc;

    if ( (pSeq = xpCplxSeqGetCurSnap( &pPipe->CM, XPGS_Upd)) == NULL ) {
	TOP_PDBG((SPKG,"%s: no source data set", pPipe->Lbl.Tag));
	return;
    }
    if ( (pSR = (XPSilarSnap*) pPipe->PM.CurView->DefaultSet) == NULL ) {
	TOP_PDBG((SPKG,"%s: no target data set", pPipe->Lbl.Tag));
	return;
    }
    pSdm = (XPSilarView*) pPipe->PM.CurView->Lbl->Me;

    switch ( t ) {
    case XPVE_Create:
	/*FALLTHROUGH*/
    case XPVE_Switch:
	/*FALLTHROUGH*/
    case XPVE_Modify:
	if ( (deg = topVecGetNum( &pSeq->Vec)-1) >= MAX_COOEF-10 ) {
	    TOP_PDBG((SPKG,"%s: truncating degree", pPipe->Lbl.Tag));
	    deg = MAX_COOEF-10;
	}
	TOPVEC_IDX_LOOP_FORW_BEGIN( &pSeq->Vec, cNum, XPCplxSeqSnapPt, pPt) {
	    if ( cNum > deg+1 )
		break;
	    Cooefs[cNum] = pPt->valRI.r;
	} TOPVEC_IDX_LOOP_FORW_END();
	jtFactor( Cooefs, &deg, zeroR, zeroI, &failB);
	TOP_PDBG((SPKG,"%s: jtFactor: (ret %d, deg %d)", pPipe->Lbl.Tag, failB, deg));
	if ( failB )
	    break;
	xpSilarSnapSetGain( pSR, Cooefs[0], 0);
	xpSilarSnapSetNum( pSR, deg);
	for ( i=0; i < deg; i++) {
	    pSilar = xpSilarGetOrCreate( pSR, i);
	    xpSilarSetLocPower( pSilar, zeroR[i], zeroI[i], 1);
	}
	xpAgentNotifyFrom( &pPipe->PM, XPNotifySN_Unspec, XPVE_Modify);
	break;
    case XPVE_TgtxxxModify:
	/*
	 * The consumer of our target data has modified the
	 * parameters of the target data.  We need to recalc everything
	 */
	_xpPipeCalcFourierPts(pSeq, 1);
	xpViewNotifyConsumer( pPipe->CM.CurView, &pPipe->CM, XPVE_Modify);
	break;
    default:
	;
    }
#endif
}

#endif

static XPNotifyWhen
_xpPipeNotifyCB( XPAgent *pSrcAg, XPViewPart *pVP, XPViewEvent ev) {
    XPPipeMod		*pPipe = (XPPipeMod*) pSrcAg->Lbl->Me;
    XPViewSnap		*pSrcSnap;
    XPPipeSide		*pSrcSide, *pTgtSide;

    if ( pSrcAg == &pPipe->cSide.Ag ) {
	pSrcSide = &pPipe->cSide;	pTgtSide = &pPipe->pSide;
    } else {
	pSrcSide = &pPipe->pSide;	pTgtSide = &pPipe->cSide;
    }
    if (  pSrcSide->Class->NotifyFunc == NULL )
	return XPNW_Null;	/* "other" dir not supported */

    switch ( ev ) {
    case XPVE_Create:
	TOP_ASSERT(SPKG, TOP_IsOff(pPipe->Flags,XPPipeF_Alive) );
	break;
    case XPVE_Param:
    case XPVE_Switch:
    case XPVE_Modify:
        if ( TOP_IsOff(pPipe->Flags,XPPipeF_Active) ) {
	    return XPNW_Defer;
        }
	/*FALLTHROUGH*/
    default:
	if ( TOP_IsOff(pPipe->Flags,XPPipeF_Alive) ) {
	    _xpPipeNotifyCB( pSrcAg, pVP, XPVE_Create);
	}
    }
    if ( (pSrcSnap = xpViewPartGetSnap( pVP, XPGS_AsIs)) == NULL )
	return XPNW_Null;

    if ( ev == XPVE_Create )		pPipe->Flags |= XPPipeF_Alive;
    else if ( ev == XPVE_Destroy )	pPipe->Flags &= ~XPPipeF_Alive;
    (*(pSrcSide->Class->NotifyFunc))( pPipe, ev, pSrcSnap, &pTgtSide->Ag);
    return XPNW_Null;
}


static TOPLogical
_xpPipeUpdateCB( XPAgent *pAg, XPViewPart *pVP) {
    XPPipeMod		*pPipe = (XPPipeMod*) pAg->Lbl->Me;
    XPViewPart		*pSrcVP;
    XPViewSnap		*pSrcSnap, *pTgtSnap;
    TOPLogical		ret;
    XPPipeSide		*pSrcSide, *pTgtSide;

    if ( pAg == &pPipe->pSide.Ag ) {
	pSrcSide = &pPipe->cSide;	pTgtSide = &pPipe->pSide;
    } else {
	pSrcSide = &pPipe->pSide;	pTgtSide = &pPipe->cSide;
    }
    TOP_ASSERT(SPKG, pVP->PntView == pTgtSide->Ag.CurView);
    if ( TOP_IsOff(pPipe->Flags,XPPipeF_Active) )
	return TOP_FALSE;
    pSrcVP = xpAgentGetCurVP( &pSrcSide->Ag);
    if ( TOP_IsOff(pPipe->Flags,XPPipeF_Alive) ) {
        _xpPipeNotifyCB( &pSrcSide->Ag, pSrcVP, XPVE_Create);
    }
    if ( (pSrcSnap = xpViewPartGetSnap( pSrcVP, XPGS_Upd)) == NULL )
	return TOP_FALSE;

    pTgtSnap = xpViewPartGetSnap( pVP, XPGS_AsIs);

    if ( (pSrcSide->Class->UpdateFunc) == NULL ) {
	topProgMsg("Pipe %s has no update function.", pAg->Lbl->Tag);
	return TOP_FALSE;
    }
    ret = (*(pSrcSide->Class->UpdateFunc))( pPipe, pSrcSnap, pTgtSnap);
    return ret;
}

static char*
_xpPipeTraceVar( ClientData cld, Tcl_Interp *ip, char *n1, char *n2, int f) {
    XPPipeMod	*pPipe = (XPPipeMod*) cld;
    char	*activeStr;
    TOPLogical	activeB;

    if ( TOP_IsOn(f,TCL_TRACE_UNSETS) ) {
	if ( n2 == NULL )
    	    xpPipeDestroy( pPipe);
	return NULL;
    }
    if ( strcmp(n2,"active")!=0 ) {
	xpAgentNotify( &pPipe->cSide.Ag, XPNW_Idle, XPVE_Param);
	return NULL;
    }
    activeStr = Tcl_GetVar2( ip, pPipe->Lbl.Name, "active", TCL_GLOBAL_ONLY);
    if ( activeStr == NULL ) {
	return "pipe(active) is missing";
    }
    if ( Tcl_GetBoolean( ip, activeStr, &activeB) != TCL_OK ) {
	return "pipe(active) has malformed boolean";
    }
    if ( activeB ) {
	pPipe->Flags |= XPPipeF_Active;
	xpAgentNotify( &pPipe->cSide.Ag, XPNW_Now, XPVE_Null);
    } else {
	pPipe->Flags &= ~XPPipeF_Active;
    }
    return NULL;
}

static DLListHandle _XpPipeClassTbl = NULL;

static void
_xpPipeClassTblInit() {
    if ( _XpPipeClassTbl != NULL )
	return;
    _XpPipeClassTbl = dlListCreate( "XPPipeClass", sizeof(XPPipeClass));
}


void
xpPipeAddClass( XPPipeClass *pNewClass) {
    XPPipeClass *pClass;

    _xpPipeClassTblInit();
    pClass = (XPPipeClass*) dlListCreateAtTail( _XpPipeClassTbl);
    *pClass = *pNewClass;
}

void
xpPipeAddStdClasses() {
    xpPipeAddClass( &_xpPipeSilarFourierCls);
    xpPipeAddClass( &_xpPipeSilarImpulseCls);
    xpPipeAddClass( &_xpPipeSilarMapCls);
    xpPipeAddClass( &_xpPipeImpulseFourierCls);
#ifdef notdef
      _xpPipeSilarImpulseNotify, NULL, sizeof(XPPipeSilarImpulse));
    xpPipeAddClass( "ImpulseFourier", "CplxSeqTime", "CplxSeqFreq", 0,
      _xpPipeImpulseFourierNotify, _xpPipeImpulseFourierNotify, 0);
    xpPipeAddClass( "ImpulseSilar", "CplxSeqTime", "SilarRep", 0,
      _xpPipeImpuseSilarNotify, _xpPipeImpuseSilarNotify, 0);
    xpPipeAddClass( "FsampImpulse", "CplxSeqFreq", "CplxSeqTime", 0,
      _xpPipeFsampImpulseNotify, NULL, 0);
#endif
}

static XPPipeClass*
_xpPipeFindClassByName( char *name, char *flags) {
    char	*s;
    int		len;

    if ( _XpPipeClassTbl == NULL )
	return NULL;
    if ( (s=strchr(name,':')) == NULL ) {
	len = strlen(name);
	if ( flags )	*flags = 'F';
    } else {
	len = s - name;
    	if ( flags )	*flags = s[1];
    }
    DLLIST_LOOP_FORW_BEGIN( _XpPipeClassTbl, XPPipeClass*, pClass) {
	if ( strncmp( name, pClass->Name, len)==0 ) {
	    return pClass;
	}
    } DLLIST_LOOP_FORW_END();
    return NULL;
}

TOPLogical
xpPipeGetTypeInfo( char *className, char **pSrcClass, char **pTgtClass) {
    XPPipeClass		*pClass;

    if ( (pClass=_xpPipeFindClassByName( className, NULL)) == NULL )
	return TOP_FALSE;
    *pSrcClass = pClass->cSide.ViewClass;
    *pTgtClass = pClass->pSide.ViewClass;
    return TOP_TRUE;
}

XPLabel*
xpPipeCreate( char *name, char *class, XPView *pSm, XPView *pTm) {
    static int		SN = 0;
    XPPipeClass		*pClass;
    XPPipeMod		*pPipe;
    Tcl_Interp		*ip = XPTheMachine.ip;
    char		dirChar;
    char		buf[100];

    if ( (pClass=_xpPipeFindClassByName( class, &dirChar)) == NULL ) {
	topErrMsg("No such pipe class ``%s''", class);
	return NULL;
    }

    if ( strcasecmp( pClass->cSide.ViewClass, pSm->Lbl.Class)!=0 ) {
	topErrMsg("Bad view src class ``%s'' (should be %s)", pSm->Lbl.Class,
	  pClass->cSide.ViewClass);
	return NULL;
    }
    if ( strcasecmp( pClass->pSide.ViewClass, pTm->Lbl.Class)!=0 ) {
	topErrMsg("Bad view tgt class ``%s'' (should be %s)", pTm->Lbl.Class,
	  pClass->pSide.ViewClass);
	return NULL;
    }

    pPipe = MEM_ALLOC(XPPipeMod);
    MEM_CLEAR(pPipe);
    xpLabelConstructor( &pPipe->Lbl, (TOPPtr)pPipe, &SN, name, class);
    pPipe->Flags = XPPipeF_Active | pClass->Flags;
    pPipe->Class = pClass;
    pPipe->cSide.Class = &pClass->cSide;
    pPipe->pSide.Class = &pClass->pSide;

    sprintf(buf,"%s-c", pPipe->Lbl.Name);
    xpLabelConstructor( &pPipe->cSide.Lbl, (TOPPtr)pPipe, NULL, buf, "PipeAg");
    xpAgentConstructor( &pPipe->cSide.Ag, &pPipe->cSide.Lbl, pSm);
    pPipe->cSide.Ag.PntPipe = pPipe;
    sprintf(buf,"%s-p", pPipe->Lbl.Name);
    xpLabelConstructor( &pPipe->pSide.Lbl, (TOPPtr)pPipe, NULL, buf, "PipeAg");
    xpAgentConstructor( &pPipe->pSide.Ag, &pPipe->pSide.Lbl, pTm);
    pPipe->pSide.Ag.PntPipe = pPipe;

    if ( dirChar == 'F' || dirChar == 'B' ) {
	if ( &pPipe->cSide.Class->NotifyFunc == NULL )	goto bad;
	pPipe->cSide.Ag.NotifyFunc = _xpPipeNotifyCB;
        pPipe->pSide.Ag.UpdateFunc = _xpPipeUpdateCB;
        pTm->Flags &= ~XPViewF_Valid;
        if ( TOP_IsOn(pPipe->cSide.Class->Flags,XPPipeF_MultiPart) )
	    pTm->Flags |= XPViewF_MultiPart;
        xpViewSetMaster( pTm, &pPipe->pSide.Ag);
    }
    if ( dirChar == 'R' || dirChar == 'B' ) {
	if ( &pPipe->pSide.Class->NotifyFunc == NULL )	goto bad;
	pPipe->pSide.Ag.NotifyFunc = _xpPipeNotifyCB;
        pPipe->cSide.Ag.UpdateFunc = _xpPipeUpdateCB;
        pSm->Flags &= ~XPViewF_Valid;
        if ( TOP_IsOn(pPipe->pSide.Class->Flags,XPPipeF_MultiPart) )
	    pSm->Flags |= XPViewF_MultiPart;
        xpViewSetMaster( pSm, &pPipe->cSide.Ag);
    }

    Tcl_SetVar2( ip, pPipe->Lbl.Name, "active", "1", TCL_GLOBAL_ONLY);
    Tcl_TraceVar2( ip, pPipe->Lbl.Name, NULL, 
      TCL_GLOBAL_ONLY|TCL_TRACE_WRITES|TCL_TRACE_UNSETS,
      _xpPipeTraceVar, (ClientData)pPipe);

    dlPtrListInsertAtTail( XPTheMachine.PipeList, (TOPPtr)pPipe);
    return & pPipe->Lbl;

bad:;
    xpPipeDestroy( pPipe);
    return NULL;
}


void
xpPipeDestroy( XPPipeMod *pPipe) {
    Tcl_Interp		*ip = XPTheMachine.ip;

    dlPtrListElemRemove( XPTheMachine.PipeList, (TOPPtr)pPipe);
    Tcl_UntraceVar2( ip, pPipe->Lbl.Name, NULL,
      TCL_GLOBAL_ONLY|TCL_TRACE_WRITES|TCL_TRACE_UNSETS, 
      _xpPipeTraceVar, (ClientData)pPipe);
    Tcl_UnsetVar2( ip, pPipe->Lbl.Name, NULL, TCL_GLOBAL_ONLY);

    xpAgentDestructor( &pPipe->cSide.Ag);
    xpLabelDestructor( &pPipe->cSide.Lbl);
    xpAgentDestructor( &pPipe->pSide.Ag);
    xpLabelDestructor( &pPipe->pSide.Lbl);
    MEM_FREE(pPipe->cSide.Ext);	pPipe->cSide.Ext = NULL;
    MEM_FREE(pPipe->pSide.Ext);	pPipe->pSide.Ext = NULL;
    xpLabelDestructor( &pPipe->Lbl);
    MEM_FREE(pPipe);
}

XPPipeMod*
xpPipeFindByName( char *name) {
    DLPTRLIST_LOOP_FORW_BEGIN( XPTheMachine.PipeList, XPPipeMod*, pPipe) {
	if ( pPipe==NULL )	continue;
	if ( strcmp(name,pPipe->Lbl.Name)==0 )
	    return pPipe;
    } DLPTRLIST_LOOP_FORW_END();
    return NULL;
}
