/**
    silar.c

    Author: Kennard White
    90/09/30 KDW: Created as zplane.c
    91/01/07 KDW: Extracted db code into silar.c

    Copyright (c) 1990-93 The Regents of the University of California.
    All rights reserved.  See the file "./copyright.h" for full copyright
    notice, limitation of liability, and disclaimer of warranty provisions.
#include "copyright.h"				COPYRIGHTENDKEY
**/
#if !defined(lint) && !defined(SABER)
static char rcsid[] = "$Header: /vol/cooley/cooley1/kennard/src/xpole/RCS/silar.c,v 1.5 93/08/27 14:28:15 kennard Exp $";
#endif

#include "topFixup.h"
#include <stdio.h>
#include <tcl.h>

#include "topStd.h"
#include "topCB.h"

#include "xpXfa.h"

#include "pole.h"
#include "filter.h"
#include "silar.h"

char _XpSilarSubPackage[] = "xp.silar";
#define SPKG _XpSilarSubPackage


/************************************************************************
	Callback mechanism for handling update notification to datamod
************************************************************************/

/**
    Traverse the silar list looking for changes.  Reflect these
    changes out to the rest of the application.
    
    XXX: Should allow changes to take place while we do this,
    but currently cant happen.

     * Note that any "global" type work done here (such as gc),
     * wont be reflected back to the orginating agent.

**/
static void 
_xpSilarSnapNotifyCB( XPViewPart *pVP, XPViewEvent ev, XPAgent *pAgOmit) {
    XPSilarSnap		*pSR = (XPSilarSnap*) xpViewPartGetSnap(pVP, XPGS_AsIs);

    pSR->ChangeMask = pSR->NextChangeMask;
    pSR->NextChangeMask = 0;
    eemSilarGrpBB( &pSR->Grp, &pSR->bbLo, &pSR->bbHi);
    xpViewPartNotifyAgents( pVP, ev, pAgOmit);
    if ( TOP_IsOn( pSR->ChangeMask, EEMSilarC_Exist|EEMSilarC_Sel) 
      && pSR->NextChangeMask == 0 ) {
	eemSilarGrpGC( &pSR->Grp);	/* also resets change flags */
    }
}

void
xpSilarSnapChange( XPSilarSnap *pSR, XPSilarChangeMask mask) {
    pSR->NextChangeMask |= mask;
}

void
xpSilarChange( XPSilar *pSilar, XPSilarChangeMask mask) {
    mask &= EEMSilarC_All;
    pSilar->flags |= mask;
    xpSilarSnapChange( XP_SILSET_PntRep(pSilar), mask);
}


/******************************************************************************
		Functions for dealing with XPSilarSnap's
******************************************************************************/

static void
_xpSilarSnapConstructor( XPSilarSnap *pSR, XPSilarView *pSView, EEMComplex *pGain) {
    pSR->PntView = pSView;
    eemSilarGrpConstructor( &pSR->Grp);
    if ( pGain == NULL ) {
        EEM_CPLX_SET( &pSR->Grp.gain, 1.0, 0.0);
    } else {
	pSR->Grp.gain = *pGain;
    }
    pSR->ChangeMask = 0;
    pSR->NextChangeMask = EEMSilarC_All;
}

static void
_xpSilarSnapDestructor( XPSilarSnap *pSR) {
    eemSilarGrpDestructor( &pSR->Grp);
}

XPSilarSnap*
xpSilarSnapCreate( XPSilarView *pSView, EEMComplex *pGain) {
    XPSilarSnap	*pSR;

    pSR = MEM_ALLOC(XPSilarSnap);
    _xpSilarSnapConstructor( pSR, pSView, pGain);
    return pSR;
}

static void xpSilarSnapDestroy( XPSilarSnap *pSR) {
    _xpSilarSnapDestructor( pSR);
    MEM_FREE(pSR);
}

static XPViewSnap* _xpSilarSnapCreateCB( XPView *pView) {
    return (XPViewSnap*) xpSilarSnapCreate( (XPSilarView*) pView, NULL);
}

static void _xpSilarSnapDestroyCB( XPViewSnap *pVS) {
    xpSilarSnapDestroy( (XPSilarSnap*) pVS);
}

void
xpSilarSnapClear( XPSilarSnap *pSR, EEMComplex *pGain) {
    if ( pGain == NULL ) {
	xpSilarSnapSetGain( pSR, 1.0, 0.0);
    } else {
	xpSilarSnapSetGain( pSR, pGain->r, pGain->i);
    }
    xpSilarSnapSetNum( pSR, 0);
}

void
xpSilarSnapSetGain( XPSilarSnap *pSR, double r, double i) {
    EEM_CPLX_SET( &XP_SILSET_Gain(pSR), r, i);
    xpSilarSnapChange( pSR, EEMSilarC_Gain);
}

#ifdef notdef
void
xpSilarSnapInsertAtTail( XPSilarSnap *pSR, XPSilar *pSilar) {
    TOPVEC_APPEND( &pSR->Grp.silars, XPSilar*) = pSilar;
}
#endif

int
xpSilarSnapGetNum( XPSilarSnap *pSR) {
    return XP_SILSET_Num( pSR);
}

void
xpSilarSnapSetNum( XPSilarSnap *pSR, int num) {
    int i, curnum = XP_SILSET_Num( pSR);

    if ( num <= curnum ) {
	/*
	 * If contracting, then set change flags for dead silars
	 * and garbage collect the dead later.
	 */
	for (i=num; i < curnum; i++) {
	    xpSilarDestroy( &XP_SILSET_Get( pSR, i));
	}
    } else {
	/* 
	 * If expanding, set the pnt pointer.  The eem level
	 * should have already initialized the new silars.
	 */
        eemSilarGrpSetNum( &pSR->Grp, num);
	for (i=curnum; i < num; i++) {
	    XP_SILSET_PntRep(&XP_SILSET_Get( pSR, i)) = pSR;
	}
    }
}

void
xpSilarSnapInsureNum( XPSilarSnap *pSR, int num) {
    if ( num > XP_SILSET_Num(pSR) )
        xpSilarSnapSetNum( pSR, num);
}

int
xpSilarSnapCalcRange( XPSilarSnap *pSR, EEMComplex *pLo, EEMComplex *pHi) {
    int		cnt = 0;
    EEMComplex	*pL, a;

    XP_SILSET_LOOP_BEGIN( pSR, pSilar) {
	if ( XP_SilarIsDeadB(pSilar) )		continue;
	pL = & pSilar->loc;
	if ( cnt == 0 )		*pLo = *pHi = *pL;
	else			eemCplxMaxRect( pL, pLo, pHi);
	++cnt;
	if ( TOP_IsOn(pSilar->flags, EEMSilarF_Conj) ) {
	    a.r = pL->r;
	    a.i = - pL->i;
	    eemCplxMaxRect( &a, pLo, pHi);
	}
    } XP_SILSET_LOOP_END();
    if ( cnt == 0 ) {
	EEM_CPLX_SET( pLo, -1.0, -1.0);
	EEM_CPLX_SET( pHi,  1.0,  1.0);
    } else {
	if ( pLo->r == pHi->r )
	    pLo->r -= .5; pHi->r += .5;
	if ( pLo->i == pHi->i )
	    pLo->i -= .5; pHi->i += .5;
    }
    return cnt;
}

void
xpSilarSnapFreqResp( XPSilarSnap *pSR, EEMComplex *pFreq, EEMComplex *pResp) {
    eemSilarGrpEval( XP_SILSET_Grp(pSR), pFreq, pResp);
}

/******************************************************************************
		Old sludge
******************************************************************************/

#ifdef notdef

XPSilarChangeMask
xpSilarSetCalcChange( XPSilarSet *pA, XPSilarSet *pB) {
    XPSilarChangeMask	mask, totmask;
    /*
     * XXX: should do diffs to set change masks right here
     * Instead, just fake it with a forced re-calc of everything
     */
    XP_SILSET_IDX_LOOP_BEGIN( pB, silNum, pSB) {
	mask = XP_SC_ALL;
	pSB->ChangeMask = mask;
	totmask |= mask;
    } XP_SILSET_IDX_LOOP_END();
    return totmask;
}

void
xpSilarTrajSetSnap( XPSilarSetSnap *pSnap) {
    XPSilarSetTraj	*pTraj = pSnap->PntTraj;
    XPSilarView		*pDB = pTraj->PntDB;
    XPSilarSet		*pNewSet;

    pTraj->CurSnap = pSnap;
    if ( pDB->CurTraj == pTraj ) {
	pNewSet = & pTraj->CurSnap->Silars;
    	pSnap->ChangeMask = xpSilarSetCalcChange( pDB->CurSet, pNewSet);
    	pDB->CurSet = pNewSet;
    	xpSilarChnageDB( pDB, pSnap->ChangeMask);
    }
}

#endif

/******************************************************************************
		Functions for dealing with XPSilar's
******************************************************************************/

void
xpSilarSnapSetPnts( XPSilarSnap *pSR) {
    XP_SILSET_LOOP_BEGIN( pSR, pSilar) {
        XP_SILSET_PntRep(pSilar) = pSR;
    } XP_SILSET_LOOP_END();
}

XPSilar*
xpSilarGetAvail( XPSilarSnap *pSR) {
    int		idx;
    XPSilar	*pSilar;
    idx = eemSilarGrpGetAvail( XP_SILSET_Grp(pSR));
    pSilar = &XP_SILSET_Get(pSR,idx);
    pSilar->power = 0;
    pSilar->flags = 0;
    XP_SILSET_PntRep(&XP_SILSET_Get( pSR, idx)) = pSR;
    xpSilarChange( pSilar, EEMSilarC_All);
    return pSilar;
}

XPSilar*
xpSilarGetOrCreate( XPSilarSnap *pSR, int idx) {
    XPSilar *pSilar;
    
    if ( idx < 0 ) {
	pSilar = xpSilarGetAvail( pSR);
    } else {
        xpSilarSnapInsureNum( pSR, idx+1);
        pSilar = &XP_SILSET_Get(pSR, idx);
    }
    return pSilar;
}



void xpSilarMove( XPSilar *pSilar, double r, double i) {
    pSilar->loc.r = r;
    pSilar->loc.i = i;
    xpSilarChange( pSilar, EEMSilarC_Loc);
}

void xpSilarSetLocPower( XPSilar *pSilar, double r, double i, int power) {
    pSilar->loc.r = r;
    pSilar->loc.i = i;
    pSilar->power = power;
    xpSilarChange( pSilar, EEMSilarC_Loc|EEMSilarC_Power);
}

void
xpSilarConfig( XPSilar *pSilar, EEMComplex *pLoc, int power, int flags) {
    TOPMask	chf, chm = 0;

    if ( pLoc != NULL ) {
        pSilar->loc = *pLoc;
	chm |= EEMSilarC_Loc;
    }
    chf = (flags ^ pSilar->flags) & (EEMSilarF_Conj|EEMSilarF_Sel);
    pSilar->flags ^= chf;
    if ( pSilar->power != power || TOP_IsOn(chf,EEMSilarF_Conj) ) {
	pSilar->power = power;
	chm |= EEMSilarC_Power;
    }
    if ( TOP_IsOn(chf, EEMSilarF_Sel) )
	chm |= EEMSilarC_Sel;
    xpSilarChange( pSilar, chm);
}

void
xpSilarCopyByIdx( XPSilarSnap *pSR, int idx, XPSilar *pSrcSilar) {
    XPSilar *pDstSilar = xpSilarGetOrCreate( pSR, idx);
    xpSilarConfig( pDstSilar,
      &pSrcSilar->loc, pSrcSilar->power, pSrcSilar->flags);
}

/**
    Mark the silar as dead and GC is later.
**/
void xpSilarDestroy( XPSilar *pSilar) {
    if ( XP_SilarIsDeadB(pSilar) )
	return;
    pSilar->power = 0;
    xpSilarChange( pSilar, EEMSilarC_Power|EEMSilarC_Exist);
}

#ifdef notdef

void xpSilarDestroyList( DLPtrListHandle silarList) {
    DLPTRLIST_LOOP_FORW_BEGIN( silarList, XPSilar*, pSilar) {
	xpSilarDestroy( pSilar);
    } DLPTRLIST_LOOP_FORW_END();
}


/* Expand to to regex or ?,* type matching using hmBAG return */
XPSilar*
xpSilarFindByName( XPSilarSnap *pSR, char *name) {
    XP_SILSET_LOOP_BEGIN( pSR, pSilar) {
	if ( strcasecmp( pSilar->Name, name)==0 )
	    return( pSilar );
    } XP_SILSET_LOOP_END();
    return( NULL );
}

#endif

/******************************************************************************
		XPSilarView
		Selecting Silars
******************************************************************************/

#ifdef notdef

void xpSilarSel( XPSilarView *pDB, XPSilar *pSilar, TOPLogical selB) {
    TOPLogical		isSelB = TOP_IsOn(pSilar->Flags,XP_SILAR_SEL);

    TOP_PDBG((SPKG,"setting select state of silar %s (%d -> %d)",
      pSilar->Name, isSelB, selB));
    if ( (isSelB?1:0) == (selB?1:0) )
	return;

    if ( selB ) {
	pSilar->Flags |= XP_SILAR_SEL;
        /* XXX: just-in-case: should remove check */
    	if ( dlPtrListElemRemove( pDB->SelList, pSilar) == 0 ) {
	    TOP_PDBG((SPKG,"silar %s: in SelList but not marked", pSilar->Name));
	} else {
            dlPtrListInsertAtHead( pDB->SelList, pSilar);
	}
    } else {
	pSilar->Flags &= ~XP_SILAR_SEL;
    	if ( dlPtrListElemRemove( pDB->SelList, pSilar) == 1 ) {
	    TOP_PDBG((SPKG,"silar %s: not in SelList but marked", pSilar->Name));
	}
    }
    xpChangeSilar( pSilar, XP_SC_SEL);
}

void xpSilarSelSetOne( XPSilarView *pDB, XPSilar *pNewSilar, int AddB) {
    if ( !AddB ) {
        DLPTRLIST_LOOP_FORW_BEGIN( pDB->SelList, XPSilar*, pOldSilar) {
	    if ( pOldSilar == pNewSilar ) {
		pNewSilar = NULL;
	    } else {
		xpSilarSel( pDB, pOldSilar, TOP_FALSE);
	    }
    	} DLPTRLIST_LOOP_FORW_END();
    }
    if ( pNewSilar )
	xpSilarSel( pDB, pNewSilar, TOP_TRUE);
}

/**
    Set the current selection set to {pNewSet}.  If {AddB} is TRUE,
    the existing set will be untouched, if FALSE, the existing
    selection set will be cleared.
**/
void xpSilarSelSet( XPSilarView *pDB, DLPtrListHandle pNewSet, TOPLogical AddB) {
    if ( !AddB ) {
	/* then want to clear the existing list */
        DLPTRLIST_LOOP_FORW_BEGIN( pDB->SelList, XPSilar*, pOldSilar) {
	    if ( dlPtrListElemRemove( pNewSet, pOldSilar) == 1 ) {
		/* not found: it was selected and is no more */
		xpSilarSel( pDB, pOldSilar, TOP_FALSE);
	    }
    	} DLPTRLIST_LOOP_FORW_END();
    }
    DLPTRLIST_LOOP_FORW_BEGIN( pNewSet, XPSilar*, pNewSilar) {
	xpSilarSel( pDB, pNewSilar, TOP_TRUE);
    } DLPTRLIST_LOOP_FORW_END();
}


#endif

extern int
_xpSilarSnapCmds( ClientData cld, Tcl_Interp *ip, int aC, char **aV);


static void
_xpSilarViewConstructor( XPView *pView, char *name) {
    XPSilarView	*pSView = (XPSilarView*) pView;

    Tcl_CreateCommand( XPTheMachine.ip, pSView->DM.Lbl.Tag, _xpSilarSnapCmds,
      (ClientData)pSView, NULL);
}

static void
_xpSilarViewDestructor( XPView *pView) {
    XPSilarView	*pSView = (XPSilarView*) pView;

    Tcl_DeleteCommand( XPTheMachine.ip, pSView->DM.Lbl.Tag);
}

/*global*/ XPViewClass XpSilarViewClass = {
    "SilarRep", sizeof(XPSilarView), NULL,
    _xpSilarViewConstructor, _xpSilarViewDestructor,
     _xpSilarViewCmds, _xpSilarSnapNotifyCB, _xpSilarNotifyTcl,
     _xpSilarSnapCreateCB, _xpSilarSnapDestroyCB
};


XPSilarSnap*
xpSilarGetSnap( XPAgent *pAgent, XPGetSnapHow how) {
    return (XPSilarSnap*) xpAgentGetSnap( pAgent, how);
}

XPSilarSnap*
xpSilarViewGetSnap( XPView *pView, XPGetSnapHow how) {
    return (XPSilarSnap*) xpViewGetSnap( pView, how);
}
