/* 
    silarDh.c :: Singularity Representation Data Handler :: XPole

    This module provides the interface between an XPSilarView
    and the xpplane widget.

    Author: Kennard White

    Copyright (c) 1990-93 The Regents of the University of California.
    All rights reserved.  See the file "./copyright.h" for full copyright
    notice, limitation of liability, and disclaimer of warranty provisions.
#include "copyright.h"				COPYRIGHTENDKEY
*/
#if !defined(lint) && !defined(SABER)
static char rcsid[] = "$Header: /vol/cooley/cooley1/kennard/src/xpole/RCS/silarDh.c,v 1.5 93/08/27 14:28:15 kennard Exp $";
#endif

#include "topFixup.h"
#include <stdio.h>
#include <tcl.h>

#include "topStd.h"
#include "topMem.h"
#include "topVec.h"
#include "eemSilar.h"

#include "xpXfa.h"
#include "xpPlane.h"

#include "pole.h"
#include "filter.h"
#include "silar.h"

extern char _XpSilarSubPackage[];
#define SPKG _XpSilarSubPackage

/**
	XPSilarSnapDH

	This stores local state needed by the data handler.
**/
typedef struct _XPSilarSnapDH XPSilarSnapDH;
struct _XPSilarSnapDH {
    XPLabel		Lbl;
    XPConsumer		CM;
    XPPlaneDataExt	*ext;		/* for xpPlane widget */
};

/**
    This function called by the XPole core (data module) to notify
    us of changes to the SilarRep data set.
**/
static XPNotifyWhen
_xpSilarDhNotifyCB( XPAgent *pAgent, XPViewPart *pVP, XPViewEvent ev) {
    XPSilarSnapDH	*pDH = (XPSilarSnapDH*) pAgent->Lbl->Me;
    XPSilarSnap		*pSR;
    
    if ( pDH == NULL )
	return XPNW_Defer;
    pSR = (XPSilarSnap*) xpViewPartGetSnap( pVP, XPGS_Upd);

    pDH->ext->bbLo = pSR->bbLo;
    pDH->ext->bbHi = pSR->bbHi;
    switch ( ev ) {
    case XPVE_Switch:
	eemSilarGrpCopy( XP_SILSET_Grp( pSR), &pDH->ext->wSilars);
	xpPlaneDataUpdateAll( pDH->ext);
	break;
    case XPVE_Modify:
	eemSilarGrpSetNum( &pDH->ext->wSilars, XP_SILSET_Num( pSR));
	XP_SILSET_IDX_LOOP_BEGIN( pSR, idx, pSilar) {
	    if ( TOP_IsOn(pSilar->flags,EEMSilarC_All) ) {
		EEM_SILAR_Get( &pDH->ext->wSilars, idx) = *pSilar;
	        xpPlaneDataUpdateOne( pDH->ext, idx);
	    }
	} XP_SILSET_IDX_LOOP_END();
	break;
    case XPVE_Destroy:
	xpPlaneDataDestroy( pDH->ext);
	break;
    default:
	;
    }
    return XPNW_Null;
}

static TOPLogical
_xpSilarDhConstructor( XPPlaneDataExt *pExt) {
    static int		SN = 0;
    XPSilarSnapDH	*pDH;
    XPView		*pView;

    if ( (pView = xpViewGetBySpec( pExt->dataSpec, "SilarRep",NULL,TOP_TRUE))
      == NULL )
	return TOP_FALSE;

    pDH = MEM_ALLOC(XPSilarSnapDH);
    xpLabelConstructor( &pDH->Lbl, (TOPPtr)pDH, &SN, NULL, "SilarDH");
    xpAgentConstructor( &pDH->CM, &pDH->Lbl, pView);
    pDH->CM.NotifyFunc = _xpSilarDhNotifyCB;
    pDH->ext = pExt;
    pExt->userData = (TOPPtr) pDH;

    xpAgentNotify( &pDH->CM, XPNW_Idle, XPVE_Switch);

    return TOP_TRUE;
}

static TOPLogical
_xpSilarDhDestructor( XPPlaneDataExt *pExt) {
    XPSilarSnapDH	*pDH = (XPSilarSnapDH*) pExt->userData;

    xpAgentDestructor( &pDH->CM);
    xpLabelDestructor( &pDH->Lbl);
    pDH->ext = NULL;

    MEM_FREE(pDH);
    pExt->userData = NULL;
    return TOP_TRUE;
}

static TOPLogical
_xpSilarDhChangeSilar( XPPlaneDataExt *pExt, XPPlaneChangeType ct, int idx) {
    XPSilarSnapDH	*pDH = (XPSilarSnapDH*) pExt->userData;
    XPViewPart		*pVP;
    XPSilarSnap		*pSR;

    if ( pDH == NULL )
	return TOP_FALSE;
    if ( (pVP = xpAgentGetCurVP( &pDH->CM)) == NULL )
	return NULL;
    pSR = (XPSilarSnap*) xpViewPartGetSnap( pVP, XPGS_AsIs);
    if ( idx == -1 ) {
	xpSilarSnapSetGain( pSR, pExt->wSilars.gain.r,  pExt->wSilars.gain.i);
    } else {
        xpSilarCopyByIdx( pSR, idx, &EEM_SILAR_Get(&pExt->wSilars,idx));
    }
    xpPlaneDataUpdateOne( pExt, idx);
    xpViewPartNotify( pVP, XPNW_Idle, XPVE_Modify, &pDH->CM);
    return TOP_TRUE;
}

static XPPlaneDataHandler _XpSilarDataHandler = {
    _xpSilarDhConstructor,
    _xpSilarDhDestructor,
    _xpSilarDhChangeSilar,
    NULL,	/* calc loci */
    NULL,	/* map loci */
};

int
xpSilarDhRegisterCmds( Tcl_Interp *ip) {
    xpRegisterDataHandler( ip, "xpole", "XPPlaneDataHandler",
      &_XpSilarDataHandler);
    return TCL_OK;
}
