/* 
    silarTcl.c :: TKPole

    TCL interface to silar.c

    Author: Kennard White
    91/Sep/4 KDW: Created as filter_t.c

    Copyright (c) 1990-93 The Regents of the University of California.
    All rights reserved.  See the file "./copyright.h" for full copyright
    notice, limitation of liability, and disclaimer of warranty provisions.
#include "copyright.h"				COPYRIGHTENDKEY
*/
#if !defined(lint) && !defined(SABER)
static char rcsid[] = "$Header: /vol/cooley/cooley1/kennard/src/xpole/RCS/silarTcl.c,v 1.5 93/08/27 14:28:14 kennard Exp $";
#endif

#include "topFixup.h"

#include <tcl.h>

#include "topStd.h"
#include "topBuf.h"
#include "topTcl.h"

#include "pole.h"
#include "filter.h"
#include "silar.h"

extern char _XpSilarSubPackage[];
#define SPKG _XpSilarSubPackage

int
_xpSilarNotifyTcl( XPView *pView, XPViewEvent ev,
  XPViewSnap *pSnap, Tcl_Interp *ip, char *basecmd) {
    XPSilarSnap	*pSR = (XPSilarSnap*) pSnap;
    return eemSilarGrpNotifyTcl( ip, XP_SILSET_Grp(pSR), basecmd,
      ev==XPVE_Modify ? EEMSilarC_All : 0);
}

static XPSilar*
_xpSilarWhichPrs( XPSilarSnap *pSR, Tcl_Interp *ip, char *which) {
    char	 *w = which, *whichEnd;
    XPSilar	 *pSilar = NULL;
    TOPLogical	CreateB = TOP_FALSE;
    int		idx;

    if ( which == NULL ) {
    	Tcl_AppendResult(ip,"WhichSilar: invalid which (missing)", NULL);
	return NULL;
    }
    if ( *w == '!' ) {
        CreateB = TOP_TRUE;  ++w;
    }
    idx = strtol( w, &whichEnd, 0);
    if ( whichEnd == w || *whichEnd != NULL ) {
    	Tcl_AppendResult(ip,"WhichSilar: invalid which ``",which,"'' (syntax)",
	  NULL);
	return NULL;
    }
    /* neg idx is ok, and is understood by GetOrCreate() */
    if ( idx >= XP_SILSET_Num(pSR) && !CreateB ) {
	Tcl_AppendResult(ip,"WhichSilar: invalid which ``",which,"'' (range)",
	  NULL);
	return NULL;
    }
    pSilar = xpSilarGetOrCreate( pSR, idx);
    return pSilar;
}

/**
    Commands actting on the specific silarrep {pSR}.  The command is
    {cmd}, and the arguments are {aC,aV}.
**/
int
_xpSilarSnapCmdsPrs( XPSilarSnap *pSR, Tcl_Interp *ip,
  char *cmd, int aC, char **aV) {
    char	c;
    int		len;
    XPSilar	*pSilar;
    EEMComplex	g;

    if ( pSR==NULL || cmd==NULL ) {
        Tcl_AppendResult(ip,"silar_cmds: ",
	  "tag|gain|num|clear|silar|insert|delete ?args?",NULL);
	return TCL_ERROR;
    }
    c = cmd[0]; len = strlen(cmd);

    /*IF*/ if ( c=='t' && strncmp(cmd,"tag",len)==0 ) {
	Tcl_SetResult( ip, pSR->PntView->DM.Lbl.Tag, TCL_STATIC);
	return TCL_OK;
    } else if ( c=='g' && strncmp(cmd,"gain",len)==0 ) {
	static char	fCmd[] = "silar_cmd gain";
	if ( aC == 0 ) {
	    eemCplxFmtTcl( ip, &XP_SILSET_Gain(pSR), TOP_FALSE, NULL);
	    return TCL_OK;
	}
	if ( eemCplxStrPrsTcl( ip, aV[0], &g) != TCL_OK ) {
	    Tcl_AppendResult(ip,"\n",fCmd,": expected gain, got ``",aV[0],"''",
	      NULL);
	    return TCL_ERROR;
	}
	xpSilarSnapSetGain( pSR, g.r, g.i);
	return TCL_OK;
    } else if ( c=='n' && strncmp(cmd,"num",len)==0 ) {
	static char	fCmd[] = "silar_cmd num";
	int numSils;
	if ( aC == 0 ) {
	    topTclRetFmt( ip, "%d", XP_SILSET_Num(pSR));
	    return TCL_OK;
	}
	if ( Tcl_GetInt( ip, aV[0], &numSils) != TCL_OK ) {
	    Tcl_AppendResult(ip,"\n",fCmd,
	      ": expected number of silars, got ``",aV[0],"''", NULL);
	    return TCL_ERROR;
	}
	xpSilarSnapSetNum(pSR, numSils);
	return TCL_OK;
    } else if ( c=='c' && strncmp(cmd,"clear",len)==0 ) {
	xpSilarSnapClear( pSR, NULL);
	return TCL_OK;
    } else if ( c=='g' && strncmp(cmd,"get",len)==0 ) {
	if ( aC == 0 ) {
	    eemSilarGrpFmtTcl( ip, XP_SILSET_Grp(pSR), TOP_FALSE);
	    return TCL_OK;
	}
    } else if ( c=='s' && strncmp(cmd,"silar",len)==0 ) {
	static char	fCmd[] = "silar_cmd silar";
        if ( (pSilar = _xpSilarWhichPrs( pSR, ip, aV[0])) == NULL ) {
	    Tcl_AppendResult(ip,"\n",fCmd,": bad which", NULL);
	    return TCL_ERROR;
	}
	if ( aC <= 1 ) {
	    eemSilarFmtTcl( ip, pSilar, TOP_FALSE);
	} else {
	    if ( eemSilarPrsTcl( ip, pSilar, aC-1, aV+1) != TCL_OK ) {
	        Tcl_AppendResult(ip,"\n",fCmd,": bad spec", NULL);
	        return TCL_ERROR;
	    }
	    xpSilarChange( pSilar, pSilar->flags & EEMSilarC_All);
	}
    } else if ( c=='i' && strncmp(cmd,"insertsilar",len)==0 ) {
	static char	fCmd[] = "silar_cmd insertsilar";
	pSilar = xpSilarGetAvail(pSR);
	if ( eemSilarPrsTcl( ip, pSilar, aC, aV) != TCL_OK ) {
	    Tcl_AppendResult(ip,"\n",fCmd,": invalid silar spec", NULL);
	    return TCL_ERROR;
	}
    } else if ( c=='d' && strncmp(cmd,"deletesilar",len)==0 ) {
	static char	fCmd[] = "silar_cmd deletesilar";
        if ( (pSilar = _xpSilarWhichPrs( pSR, ip, aV[0])) == NULL ) {
	    Tcl_AppendResult(ip,"\n",fCmd,": bad which", NULL);
	    return TCL_ERROR;
	}
	xpSilarDestroy( pSilar);
    } else {
	return TCL_CONTINUE;
    }
    return TCL_OK;
}

int
_xpSilarSnapCmds( ClientData cld, Tcl_Interp *ip, int aC, char **aV) {
    int			r;
    XPSilarView		*pDM = (XPSilarView*) cld;
    XPSilarSnap		*pSR = xpSilarViewGetSnap( &pDM->DM, XPGS_AsIs);
    char		*cmd;

    if ( pSR == NULL ) {
	topTclRetFmt( ip, "%s: no current set", aV[0]);
	return TCL_ERROR;
    }
    cmd = aV[1];
    if ( (r=_xpSilarSnapCmdsPrs( pSR, ip, cmd, aC-2, aV+2))
      == TCL_OK )
	return TCL_OK;
    if ( r == TCL_CONTINUE ) {
	Tcl_AppendResult(ip,aV[0],": bad sub-command ``",cmd,"''",NULL);
	return TCL_ERROR;
    }
    return TCL_ERROR;
}

int
_xpSilarViewCmds( Tcl_Interp *ip, XPView *pView, char *cmd, int aC, char **aV) {
    /* XPSilarView *pSView = (XPSilarView*) pView; */
    XPSilarSnap *pSR = xpSilarViewGetSnap( pView, XPGS_AsIs);

    if ( pSR == NULL ) {
	topTclRetFmt( ip, "%s: no current set", aV[0]);
	return TCL_ERROR;
    }
    return _xpSilarSnapCmdsPrs( pSR, ip, cmd, aC, aV);
}
