/* 
    Author: Kennard White

    Copyright (c) 1990-93 The Regents of the University of California.
    All rights reserved.  See the file "./copyright.h" for full copyright
    notice, limitation of liability, and disclaimer of warranty provisions.
#include "copyright.h"				COPYRIGHTENDKEY
*/
#if !defined(lint) && !defined(SABER)
static char rcsid[] = "$Header: /vol/cooley/cooley1/kennard/src/xpole/RCS/viewTcl.c,v 1.6 93/08/27 14:28:10 kennard Exp $";
#endif

#include "topFixup.h"
#include <stdio.h>
#include <tcl.h>

#include "topStd.h"
#include "topBuf.h"

#include "pole.h"
#include "filter.h"
#include "silar.h"
#include "cplxseq.h"
#include "pipe.h"

extern char _XpTrajSubPackage[];
#define SPKG _XpTrajSubPackage


/****************************************************************************
 *
 * There is a fine distinsion between a DataSet and DataMod.
 * A DataMod refers to the entire set of data accross an entire
 * trajectory of snaps, while a DataSet refers to just a single
 * snap within the trajectory.
 *
 * Thus a DataSetCmd is a command acting on a specific snap,
 * and a DataModCmd acts globally on all the data.
 *
 ****************************************************************************/


/****************************************************************************
 *
 * Functions to build up trajectories in bulk (many snaps)
 *
 ****************************************************************************/


#ifdef notdef
static int
_xpConvertSilarRepPrs( Tcl_Interp *ip, XPView *pView, int aC, char **aV, 
  XPViewSnap **ppSet) {
    EEMComplex		gain;
    XPSilarSnap		*pSR;


    if ( eemCplxStrPrsTcl( ip, aV[0], &gain) != TCL_OK ) {
	topTclAppendResultFmt( ip, "SilarRep: gain: %s", aV[0]);
	return TCL_ERROR;
    }
    pSR = xpSilarSnapCreate( (XPSilarView*)pView, &gain);
#ifdef notdef
	XXXX
    for ( i=1; i < aC; i++) {
	if ( _xpSilarStrPrs( ip, &tmpSilar, aV[i]) != TCL_OK ) {
	    topTclAppendResultFmt( ip, "SilarRep: bad silar: %s", aV[i]);
	    return TCL_ERROR;
        }
	if ( (pSilar = xpSilarCreate( pSR, tmpSilar.Name,
	  &tmpSilar.pz.Loc, tmpSilar.pz.Power, tmpSilar.Flags))== NULL) {
	    topTclAppendResultFmt( ip, "SilarRep: create failed: %s", aV[i]);
	    return TCL_ERROR;
	}
    }
#endif
    *ppSet = (XPViewSnap*) pSR;
    return TCL_OK;
}
#endif

/*
 * This "AgentTcl" should be eliminate and replace by cb functions
 * in the datamod header
 */
typedef struct _XPAgentTcl XPAgentTcl;
struct _XPAgentTcl {
    XPLabel	Lbl;
    XPConsumer	CM;
    Tcl_Interp	*ip;
    char	*notifyProc;
};

static XPNotifyWhen
_xpAgentTclNotifyCB( XPAgent *pAgent, XPViewPart *pVP, XPViewEvent ev) {
    XPAgentTcl		*pAt = (XPAgentTcl*) pAgent->Lbl->Me;
    XPView		*pView = pAt->CM.CurView;
    XPViewSnap		*pSnap;
    char*		lbuf;
    int			r;

    if ( (pSnap = xpAgentGetSnap( &pAt->CM, XPGS_Upd)) == NULL )
	return XPNW_Null;
    lbuf = topBufLFmt( "%s {%s}", pAt->notifyProc, pView->Lbl.Tag);
    if ( pView->Class->NotifyTclFunc == NULL ) {
        r = topTclBgEval( pAt->ip, SPKG, lbuf);
    } else {
        r = (*(pView->Class->NotifyTclFunc))( pView, ev, pSnap, pAt->ip, lbuf);
    }
    topBufUnlock(lbuf);

    if ( ev == XPVE_Destroy ) {
	xpAgentDestructor( &pAt->CM);
	xpLabelDestructor( &pAt->Lbl);
	MEM_FREE(pAt->notifyProc);
	pAt->notifyProc = NULL;
	MEM_FREE(pAt);
    }
    return XPNW_Null;
}



/**
    Return format is "filter name class tag xfdom"
    Note that LabelFmt does "name class tag"
**/
void
_xpViewInfoFmtTcl( Tcl_Interp *ip, XPView *pView, TOPLogical nestB) {
    XPFilter	*pFltr = pView->PntFltr;
    char	buf[100];

    if ( nestB )		Tcl_AppendResult(ip," {",NULL);
    Tcl_AppendResult(ip,"\"",pFltr->Name,"\"", NULL);
    xpLabelFmtTcl(ip, &pView->Lbl, TOP_FALSE);
    buf[0] = pView->XfDomain;
    buf[1] = '\0';
    Tcl_AppendResult(ip,buf,NULL);
    if ( nestB )		Tcl_AppendResult(ip,"}",NULL);
}

/**
    Agent Format: name class tag flags viewspec pntpipe
**/
static void
xpAgentFmtTcl( Tcl_Interp *ip, XPAgent *pAgent, TOPLogical nestB) {
    char	buf[100];
    XPView	*pView = pAgent->CurView;

    if ( nestB )		Tcl_AppendResult(ip," {",NULL);
    xpLabelFmtTcl( ip, pAgent->Lbl, TOP_FALSE);
    sprintf(buf," %d %s.%s %s", pAgent->Flags,
      pView->PntFltr->Name, pView->Lbl.Name,
      pAgent->PntPipe ? pAgent->PntPipe->Lbl.Name : "");
    Tcl_AppendResult(ip,buf,NULL);
    if ( nestB )		Tcl_AppendResult(ip," }",NULL);
}


/**
    Process commands for a specific agent {pAgent}.  The agent command
    is given by {cmd}, and {aC,aV} are the argments to the command.
**/
int
_xpAgentCmdsPrs( XPAgent *pAgent, Tcl_Interp *ip, char *cmd, int aC, char **aV){
    int		len, r;
    char	c;

    if ( pAgent==NULL || cmd==NULL ) {
	Tcl_AppendResult(ip,"agent_cmds: destroy|info",NULL);
	return TCL_ERROR;
    }
    c = *cmd; len = strlen(cmd);
    /*IF*/ if ( c == 'i' && strncmp(cmd,"info",len)==0 ) {
	xpAgentFmtTcl( ip, pAgent, TOP_FALSE);
    } else if ( c == 'd' && strncmp(cmd,"destroy",len)==0 ) {
	xpAgentDestructor( pAgent);
#ifdef notdef
    } else if ( strcasecmp( aV[1], "head")==0 ) {
	if ( (s=xpAgentChangeCur( pAgent, XPVE_Head)) != NULL) {
	    topTclRetFmt( ip, "%s %s: %s", aV[0], aV[1], s);
	    return TCL_ERROR;
	}
    } else if ( strcasecmp( aV[1], "tail")==0 ) {
	if ( (s=xpAgentChangeCur( pAgent, XPVE_Tail)) != NULL) {
	    topTclRetFmt( ip, "%s %s: %s", aV[0], aV[1], s);
	    return TCL_ERROR;
	}
    } else if ( strcasecmp( aV[1], "next")==0 ) {
	if ( (s=xpAgentChangeCur( pAgent, XPVE_Next)) != NULL) {
	    topTclRetFmt( ip, "%s %s: %s", aV[0], aV[1], s);
	    return TCL_ERROR;
	}
    } else if ( strcasecmp( aV[1], "prev")==0 ) {
	if ( (s=xpAgentChangeCur( pAgent, XPVE_Prev)) != NULL) {
	    topTclRetFmt( ip, "%s %s: %s", aV[0], aV[1], s);
	    return TCL_ERROR;
	}
#endif
    } else {
	if ( pAgent->TclCB != NULL ) {
	    r = (*(pAgent->TclCB))( pAgent->Lbl, ip, aC, aV);
	    if ( r == TCL_OK )	return TCL_OK;
	    if ( r != TCL_CONTINUE ) {
		Tcl_AppendResult(ip,"\nAgent module cmd failed.",NULL);
		return TCL_ERROR;
	    }
	}
	return TCL_CONTINUE;
    }
    return TCL_OK;
}

/**
    Implements the "agent" command.  Trys the global agent commands,
    and then attempts the "agent agentcmd agentspec args ..." form.
**/
int
_xpAgentCmd( ClientData cld, Tcl_Interp *ip, int aC, char **aV) {
    XPAgent	*pAgent;
    char	*cmd, c, *spec;
    int		len, r;

    if ( aC < 2 ) {
	Tcl_AppendResult(ip,
	  "usage:\t",aV[0]," get|info ?agentspec args?",
	  "\nor:\t", aV[0]," agent_cmd agentspec ?args?; with ",NULL);
	_xpAgentCmdsPrs( NULL, ip, NULL, aC, aV);
	return TCL_ERROR;
    }
    cmd = aV[1]; c = *cmd; len = strlen(cmd);
    spec = aV[2];

    /*IF*/ if ( c == 'g' && strncmp(cmd,"get",len)==0 ) {
#ifdef notdef
	dataClass = aC >= 4 ? aV[3] : NULL;
	if ( (pView = xpViewGetBySpec( spec, dataClass, NULL, TOP_TRUE)) == NULL ) {
	    Tcl_AppendResult(ip,aV[0]," ",aV[1],
	      ": Can't find nor create view ``",spec, "''",NULL);
	    return TCL_ERROR;
	}
    	_xpViewInfoFmtTcl( ip, pView, TOP_FALSE);
#endif
	return TCL_ERROR;
    } else if ( c=='i' && strncmp(cmd,"info",len)==0 && aC==2 ) {
	DLPTRLIST_LOOP_FORW_BEGIN( XPTheMachine.AgentList, XPAgent*, pAgent) {
	    if ( pAgent == NULL )	continue;
	    xpAgentFmtTcl( ip, pAgent, TOP_TRUE);
	} DLPTRLIST_LOOP_FORW_END();
    } else {
	if ( aC >= 3 ) {
	    if ( (pAgent=xpAgentFindByName( aV[2], NULL)) == NULL ) {
	        Tcl_AppendResult(ip,aV[0]," ",cmd,": no such agent ``",
		  aV[2],"''", NULL);
	        return TCL_ERROR;
	    }
	    if ( (r=_xpAgentCmdsPrs( pAgent, ip, cmd, aC-3, aV+3)) == TCL_OK )
		return TCL_OK;
	    if ( r!=TCL_CONTINUE ) {
		Tcl_AppendResult(ip,"\n",aV[0]," ",cmd,": failed",NULL);
		return TCL_ERROR;
	    }
	}
	Tcl_AppendResult(ip,aV[0],": invalid command ``",cmd,"''",NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/**
    Process commands for a specific view {pView}.  The view command
    is given by {cmd}, and {aC,aV} are the argments to the command.
**/
    
int
_xpViewCmdsPrs( XPView *pView, Tcl_Interp *ip, char *cmd, int aC, char **aV){
    XPView	*pTgtView;
    char	c;
    int		len, r;
    XPFilter	*pFltr = pView->PntFltr;

    if ( pView==NULL || cmd==NULL ) {
	Tcl_AppendResult(ip,"view_cmds: agents|destroy|info|pipe",NULL);
	return TCL_ERROR;
    }
    c = *cmd; len = strlen(cmd);
    /*IF*/ if ( c == 'a' && strncmp(cmd,"agents",len)==0 ) {
	static char	fCmd[] = "view_cmd consumer";
	XPAgentTcl	*pAg;
	static int	_XpAgentTclSN = 0;

	if ( aC==0 ) {
	    XPVIEWPART_LOOP_FORW_BEGIN( pView, pVP) {
		XPAGENT_LOOP_FORW_BEGIN( pVP, pAgent) {
		    if ( pAgent == NULL )		continue;
		    xpAgentFmtTcl( ip, pAgent, TOP_TRUE);
		} XPAGENT_LOOP_FORW_END();
	    } XPVIEWPART_LOOP_FORW_END();
	} else {
	    if ( aC != 1 ) {
		Tcl_AppendResult(ip,fCmd," args: ?procname?",NULL);
		return TCL_ERROR;
	    }
	    pAg = MEM_ALLOC(XPAgentTcl);
	    xpLabelConstructor( &pAg->Lbl, (TOPPtr)pAg, &_XpAgentTclSN,
	      NULL, "AgentTcl");
	    xpAgentConstructor( &pAg->CM, &pAg->Lbl, pView);
	    pAg->CM.NotifyFunc = _xpAgentTclNotifyCB;
	    pAg->notifyProc = memStrSave( aV[0]);
	    pAg->ip = ip;
	    xpAgentFmtTcl( ip, &pAg->CM, TOP_FALSE);
	}
    } else if ( c == 'd' && strncmp(cmd,"description",len)==0 ) {
	if ( aC!=0 && aC!=1 )	goto extra_args;
	if ( aC==0 ) {
	    Tcl_SetResult(ip,pView->Desc?pView->Desc:"",TCL_STATIC);
	} else {
	    MEM_FREE(pView->Desc);
	    pView->Desc = memStrSave(aV[0]);
	}
    } else if ( c == 'd' && strncmp(cmd,"destroy",len)==0 ) {
	if ( aC!=0 )	goto extra_args;
	xpViewDestroy( pView);
    } else if ( c == 'e' && strncmp(cmd,"event",len)==0 ) {
	if ( aC!=0 )	goto extra_args;
	xpViewNotify( pView, XPNW_Idle, XPVE_Modify, NULL);
    } else if ( c == 'i' && strncmp(cmd,"info",len)==0 ) {
	if ( aC!=0 )	goto extra_args;
    	_xpViewInfoFmtTcl( ip, pView, TOP_FALSE);
    } else if ( c == 'm' && strncmp(cmd,"master",len)==0 ) {
	if ( aC!=0 && aC!=1 )	goto extra_args;
	if ( aC==0 ) {
	    if ( pView->MasterAgent != NULL ) {
		xpAgentFmtTcl( ip, pView->MasterAgent, TOP_FALSE);
	    } else {
		Tcl_ResetResult(ip);
	    }
	} else {
	    ;
	}
    } else if ( c == 'p' && strncmp(cmd,"pipe",len)==0 ) {
	static char	fCmd[] = "view_cmd pipe";
	char		*srcClass, *tgtClass, *pipeClass, *pipeName;
	XPLabel	*pPipeLbl;

	if ( aC != 2 && aC != 3 ) {
	    Tcl_AppendResult(ip,fCmd," args: pipeClass tgtView ?name?",
	      NULL);
	    return TCL_ERROR;
	}
	pipeClass = aV[0];
	pipeName = aC == 3 ? aV[2] : NULL;
	if ( ! xpPipeGetTypeInfo( pipeClass, &srcClass, &tgtClass) ) {
	    Tcl_AppendResult(ip,fCmd,": No pipe class ``",pipeClass,"''",NULL);
	    return TCL_ERROR;
	}
	if ( (pTgtView=xpViewGetBySpec( aV[1], tgtClass, pFltr, TOP_TRUE))==NULL) {
	    Tcl_AppendResult(ip,fCmd,": No view ``",aV[1],"''",NULL);
	    return TCL_ERROR;
	}
	if ( (pPipeLbl=xpPipeCreate( pipeName, pipeClass, pView, pTgtView))
	  == NULL ) {
	    Tcl_AppendResult(ip,fCmd,": Pipe create failed",NULL);
	    return TCL_ERROR;
	}
	Tcl_SetResult( ip, pPipeLbl->Name, TCL_STATIC);
    } else if ( c == 'p' && strncmp(cmd,"primary",len)==0 ) {
	if ( aC!=0 && aC!=1 )	goto extra_args;
	if ( aC==0 ) {
	    if ( pView->PrimaryPart != NULL 
	      && pView->PrimaryPart->RangeOwner != NULL) {
		xpAgentFmtTcl( ip, pView->PrimaryPart->RangeOwner, TOP_FALSE);
	    } else {
		Tcl_ResetResult(ip);
	    }
	} else {
	    ;
	}
    } else {
	if ( pView->Class->CmdFunc != NULL ) {
	    if ( (r=(*(pView->Class->CmdFunc))( ip, pView, cmd, aC, aV)) == TCL_OK )
		return TCL_OK;
	    if (r!=TCL_CONTINUE) {
		Tcl_AppendResult(ip,"view_cmd ",cmd,": cmd failed.",NULL);
		return TCL_ERROR;
	    }
	}
	return TCL_CONTINUE;
    }
    return TCL_OK;

extra_args:
    Tcl_AppendResult(ip,"view_cmd ",cmd,": extra arguments",NULL);
    return TCL_ERROR;
}

/**
    Implements the "view" command.  Trys the global view commands,
    and then attempts the "view viewcmd viewspec args ..." form.
**/
int
_xpViewCmd( ClientData cld, Tcl_Interp *ip, int aC, char **aV) {
    XPView	*pView;
    char	*cmd, c, *spec, *dataClass;
    int		len, r;

    if ( aC < 2 ) {
	Tcl_AppendResult(ip,
	  "usage:\t",aV[0]," get viewspec ?args?",
	  "\nor:\t", aV[0],"viewcmd viewspec ?args?; with ",NULL);
	_xpViewCmdsPrs( NULL, ip, NULL, aC, aV);
	return TCL_ERROR;
    }
    cmd = aV[1]; c = *cmd; len = strlen(cmd);
    spec = aV[2];

    /*IF*/ if ( c == 'g' && strncmp(cmd,"get",len)==0 ) {
	dataClass = aC >= 4 ? aV[3] : NULL;
	topTclCaptureMsgsBegin(ip);
	pView = xpViewGetBySpec( spec, dataClass, NULL, TOP_TRUE);
	topTclCaptureMsgsEnd(ip);
	if ( pView == NULL ) {
	    Tcl_AppendResult(ip,aV[0]," ",aV[1],
	      ": Can't find nor create view ``",spec, "''",NULL);
	    return TCL_ERROR;
	}
    	_xpViewInfoFmtTcl( ip, pView, TOP_FALSE);
    } else {
	if ( aC >= 3 ) {
	    if ( (pView=xpViewFindBySpec( aV[2], NULL, NULL, TOP_FALSE)) == NULL ) {
	        Tcl_AppendResult(ip,aV[0]," ",cmd,": no such view ``",
		  aV[2],"''", NULL);
	        return TCL_ERROR;
	    }
	    if ( (r=_xpViewCmdsPrs( pView, ip, cmd, aC-3, aV+3)) == TCL_OK )
		return TCL_OK;
	    if ( r!=TCL_CONTINUE ) {
		Tcl_AppendResult(ip,"\n",aV[0]," ",cmd,": failed",NULL);
		return TCL_ERROR;
	    }
	}
	Tcl_AppendResult(ip,aV[0],": invalid command ``",cmd,"''",NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}
