#if !defined(_XPVIEW_H_)
#define _XPVIEW_H_ 1
/* 
    RCS: $Header: /vol/cooley/cooley1/kennard/src/xpole/RCS/xpView.h,v 1.1 93/08/27 14:28:01 kennard Exp $
    Author: Kennard White

    Copyright (c) 1990-93 The Regents of the University of California.
    All rights reserved.  See the file "./copyright.h" for full copyright
    notice, limitation of liability, and disclaimer of warranty provisions.
#include "copyright.h"				COPYRIGHTENDKEY
*/

typedef struct _XPViewClass XPViewClass;
typedef struct _XPViewSnap XPViewSnap;
typedef struct _XPViewPart XPViewPart;
typedef struct _XPView XPView;

typedef struct _XPAgent XPAgent, XPConsumer, XPProducer;
typedef XPConsumer XPConsumerMod;
typedef XPProducer XPProducerMod;
typedef struct _XPPipeClassSide XPPipeClassSide;
typedef struct _XPPipeClass XPPipeClass;
typedef struct _XPPipeMod XPPipeMod;
typedef enum _XPNotifyWhen XPNotifyWhen;


/******************************************************************************
    {Label, Data, Producer, Consumer, Pipe} Modules

    These modules are the abstractions through which different parts
    of the program communicate.

    XPView
	A data modules lives within a trajectory, and is essentially a
	control structure for some data within that trajectory.  Each
	data module is bound, via its class name, to a particular type
	of dataset (e.g., CplxSeq, SilarRep).  The dataset that is currently
	bound to the data module is referenced by the {Cur} pointer.
	The referenced data set may either (1) live within a snap {CurSnap}
	of the trajectory; or (2) belong exclusivly to the data module.
	The {Idx} refers to the data module's index within
	the trajectory's DataModVec.  If the referenced data lives in a snap,
	then this index will also index the snap's dataset vector.
	Each data module has a list of consumers that wish
	to be notified when data has changed.  It may also have a producer
	module attached, which is the producer which has exclusive right
	to modify the contents of the data module.

******************************************************************************/

struct _XPViewSnap {
    XPView*		PntView;
    XPViewPart*		PntVP;
};

typedef void (XPViewDestroyFunc)( XPView *pView);

/**
    When a consumer has attached itself to a data module it may be notified
    of changes to the data module and/or set.
    -	Create	This is guarenteed to be the first message that an agent
		receives.  It carries the same meaning as {Switch} below,
		but can also be used by an agent to perform initialization.
    -	Destroy	The agent and/or view is being destroyed, thus the agent
		should cleanup all references.  Either the agent will
		be de-allocated soon, or a Create will be issued.
    -	Modify	The viewsnap was modified in some way.  This change
		is always done without moving the viewsnap, or its contents,
		thus any pointer the consumer had saved away are still valid.
		Some viewsnap's will have additional information internally
		indicating what changed.
    -	Switch  This is really the combination of two events that appear
		identical to the agent:
		a) The current snap for the agent has changed, thus
		   "everything" has changed.  Any pointers saved by the
		   agent are no longer valid.
		b) A "major" modify has taken place, such that even though
		   it may be the same snap, a lot has changed.  This typically
		   occurs when several "modifies" are collapsed together.
    -	ParamChange
		Some parameters of the viewsnap have changed.  Typically,
		this occurs when the consumer wants to see a different region
		of the same view.  This is only meaningful for 
		pipe-driven view's, and is delivered via xpViewParamNotify().
**/

typedef enum _XPViewEvent XPViewEvent;
enum _XPViewEvent {
    XPVE_Null,
    XPVE_Create,
    XPVE_Param,
    XPVE_Modify,
    XPVE_Switch,
    XPVE_Destroy,
    XPVE_Avail,
    XPVE_Head,
    XPVE_Tail,
    XPVE_Next,
    XPVE_Prev,
};

#define XPVEM_Switch TOP_BIT((int)XPVE_Switch)
#define XPVEM_Modify TOP_BIT((int)XPVE_Modify)
#define XPVEM_Param TOP_BIT((int)XPVE_Param)

#define XPNotifySN_Unspec	-1	/* make sn for me */
#define XPNotifySN_Idle		-2	/* make sn for me, do idle update */


/**
	XPView

	A view is either a master or slave (XPViewF_Master).  
	A master view is the source for its own data; it is not generated
	from another view (e.g., Silars attached to planewidget,
	or Fsamps attached to seqwidget).
	A slave view is calculated from another view, typically by a pipe.
	The data expressed in a particular slave view is controled in
	two distinct ways:
	  -	Pipe parameters (e.g., type of silar-mapping)
	  -	View parameters (e.g., the index values of a freq resp)

	A master view has neither pipe params nor view params: that data
	is what it is.  When view params of a slave view are changed,
	{XPVE_ParamChange} is sent to the view's {MasterAgent}.
	Note: the existing of a MasterAgent implies that the view
	is a slave view.

	If a slave view's {MasterAgent} will respond to parameter changes,
	it will set the {XPViewF_Params} flag.  If this flag is not
	set, no agent should modify the parameters.  A master view
	will generally never set this flag.

	Master/slave-ness is distinct from a read-only-ness.  A master
	view may be modified via a user-interface, and a slave view is
	modified whenever its master agent wants to, or whenever the
	view parameters change.  Thus in a writable, master view,
	an agent may modify the data itself but not the parameters.

	To further complicate things, if {XPViewF_Params} is set,
	then each agent "owns" a distinct {XPViewSnap}, and can control
	the parameters of that snap independently.  This is established
	when the agent first attaches itself to the view.  Each such
	"owned" view is described by a {XPViewPart} structure.  Note
	that in general there will be more agents than owners.
	(Every owned view will be produce by the same producer agent.)
**/
struct _XPViewPart {
    XPView*		PntView;	/* the View which we belong to */
    char*		Name;		/* a name assoc. with part */
    TOPMask		Flags;		/* XPViewF_XXX, below */
    TOPMask		Events;		/* pending events */
    XPViewSnap*		DfltVS;		/* default ViewSnap */
    DLPtrListHandle	AgentList;	/* attached agents */
    int			LockCnt;	/* locks onto AgentList (prevent dels)*/
    XPAgent*		RangeOwner;	/* who controls my range */
};

struct _XPViewClass {
    char		*Name;
    int			byteSize;
    int*		SerialPtr;
    void		(*ConstructorFunc) ARGS((XPView *pView, char *class));
    void		(*DestructorFunc) ARGS((XPView *pView));
    int			(*CmdFunc) ARGS_TCL((Tcl_Interp*ip, XPView *pGenDM,
			  char *cmd, int aC, char **aV));
    void		(*NotifyFunc) ARGS((XPViewPart *pVP, XPViewEvent t,
			  XPAgent *pAgOmit));
    int			(*NotifyTclFunc) ARGS_TCL(( XPView *pView,
			  XPViewEvent ev,
			  XPViewSnap *pSnap, Tcl_Interp *ip, char *basecmd));
    XPViewSnap*		(*SnapCreateFunc) ARGS((XPView *pView));
    void		(*SnapDestroyFunc) ARGS((XPViewSnap *pVS));
};

struct _XPView {
    XPLabel		Lbl;
    XPFilter*		PntFltr;
    int			Idx;		/* within traj's DataModVec */
    int			Flags;
    XPXformDomain	XfDomain;
    double		SampleTime;
    char*		Desc;		/* dynamic, unformatted text desc */
    int			LockCnt;
    XPAgent*		MasterAgent;
    XPViewPart*		PrimaryPart;
    DLListHandle	PartList;		/* of XPViewPart */
    XPViewClass*	Class;
};
#define XPViewF_IdleCalc	TOP_BIT(0)
#define XPViewF_DoClean		TOP_BIT(1)
/* #define XPViewF_Master		TOP_BIT(2) */
#define XPViewF_Params		TOP_BIT(3)
#define XPViewF_Valid		TOP_BIT(4)	/* view is update-to-date */
#define XPViewF_UpdateLock	TOP_BIT(5)	/* ViewUpdate() in progress */
#define XPViewF_MultiPart	TOP_BIT(6)

#define XPVIEWPART_LOOP_FORW_BEGIN(pView,pVP) \
    DLLIST_LOOP_FORW_BEGIN((pView)->PartList, XPViewPart*, pVP)
#define XPVIEWPART_LOOP_FORW_END() \
    DLLIST_LOOP_FORW_END()

#define XPAGENT_LOOP_FORW_BEGIN(pVP,pAgent) \
    DLPTRLIST_LOOP_FORW_BEGIN((pVP)->AgentList, XPAgent*, pAgent)
#define XPAGENT_LOOP_FORW_END() \
    DLPTRLIST_LOOP_FORW_END()

#define XPVIEW_LOOP_FORW_BEGIN(pFltr,pView) \
    TOPPTRVEC_LOOP_FORW_BEGIN(&(pFltr)->Views, XPView*, pView)
#define XPVIEW_LOOP_FORW_END() \
    TOPPTRVEC_LOOP_FORW_END()


typedef enum _XPGetSnapHow XPGetSnapHow;
enum _XPGetSnapHow {
    XPGS_Null,
    XPGS_AsIs,			/* just return it */
    XPGS_Upd,			/* updated it first */
};

enum _XPNotifyWhen {
    XPNW_Null,
    XPNW_Now,
    XPNW_Idle,
    XPNW_Defer,
};

typedef void (XPAgentNotifyFunc)( XPAgent *pAgent, XPViewEvent t);
typedef int (XPAgentTclFunc)( XPLabel *pLbl, Tcl_Interp *ip,
  int aC, char **aV);

typedef XPAgentNotifyFunc XPModNotifyFunc;

/**
    The Agent flags have the following meanings:
	LocalCur: This agent wants to step through the trajectory
		indepently of all other agents.  Thus the {CurSnap}
		indicates which snap to use; normally
		{CurView->PntTraj->CurSnap} is used.
    	AutoAdvance: This agent wants to use the "most recent" snap
		available.  {CurSnap} will be automatically adjusted
		to be the most recent snap in the trajectory.
	IdleCreate: There is an outstanding idle-callback that will
		eventually deliver a Create Notify message.
	Alive: A create notify has been delivered and a destroy has not.
	Active: The agent is currently active.  If not active, all
		messages will be recorded but not dilivered.  When
		the agent becomes active again, summaries of all recorded
		messages will be delivered.
**/
typedef TOPMask	XPAgentFlags;
#define XPAgF_LocalCur	TOP_BIT(0)
#define XPAgF_AutoAdvance	TOP_BIT(1)
#define XPAgF_IdleCreate	TOP_BIT(2)
#define XPAgF_Alive		TOP_BIT(3)
#define XPAgF_Active		TOP_BIT(4)

struct _XPAgent {
    XPLabel*		Lbl;
    XPPipeMod*		PntPipe;
    XPAgentFlags	Flags;
    TOPMask		DmFlags;	/* DM-specific flags */
    XPView*		CurView;
    XPViewSnap*		CurSet;
    XPAgentTclFunc*	TclCB;
    XPViewPart*		ViewPart;
    XPNotifyWhen	(*NotifyFunc) ARGS((XPAgent *pAg, XPViewPart *pVP,
			  XPViewEvent ev));
    TOPLogical		(*UpdateFunc) ARGS((XPAgent *pAg, XPViewPart *pVP));
    TOPMask		Events;
};

/*
 * XPView, XPAgent
 */

extern void		xpViewClassAdd( XPViewClass *pNewClass);
extern XPViewClass*	 xpViewClassFindByName( char *name);

extern void		xpViewConstructor( XPView *pView, 
			  XPFilter *pFltr, XPXformDomain xfd);
extern void		xpViewDestructor( XPView *pView);
extern void		xpViewDestroy( XPView *pView);

extern void		xpViewAddAgent( XPView *pView, XPAgent *pAgent, char *partName);
extern void		xpViewRemoveAgent( XPView *pView, XPAgent *pAgent);

extern void		xpViewPartNotifyAgents( XPViewPart *pVP, XPViewEvent ev,
			  XPAgent *pAgOmit);
extern void		xpViewNotifyAgents( XPView *pView,
			  XPViewEvent ev, XPAgent *pAgOmit);
extern void		xpViewPartNotify( XPViewPart *pVP, XPNotifyWhen when,
			  XPViewEvent ev, XPAgent *pAgOmit);
extern void		xpViewNotify( XPView *pView, XPNotifyWhen when,
			  XPViewEvent ev, XPAgent *pAgOmit);



extern XPViewSnap*	xpViewPartGetSnap( XPViewPart *pVP, XPGetSnapHow how);
extern XPViewSnap*	xpViewGetSnap( XPView *pView, XPGetSnapHow how);
extern XPViewPart*	xpAgentGetCurVP( XPAgent *pAg);
extern XPViewSnap*	xpAgentGetSnap( XPAgent *pAg, XPGetSnapHow how);

extern void		xpAgentNotify( XPAgent *pAg, XPNotifyWhen when,
			  XPViewEvent ev);


extern void		xpViewSetMaster( XPView *pView, XPAgent *pAg);

extern void		xpAgentConstructor( XPAgent *pAgent, 
			  XPLabel *pLbl, XPView *pSrcData);
extern void		xpAgentDestructor( XPAgent *pAgent);

extern XPAgent*		xpAgentFindByName ARGS((char *name, char *class));



#ifdef notdef
extern void		xpAgentAddTcl( XPAgent *pAgent, XPAgentTclFunc *pTclCB);
extern int		_xpAgentCmdsPrs( XPAgent *pAgent, Tcl_Interp *ip, 
			  char *cmd, int aC, char **aV);
#endif

extern int		_xpViewCmd ARGS_TCL((ClientData cld, Tcl_Interp *ip,
			  int aC, char **aV));


extern XPView*	xpViewCreateByClass( XPFilter *pFltr,
		  char *name, char *class);
extern XPView*	xpViewFindByName( XPFilter *pFltr, char *name, 
		  char *class);
extern XPView*	xpViewGetByName( XPFilter *pFltr, 
		  char *name, char *class);
extern XPView*	xpViewFindBySpec( char *name, char *class,
		  XPFilter *pFltr, TOPLogical hintB);
extern XPView*	xpViewGetBySpec( char *name, char *class, 
		  XPFilter *pFltr, TOPLogical hintB);

extern char*		xpViewGetFullName( XPView *pView);


/*
 * viewTcl.c
 */
extern void		_xpViewInfoFmtTcl ARGS_TCL((Tcl_Interp *ip,
			  XPView *pView, TOPLogical nestB));
extern int		 _xpViewCmdsPrs ARGS_TCL((XPView *pView, Tcl_Interp *ip,
			  char *cmd, int aC, char **aV));
extern int		_xpViewCmd ARGS_TCL((ClientData cld, Tcl_Interp *ip,
			  int aC, char **aV));

extern int		_xpAgentCmd ARGS_TCL((ClientData cld, Tcl_Interp *ip,
			  int aC, char **aV));

#endif /* _XPVIEW_H_ */
